require 'cgi'

# This class is useful for copying a block of text from a LaTeX file into
# an HTML file. It finds all text between two markers, translates between
# the two formats, and puts the text between the same two markers in the
# HTML file (replacing anything there).
#
# It even works when the LaTeX text is broken up (begin, stuff, end, text
# in the middle, begin, more stuff, end).

class CopyTexToHTML

    SIMPLE_TRANSLATIONS = {
	/\\begin{itemize}/ => '<ul>',
	/\\end{itemize}/ => '</ul>',
	/\\begin{enumerate}/ => '<ol>',
	/\\end{enumerate}/ => '</ol>',
	/\\item/ => '<li>',
	/\\\\/ => '<br>',
	/(''|``)/ => '&quot;',
	# my "dvversion", "riceversion", "twiceversion" commands
	/\\[a-z]+version/ => '<!--#include virtual="/version.txt" -->',
	/\\filename{(.*?)}/ => '<tt>\1</tt>',
	/\\classname{(.*?)}/ => '<tt>\1</tt>',
	/\\emailstyle{(.*?)}/ => '<a href="mailto:\1"><tt>\1</tt></a>',
	/\\textbf{(.*?)}/ => '<b>\1</b>',
	/\\texttt{(.*?)}/ => '<code>\1</code>',
	/\\textit{(.*?)}/ => '<i>\1</i>',
	/\\secname{(.*?)}(\s*,)?/ => '<i>\1</i>',
	/Section\s*\\ref{(.*?)}(\s*,)?/ => 'the User\'s Manual section named',
	/\\ldots/ => '...',
	/\\subsection{(.*?)}/ => '<h2>\1</h2>',
	/%\s*(.*)/ => '<!-- \1 -->',
	/\\copyright{}/ => '&copy;',
	/\\LaTeXe?(\{\})?/ => 'LaTeX2e',
	/\\htlink{(.*?)}{(.*?)}/ => '<a href="\2">\1</a>',
	/\\htmailto{(.*?)}/ => '<a href="mailto:\1">\1</a>', # /
#	/\\\\-/ => '',
    }
    # The comment at the end of the htmailto regex is only there to
    # un-confuse the Emacs ruby-mode font-lock code.

    # Read LaTeX from texfile inbetween begin_marker and end_marker,
    # translate it to HTML, and write it to htmlfile.
    def run(texfile, htmlfile, begin_marker, end_marker)
	write_html(htmlfile, read_tex(texfile, begin_marker, end_marker),
		   begin_marker, end_marker)
    end

    # Given a line of text, perform upon it all of the substutions in
    # xlations and return the line.
    def translate(line, xlations)
	line = CGI.escapeHTML(line)
	xlations.each { | regex, substitution |
	    line.gsub!(regex, substitution)
	}
	return line
    end

    # Open the LaTeX file and read the section between begin_marker and
    # end_marker. Save each line, translating it from LaTeX to HTML. Return
    # the lines as an array.
    def read_tex(texfile, begin_marker, end_marker)
	html = Array.new()
	inside = false
	possibly_new_paragraph = false
	IO.readlines(texfile).each { | line |
	    case line
	    when begin_marker
		inside = true
		next
	    when end_marker
		inside = false
		next
	    end
	    
	    if inside
		if line.strip == ''	# blank line
		    possibly_new_paragraph = true
		else		# non-blank line
		    xlated_text = translate(line, SIMPLE_TRANSLATIONS)
		    if possibly_new_paragraph
			if xlated_text.strip !~ /^\</
			    html << "<p>\n"
			end
			possibly_new_paragraph = false
		    end
		    html << xlated_text
		end
	    end
	}
	return html
    end

    # Given an array of HTML lines and two begin/end regexes, put the HTML
    # between the corresponding begin/end lines in the HTML file.
    def write_html(html_file_name, html_lines, begin_marker, end_marker)
	File.open(html_file_name + '.tmp', 'w') { | f |
	    inside = false
	    IO.readlines(html_file_name).each { | line |
		case line
		when begin_marker
		    inside = true
		    f.print line
		    f.print html_lines
		when end_marker
		    inside = false
		end
		
		f.print line unless inside
	    }
	}
	File.rename(html_file_name + '.tmp', html_file_name)
    end

end
