package jimm.datavision;
import java.util.Observer;
import java.util.Observable;

/**
 * A border is a visual decoration around a report field. Each of its four
 * edges (top, left, bottom, and right) is a <code>BorderEdge</code> (and
 * may be <code>null</code>).
 *
 * @see BorderEdge
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Border
    extends Observable
    implements Dirtiable, Writeable, Cloneable, Observer
{

protected static final int BORDER_LINE_SPACE_MULT = 3;

protected Field field;
protected BorderEdge top, left, bottom, right; // May be null
protected boolean dirty;

/**
 * Constructs a new border for the specified field.
 *
 * @param field a report field
 */
public Border(Field field) {
    this.field = field;
}

/**
 * Returns a clone of this border. All edges are cloned as well.
 *
 * @param other a border
 */
public Object clone() {
    Border b = new Border(field);
    BorderEdge e;

    if ((e = getTop()) == null || e.getNumber() == 0)
	b.setTop(null);
    else
	b.setTop((BorderEdge)e.clone());

    if ((e = getBottom()) == null || e.getNumber() == 0)
	b.setBottom(null);
    else
	b.setBottom((BorderEdge)e.clone());

    if ((e = getLeft()) == null || e.getNumber() == 0)
	b.setLeft(null);
    else
	b.setLeft((BorderEdge)e.clone());

    if ((e = getRight()) == null || e.getNumber() == 0)
	b.setRight(null);
    else
	b.setRight((BorderEdge)e.clone());

    b.setDirty(false);
    return b;
}

public void finalize() {
    if (top != null) top.deleteObserver(this);
    if (left != null) left.deleteObserver(this);
    if (bottom != null) bottom.deleteObserver(this);
    if (right != null) right.deleteObserver(this);
}

public void update(Observable o, Object arg) {
    setChanged();
    notifyObservers(arg);
}

/**
 * Returns <code>true</code> if this border or any of its edges have been
 * modified.
 *
 * @return <code>true</code> if this border has been modified
 */
public boolean isDirty() {
    if (dirty) return true;
    if (top != null && top.isDirty()) return true;
    if (left != null && left.isDirty()) return true;
    if (bottom != null && bottom.isDirty()) return true;
    if (right != null && right.isDirty()) return true;
    return false;
}

/**
 * Sets or clears the dirty bit. If <i>dirty</i> is <code>false</code>,
 * clears the dirty bit for each edge.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) {
    this.dirty = dirty;
    if (!this.dirty) {
	if (top != null) top.setDirty(false);
	if (left != null) left.setDirty(false);
	if (bottom != null) bottom.setDirty(false);
	if (right != null) right.setDirty(false);
    }
}

/**
 * Returns the border's top edge. May return <code>null</code>.
 *
 * @return the top edge
 */
public BorderEdge getTop() { return top; }

/**
 * Sets the top edge. <i>newTop</i> may be <code>null</code>.
 *
 * @param newTop the new edge
 */
public void setTop(BorderEdge newTop) {
    if (top != newTop) {
	if (top != null) top.deleteObserver(this);
	top = newTop;
	if (top != null) top.addObserver(this);
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the border's left edge. May return <code>null</code>.
 *
 * @return the left edge
 */
public BorderEdge getLeft() { return left; }

/**
 * Sets the left edge. <i>newLeft</i> may be <code>null</code>.
 *
 * @param newLeft the new edge
 */
public void setLeft(BorderEdge newLeft) {
    if (left != newLeft) {
	if (left != null) left.deleteObserver(this);
	left = newLeft;
	if (left != null) left.addObserver(this);
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the border's bottom edge. May return <code>null</code>.
 *
 * @return the bottom edge
 */
public BorderEdge getBottom() { return bottom; }

/**
 * Sets the bottom edge. <i>newBottom</i> may be <code>null</code>.
 *
 * @param newBottom the new edge
 */
public void setBottom(BorderEdge newBottom) {
    if (bottom != newBottom) {
	if (bottom != null) bottom.deleteObserver(this);
	bottom = newBottom;
	if (bottom != null) bottom.addObserver(this);
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the border's right edge. May return <code>null</code>.
 *
 * @return the right edge
 */
public BorderEdge getRight() { return right; }

/**
 * Sets the right edge. <i>newRight</i> may be <code>null</code>.
 *
 * @param newRight the new edge
 */
public void setRight(BorderEdge newRight) {
    if (right != newRight) {
	if (right != null) right.deleteObserver(this);
	right = newRight;
	if (right != null) right.addObserver(this);
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * For each edge, hand the lines that make up that edge to the specified
 * line drawer's <code>drawLine</code> method.
 *
 * @param ld a line drawer
 */
public void eachLine(LineDrawer ld) {
    Rectangle rect = field.bounds;
    Section section = field.section;
    BorderEdge edge;
    Line line;

    edge = top;
    if (edge != null && edge.getThickness() > 0 && edge.getNumber() > 0) {
	line = new Line(section, edge.getThickness(), new Point(rect.x, rect.y),
			new Point(rect.x + rect.width, rect.y));
	for (int i = 0; i < edge.getNumber(); ++i) {
	    ld.drawLine(line);
	    line.points[0].y += BORDER_LINE_SPACE_MULT;
	    line.points[1].y += BORDER_LINE_SPACE_MULT;
	}
    }

    edge = bottom;
    if (edge != null && edge.getThickness() > 0 && edge.getNumber() > 0) {
	line = new Line(section, edge.getThickness(),
			new Point(rect.x, rect.y + rect.height),
			new Point(rect.x + rect.width, rect.y + rect.height));
	for (int i = 0; i < edge.getNumber(); ++i) {
	    ld.drawLine(line);
	    line.points[0].y -= BORDER_LINE_SPACE_MULT;
	    line.points[1].y -= BORDER_LINE_SPACE_MULT;
	}
    }

    edge = left;
    if (edge != null && edge.getThickness() > 0 && edge.getNumber() > 0) {
	line = new Line(section, edge.getThickness(), new Point(rect.x, rect.y),
			new Point(rect.x, rect.y + rect.height));
	for (int i = 0; i < edge.getNumber(); ++i) {
	    ld.drawLine(line);
	    line.points[0].x += BORDER_LINE_SPACE_MULT;
	    line.points[1].x += BORDER_LINE_SPACE_MULT;
	}
    }

    edge = right;
    if (edge != null && edge.getThickness() > 0 && edge.getNumber() > 0) {
	line = new Line(section, edge.getThickness(),
			new Point(rect.x + rect.width, rect.y),
			new Point(rect.x + rect.width, rect.y + rect.height));
	for (int i = 0; i < edge.getNumber(); ++i) {
	    ld.drawLine(line);
	    line.points[0].x -= BORDER_LINE_SPACE_MULT;
	    line.points[1].x -= BORDER_LINE_SPACE_MULT;
	}
    }
}
	    
/**
 * Writes this border as an XML tag. Asks each edge to write itself as well.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<border>");
    out.indent();

    if (top != null) top.writeXML(out, "top");
    if (bottom != null) bottom.writeXML(out, "bottom");
    if (left != null) left.writeXML(out, "left");
    if (right != null) right.writeXML(out, "right");

    out.outdent();
    out.println("</border>");
}

}

