package jimm.datavision;
import jimm.datavision.layout.*;
import jimm.datavision.gui.DesignWin;
import java.io.*;

/**
 * This is the DataVision application class. It opens a design window
 * for each report specified on the command line. If none are specified,
 * it openes a new design window on a new, empty report.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class DataVision {

protected static final String DEFAULT_CHARACTER_SEPARATOR = ",";

char layoutEngineChoice;
String dbPassword;
String paramXMLFile;

/**
 * This main application method opens a design window for each report
 * specified on the command line. If none are specified, it openes a new
 * design window on a new, empty report.
 *
 * @param args command line array; each element is assumed to be a report
 * file name.
 */
public static void main(String[] args) {
    Getopts g = new Getopts("cdhlxnp:s:r:", args);
    DataVision dv = new DataVision();

    dv.layoutEngineChoiceFromOptions(g); // Grab layout engine
    if (g.hasOption('n')) {	// No password/empty password
	if (g.hasOption('p'))
	    usage("-n and -p are mutually exclusive");
	dv.dbPassword = "";
    }
    else
	dv.dbPassword = g.option('p', null); // Grab password; default is null

    if (dv.hasLayoutEngine() && !g.hasOption('n') && !g.hasOption('p'))
	usage("when layout engine specified either -n or -p is required");

    dv.paramXMLFile = g.option('r', null); // Parameter XML file name or null

    if (g.argc() == 0) {
	if (dv.hasLayoutEngine()) // Have layout engine but no file
	    usage("xml_file required when layout engine specified");
	else
	    new DesignWin(null).show();	// No layout engine, no file
    }
    else {			// Loop over input files
	for (int i = 0; i < g.argc(); ++i) {
	    String fileName = g.argv(i);
	    try {
		if (dv.hasLayoutEngine()) {
		    Report report = new Report();
		    if (dv.dbPassword != null)
			report.setDatabasePassword(dv.dbPassword);
		    report.readFile(fileName); // Must come after password set
		    if (dv.paramXMLFile != null) // Must come after file read
			report.setParameterXMLFile(dv.paramXMLFile);
		    report.setLayoutEngine(dv.createLayoutEngine(fileName, g));
		    report.run();
		}
		else {
		    new DesignWin(args[i], dv.dbPassword).show();
		}
	    }
	    catch (Exception e) { // Global catch-all
		ErrorHandler.error(e);
	    }
	}
    }
}

boolean hasLayoutEngine() { return layoutEngineChoice != '\0'; }

void layoutEngineChoiceFromOptions(Getopts g) {    
    String errMsg = "only one layout engine may be specified";

    layoutEngineChoice = '\0';
    if (g.hasOption('c'))
	layoutEngineChoice = 'c';
    if (g.hasOption('d')) {
	if (layoutEngineChoice != '\0') usage(errMsg);
	layoutEngineChoice = 'd';
    }
    if (g.hasOption('h')) {
	if (layoutEngineChoice != '\0') usage(errMsg);
	layoutEngineChoice = 'h';
    }
    if (g.hasOption('l')) {
	if (layoutEngineChoice != '\0') usage(errMsg);
	layoutEngineChoice = 'l';
    }
    if (g.hasOption('x')) {
	if (layoutEngineChoice != '\0') usage(errMsg);
	layoutEngineChoice = 'x';
    }
}

LayoutEngine createLayoutEngine(String fileName, Getopts g) throws IOException
{
    LayoutEngine le = null;

    // Create new file name.
    String fileNameSansExtension;
    int pos = fileName.lastIndexOf('.');
    if (pos == -1)
	fileNameSansExtension = new String(fileName);
    else
	fileNameSansExtension = fileName.substring(0, pos);

    String outFileName = null;
    PrintWriter out = null;
    switch (layoutEngineChoice) {
    case 'c':
	char sep = g.option('s', DEFAULT_CHARACTER_SEPARATOR).charAt(0);
	outFileName = fileNameSansExtension + (sep == ',' ? ".csv" : ".tab");
	if (outFileName.equals(fileName))
	    outFileName = fileNameSansExtension + "_out"
		+ (sep == ',' ? ".csv" : ".tab");
	out = new PrintWriter(new FileWriter(outFileName));
	le = new CharSepLE(out, sep);
	break;
    case 'd':
	outFileName = fileNameSansExtension + ".sgml";
	if (outFileName.equals(fileName))
	    outFileName = fileNameSansExtension + "_out.sgml";
	out = new PrintWriter(new FileWriter(outFileName));
	le = new DocBookLE(out);
	break;
    case 'h':
	outFileName = fileNameSansExtension + ".html";
	if (outFileName.equals(fileName))
	    outFileName = fileNameSansExtension + "_out.html";
	out = new PrintWriter(new FileWriter(outFileName));
	le = new HTMLLE(out);
	break;
    case 'l':
	outFileName = fileNameSansExtension + ".tex";
	if (outFileName.equals(fileName))
	    outFileName = fileNameSansExtension + "_out.tex";
	out = new PrintWriter(new FileWriter(outFileName));
	le = new LaTeXLE(out);
	break;
    case 'x':
	outFileName = fileNameSansExtension + ".xml";
	if (outFileName.equals(fileName))
	    outFileName = fileNameSansExtension + "_out" + ".xml";
	IndentWriter iout = new IndentWriter(new FileWriter(outFileName));
	le = new XMLLE(iout);
	break;
    }
    return le;
}

/**
 * Prints a usage message and an optional extra error message to System.err
 * and exits.
 *
 * @param errMsg string to print; may be <code>null</code>
 */
public static void usage(String errMsg) {
    if (errMsg != null)
	System.err.println(errMsg);
    System.err.println(
"usage: jimm.datavision.DataVision [(-c|-d|-h|-l|-x)] [-s char]\n" +
"        [-n | -p db_password] [-r param_xml_file] [xml_file...]\n" +
"\n" +
"    -c               Use char-delimited layout engine (default separator\n" +
"                         is comma; default output to xml_file.csv)\n" +
"    -d               Use DocBook layout engine (output xml_file.sgml)\n" +
"    -h               Use HTML layout engine (output xml_file.html)\n" +
"    -l               Use LaTeX layout engine (output xml_file.tex)\n" +
"    -x               Use XML layout engine (output xml_file_out.xml)\n" +
"    -s char          Character for -c char-delimited layout engine\n" +
"                         (if whitespace, use quotes around char)\n" +
"    -n               No password required/empty password\n" +
"    -p db_password   Database password\n" +
"    -r param_xml     Read XML file for parameter values\n" +
"\n" +
"The options -n and -p are mutually exclusive.\n" +
"If a layout engine specified, xml_file and either -n or -p are required.\n");

    System.exit(1);
}

}
