package jimm.datavision;
import java.util.Observer;
import java.util.Observable;

/**
 * <code>Element</code> is the abstract superclass of <code>Field</code>
 * and <code>Line</code>. These are the visual elements of a report section.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */

public abstract class Element
    extends Observable
    implements Dirtiable, Writeable, Observer
{

protected Section section;
protected boolean visible;
protected boolean dirty;

/**
 * Constructor.
 *
 * @param section the report section containing this element.
 */
public Element(Section section) {
    this.section = section;
    visible = true;
}

public void update(Observable o, Object arg) {
    setChanged();
    notifyObservers(arg);
}

/**
 * Returns <code>true</code> if this element has been modified.
 *
 * @return <code>true</code> if this element has been modified
 */
public boolean isDirty() { return dirty; }

/**
 * Sets or clears the dirty bit.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) { this.dirty = dirty; }

/**
 * Returns the section that containts this field.
 *
 * @return the section
 */
public Section getSection() { return section; }

/**
 * Modifies the section to which this field belongs. Only called by
 * section itself.
 *
 * @param s the section
 */
public void setSection(Section s) {
    if (s != section) {
	section = s;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the visible state of this element.
 *
 * @return the visible state
 */
public boolean isVisible() { return visible; }

/**
 * Sets the visibility of this element.
 *
 * @param newVisible the new visible state
 */
public void setVisible(boolean newVisible) {
    if (visible != newVisible) {
	visible = newVisible;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Writes this element as an XML tag. This abstract method is overridden
 * by the <code>Field</code> and <code>Line</code> subclasses.
 *
 * @param out a writer that knows about indenting XML tags
 */
public abstract void writeXML(IndentWriter w);

}
