package jimm.datavision;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.text.DecimalFormat;

/**
 * The abstract superclass of visual report fields that display text
 * labels, database columns, special values, subtotal values, and formulas.
 * A field has a bounds rectangle that determines its position within a
 * section and an associated format for determining how to print the field.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public abstract class Field extends Element implements Draggable {

public static final double DEFAULT_WIDTH = 120;
public static final double DEFAULT_HEIGHT = 16;

protected static HashMap fields = new HashMap();

protected Long id;
protected Rectangle bounds;
protected Format format;	// Possibly null
protected Border border;	// Possibly null
protected Object value;		// String or id (column, formula, etc.)

/**
 * This factory method constructs and returns a new instance of a subclass
 * of <code>Field</code> based on the <i>type</i> string.
 * <p>
 * If <i>id</i> is <code>null</code>, generates a new id number. This number
 * is one higher than any previously-seen id number. This does <em>not</em>
 * guarantee that no later field will be created manually with the same
 * id number.
 *
 * @param id the unique identifier for the new field; if <code>null</code>,
 * generate a new id
 * @param section the report section containing the field
 * @param type one of "special", "text", "column", "subtotal", "formula",
 * or "parameter"; found in report XML
 * @see Utils#generateNewId
 */
public static Field create(Long id, Section section, String type, Object value)
{
    if (id == null)
	id = Utils.generateNewId(fields.keySet());

    if (type == null || type.length() == 0)
	throw new IllegalArgumentException("Field " + id
					   + ": type may not be empty");
	
    type = type.toLowerCase();
    if (type.equals("special"))
	return new SpecialField(id, section, value);
    else if (type.equals("text"))
	return new TextField(id, section, value);
    else if (type.equals("column"))
	return new ColumnField(id, section, value);
    else if (type.equals("subtotal"))
	return new SubtotalField(id, section, value);
    else if (type.equals("formula"))
	return new FormulaField(id, section, value);
    else if (type.equals("parameter"))
	return new ParameterField(id, section, value);

    throw new IllegalArgumentException("Field " + id
				       + ": unknown field type \"" + type
				       + "\"");
}

/**
 * Creates a field from a drag string. <i>str</code> should a string created
 * by some field's <code>dragString</code> method.
 *
 * @param str a drag string
 * @return a new field
 * @see #dragString
 */
public static Field createFromDragString(String str) {
    int pos = str.indexOf(":");
    if (pos == -1)
	return null;

    String type = str.substring(0, pos);
    String value = str.substring(pos + 1);
    return create(null, null, type, value);
}

/**
 * Adds a field to the static map of all fields. Called from constructors.
 *
 * @param f a field
 */
protected static void addInstance(Field f) {
    fields.put(f.id, f);
}

/**
 * Given an id, returns the field that has that id. If no field with the
 * specified id exists, returns <code>null</code>.
 *
 * @return a field, or <code>null</code> if no field with the specified
 * id exists
 */
public static Field find(Object id) {
    if (id instanceof String)
	id = new Long((String)id);
    return (Field)fields.get(id);
}

/**
 * Constructor.
 *
 * @param id the unique identifier for the new field
 * @param section the report section containing the field
 * @param value the value this field represents visually
 */
protected Field(Long id, Section section, Object value) {
    super(section);
    this.id = id;
    this.value = value;
    bounds = new Rectangle(0, 0, DEFAULT_WIDTH, DEFAULT_HEIGHT);
    bounds.addObserver(this);
    Field.addInstance(this);
}

public void finalize() {
    bounds.deleteObserver(this);
    if (format != null) format.deleteObserver(this);
    if (border != null) border.deleteObserver(this);
}

/**
 * Returns <code>true</code> if this field or any of its attributes have been
 * modified.
 *
 * @return <code>true</code> if this field has been modified
 */
public boolean isDirty() {
    if (super.isDirty()) return true;
    return bounds.isDirty() || (format != null && format.isDirty())
	|| (border != null && border.isDirty());
}

/**
 * Sets or clears the dirty bit. When clearing, the dirty bits of all
 * attributes are also cleared.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) {
    super.setDirty(dirty);
    if (!dirty) {
	bounds.setDirty(false);
	if (format != null) format.setDirty(false);
	if (border != null) border.setDirty(false);
    }
}

/**
 * Returns the unique id for this field.
 *
 * @return the unique id
 */
public Long getId() { return id; }

/**
 * Returns the bounds rectangle for this field.
 *
 * @return the bounds rectangle
 */
public Rectangle getBounds() { return bounds; }

/**
 * Sets the bounds rectangle.
 *
 * @param newBounds the new bounds rectangle
 */
public void setBounds(Rectangle newBounds) {
    if (bounds != newBounds) {
	bounds.deleteObserver(this);
	bounds = newBounds;
	bounds.addObserver(this);
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the format for this field. May return <code>null</code>.
 *
 * @return the format, possibly <code>null</code>
 */
public Format getFormat() { return format; }

/**
 * Sets the format. If this field already has a format, you can just modify
 * it instead of giving it a completely new one.
 *
 * @param newBounds the format
 */
public void setFormat(Format newFormat) {
    if (format != newFormat) {
	if (format != null) format.deleteObserver(this);
	format = newFormat;
	if (format != null) format.addObserver(this);
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the border for this field. May return <code>null</code>.
 *
 * @return the border, possibly <code>null</code>
 */
public Border getBorder() { return border; }

/**
 * Sets the border.
 *
 * @param newBorder the new border
 */
public void setBorder(Border newBorder) {
    if (border != newBorder) {
	if (border != null) border.deleteObserver(this);
	border = newBorder;
	if (border != null) border.addObserver(this);
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the value for this field. May return <code>null</code>.
 *
 * @return the border, possibly <code>null</code>
 */
public Object getValue() { return value; }

/**
 * Sets the value.
 *
 * @param newValue the new value
 */
public void setValue(Object newValue) {
    if (value != newValue) {
	value = newValue;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns a string ready for display in a report.
 *
 * @param obj an object; usually the return value from <code>getValue</code>
 * @return a string, possibly new, possibly the same
 */
public String formatted(Object obj) {
    if (format == null || format.format == null)
	return obj.toString();

    if (obj instanceof Number) {
	DecimalFormat formatter = new DecimalFormat(format.format);
	return formatter.format((Number)obj);
    }
    else if (obj instanceof Date) {
	SimpleDateFormat formatter = new SimpleDateFormat(format.format);
	return formatter.format((Date)obj);
    }
    else
	return obj.toString();
}

/**
 * Returns the string that specifies this field's type in the report XML.
 * Must be overridden.
 *
 * @return a string representing this field's type; used in XML files
 */
public abstract String typeString();

/**
 * Returns a string representing the field in the GUI during report design.
 *
 * @return a string useful for display in the design GUI
 */
public String designLabel() { return formulaString(); }

/**
 * Returns a string representing the field as it appears in a formula.
 *
 * @return a string useful in a formula
 */
public abstract String formulaString();

/**
 * Returns <code>true</code> if this field contains a reference to the
 * specified field. Most fields return <code>false</code>; only a subtotal
 * or formula field would possibly return <code>true</code>.
 *
 * @param f a field
 * @return <code>true</code> if this field contains a reference to the
 * specified field
 * @see FormulaField
 * @see SubtotalField
 */
public boolean refersTo(Field f) {
    return false;
}

/**
 * Returns <code>true</code> if this field contains a reference to the
 * specified formula. Most fields return <code>false</code>; only a
 * subtotal or formula field would possibly return <code>true</code>.
 *
 * @param f a formula
 * @return <code>true</code> if this field contains a reference to the
 * specified field
 * @see FormulaField
 * @see SubtotalField
 */
public boolean refersTo(Formula f) {
    return false;
}

/**
 * Returns <code>true</code> if this field contains a reference to the
 * specified parameter. Most fields return <code>false</code>; only a
 * subtotal or formula field would possibly return <code>true</code>.
 *
 * @param p a parameter
 * @return <code>true</code> if this field contains a reference to the
 * specified field
 * @see FormulaField
 * @see SubtotalField
 */
public boolean refersTo(Parameter p) {
    return false;
}

/**
 * Returns this fields formatted value, ready for display in the report.
 *
 * @return the report display string
 */
public String toString() {
    if (!visible) return null;

    Object v = getValue();
    if (v == null) return "";

    return formatted(v);
}

/**
 * Writes this field as an XML tag. Writes bounds, border, and format.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<field id=\"" + id + "\" type=\"" + typeString()
		+ "\" value=\"" + value + "\">");
    out.indent();

    writeFieldGuts(out);

    out.outdent();
    out.println("</field>");
}

/**
 * Writes objects contained within this field (bounds, border, and format.
 *
 * @param out a writer that knows about indenting XML tags
 */
protected void writeFieldGuts(IndentWriter out) {
    bounds.writeXML(out);
    if (format != null) format.writeXML(out);
    if (border != null) border.writeXML(out);
}

}
