package jimm.datavision;
import java.util.Observable;

/**
 * A format describes how to display a field. It specifies font, size,
 * attributes (bold, italic, underline, wrap), alignment, and print
 * format.
 * <p>
 * Not all fields of a format are used. Currently, underline, wrap, and
 * print format are ignored.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Format
    extends Observable
    implements Dirtiable, Writeable, Cloneable
{

/** Used to specify left alignment. */
public static final int ALIGN_LEFT = 0;
/** Used to specify center alignment. */
public static final int ALIGN_CENTER = 1;
/** Used to specify right alignment. */
public static final int ALIGN_RIGHT = 2;
/** The default format, used when a field has no format specified. */
public static final Format DEFAULT = new Format();

protected static final int DEFAULT_SIZE = 11;

protected String font;		// Name
protected double size;		// In points
protected boolean bold;
protected boolean italic;
protected boolean underline;
protected boolean wrap;
protected int align;		// ALIGN_{LEFT,CENTER,RIGHT}
protected String format;	// Output formatting string
protected boolean dirty;

/**
 * Returns an <code>ALIGN_*</code> constant, given one of "left",
 * "center", or "right". If the specified string is null or is not one
 * of these values, <code>ALIGN_LEFT</code> is returned.
 *
 * @param s the string "left", "center", or "right" (case is not
 * significant)
 * @return one of <code>ALIGN_LEFT</code>, <code>ALIGN_CENTER</code>,
 * or <code>ALIGN_RIGHT</code>
 */
public static int alignFromString(String s) {
    if (s == null) return ALIGN_LEFT;
    String copy = s.toLowerCase();
    if ("center".equals(copy)) return ALIGN_CENTER;
    if ("right".equals(copy)) return ALIGN_RIGHT;
    return ALIGN_LEFT;
}

/**
 * Given an <code>ALIGN_*</code> constant, return the string used
 * to represent that value in a report XML file. If <i>align</i> is not
 * one of those values, returns "left".
 *
 * @param align one of "left", "center", or "right"
 */
public static String alignToString(int align) {
    switch (align) {
    case ALIGN_RIGHT: return "right";
    case ALIGN_CENTER: return "center";
    default: return "left";
    }
}

/**
 * Constructor. Fills in default values.
 */
public Format() {
    size = DEFAULT_SIZE;
    bold = italic = underline = wrap = false;
    align = ALIGN_LEFT;
}

/**
 * Returns a clone of this format.
 *
 * @param other a format
 */
public Object clone() {
    Format f = new Format();
    f.setSize(getSize());
    f.setBold(isBold());
    f.setItalic(isItalic());
    f.setUnderline(isUnderline());
    f.setWrap(isWrap());
    f.setAlign(getAlign());
    f.setFormat(getFormat());
    return f;
}

/**
 * Returns <code>true</code> if this format has been modified.
 *
 * @return <code>true</code> if this format has been modified
 */
public boolean isDirty() { return dirty; }

/**
 * Sets or clears the dirty bit.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) { this.dirty = dirty; }

/**
 * Returns the font for this format.
 *
 * @return the font
 */
public String getFont() { return font; }

/**
 * Sets the font
 *
 * @param newFont the new font
 */
public void setFont(String newFont) {
    if (font != newFont && (font == null || !font.equals(newFont))) {
	font = newFont;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the size for this format.
 *
 * @return the size
 */
public double getSize() { return size; }

/**
 * Sets the size
 *
 * @param newSize the new size
 */
public void setSize(double newSize) {
    if (size != newSize) {
	size = newSize;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the bold state.
 *
 * @return the bold state
 */
public boolean isBold() { return bold; }

/**
 * Sets the bold state.
 *
 * @param newBold the new value
 */
public void setBold(boolean newBold) {
    if (bold != newBold) {
	bold = newBold;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the italic state.
 *
 * @return the italic state
 */
public boolean isItalic() { return italic; }

/**
 * Sets the italic state.
 *
 * @param newItalic the new value
 */
public void setItalic(boolean newItalic) {
    if (italic != newItalic) {
	italic = newItalic;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the underline state.
 *
 * @return the underline state
 */
public boolean isUnderline() { return underline; }

/**
 * Sets the underline state.
 *
 * @param newUnderline the new underline state
 */
public void setUnderline(boolean newUnderline) {
    if (underline != newUnderline) {
	underline = newUnderline;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the wrap state.
 *
 * @return the wrap state
 */
public boolean isWrap() { return wrap; }

/**
 * Sets the wrap state.
 *
 * @param newWrap the new wrap state
 */
public void setWrap(boolean newWrap) {
    if (wrap != newWrap) {
	wrap = newWrap;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the alignment.
 *
 * @return one of the <code>ALIGN_*</code> values
 */
public int getAlign() { return align; }

/**
 * Sets the alignment.
 *
 * @param newAlign one of the <code>ALIGN_*</code> values
 */
public void setAlign(int newAlign) {
    if (align != newAlign) {
	align = newAlign;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the format string for this field. May return <code>null</code>.
 *
 * @return the format string, possibly <code>null</code>
 */
public String getFormat() { return format; }

/**
 * Sets the format string.
 *
 * @param newFormat the new format string
 */
public void setFormat(String newFormat) {
    if (format != newFormat && (format == null || !format.equals(newFormat))) {
	format = newFormat;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns a string representation of this format. Mainly used for
 * debugging.
 *
 * @return pretty string, inn't it?
 */
public String toString() {
    return "Format[size=" + size + ", bold=" + bold + ", italic=" + italic
	+ ", underline=" + underline + ", wrap=" + wrap + "align="
	+ (align == ALIGN_LEFT
	   ? "left" : (align == ALIGN_CENTER ? "center" : "right"))
	+ ", format=" + format + ", dirty=" + dirty + "]";
}    

/**
 * Writes this format as an XML tag.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.print("<format");
    if (font != null) out.print(" font=\"" + Utils.escapeXML(font) + "\"");
    if (size != 0) out.print(" size=\"" + size + "\"");
    if (bold) out.print(" bold=\"" + bold + "\"");
    if (italic) out.print(" italic=\"" + italic + "\"");
    if (underline) out.print(" underline=\"" + underline + "\"");
    if (wrap) out.print(" wrap=\"" + wrap + "\"");
    out.print(" align=\"" + Format.alignToString(align) + "\"");
    if (format != null) out.print(" format=\"" + format + "\"");
    out.println(" />");
}

}
