package jimm.datavision;
import jimm.datavision.sql.Column;
import java.io.*;
import java.util.*;

/**
 * A formula is an expression evaluated at runtime. It may contain database
 * column values, other formulas, special values, subtotals, strings, or
 * numbers.
 * <p>
 * Formulas use a RPN (Reverse Polish Notation) syntax, because that is
 * easy to parse and execute. This will change some day.
 * <p>
 * Before being evaluated, the following substitutions are made withing
 * the evaluation string of a formula:
 * <ul>
 * <li>{<i>table_name.column_name</i>} is replaced by the current value
 * of the column <i>table_name.column_name</i>.</li>
 * <li>{@<i>id_number</i>} is replaced by the results of evaluating the
 * formula whose id is <i>id_number</i>.</li>
 * <li> {%<i>special_value_name</i>} is replaced by a special value
 * (report title, report run date, page number, or record number).</li>
 * <li> {?<i>id_number</i>} is replaced by a parameter value (string,
 * number, or date).</li>
 * <ul>
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Formula
    extends Observable
    implements Writeable, Draggable, Observer
{

protected static final HashMap formulas = new HashMap();

protected Long id;
protected Report report;
protected String name;
protected String evalString;
protected Stack stack;
protected String tok;
protected int tokType;
protected boolean dirty;
protected ArrayList observedContents;

/**
 * Returns the formula with the specified id or <code>null</code>
 * if one is not found.
 *
 * @param id the formula id
 * @return the formula with that id or <code>null</code> if one is not found
 */
public static Formula find(Object id) {
    if (id instanceof String)
	id = new Long((String)id);
    return (Formula)formulas.get(id);
}

/**
 * Returns the formula with the specified name or <code>null</code>
 * if one is not found.
 *
 * @param name the name string
 * @return the formula with that name or <code>null</code> if one is not found
 */
public static Formula findName(String name) {
    if (name == null)
	return null;

    for (Iterator iter = formulas.values().iterator(); iter.hasNext(); ) {
	Formula f = (Formula)iter.next();
	if (name.equals(f.getName()))
	    return f;
    }
    return null;
}

/**
 * Constructor.
 *
 * @param id the unique identifier for the new formula; if <code>null</code>,
 * generate a new id
 * @param report the report containing this formula
 * @param name the formula name
 */
public Formula(Long id, Report report, String name) {
    this(id, report, name, null);
}

/**
 * Constructor. If <i>id</i> is <code>null</code>, generates a new id number.
 * This number is one higher than any previously-seen id number. This does
 * <em>not</em> guarantee that no later formula will be created manually with
 * the same id number.
 *
 * @param id the unique identifier for the new formula; if <code>null</code>,
 * generate a new id
 * @param report the report containing this formula
 * @param name the formula name
 * @param evalString the string to evaulate at runtime.
 */
public Formula(Long id, Report report, String name, String evalString) {
    if (id == null)
	id = Utils.generateNewId(formulas.keySet());

    this.report = report;
    this.id = id;
    this.name = name;
    this.evalString = evalString;
    formulas.put(id, this);
    dirty = false;

    // I'd like to start observing the contents of the eval string here,
    // but the other formulas may not yet be defined (for example, when
    // reading in formulas from an XML file). That's why we start observing
    // contents when someone asks for the eval string.
    observedContents = null;
}

public void finalize() {
    stopObservingContents();
    try {
	super.finalize();
    }
    catch (Throwable t) {}
}

public void update(Observable o, Object arg) {
    setChanged();
    notifyObservers(arg);
}

/**
 * Returns <code>true</code> if this formula has been modified.
 *
 * @return <code>true</code> if this formula has been modified
 */
public boolean isDirty() { return dirty; }

/**
 * Sets or clears the dirty bit.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) { this.dirty = dirty; }

/**
 * Returns the unique id for this field.
 *
 * @return the unique id
 */
public Long getId() { return id; }

/**
 * Returns the name for this formula.
 *
 * @return the name
 */
public String getName() { return name; }

/**
 * Sets the name.
 *
 * @param newName the new name
 */
public void setName(String newName) {
    if (name != newName && (name == null || !name.equals(newName))) {
	name = newName;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the eval string for this formula.
 *
 * @return the eval string
 */
public String getEvalString() {
    // I'd like to start observing the contents of the eval string as soon
    // as we are constructed, but the other formulas may not yet be defined
    // (for example, when reading in formulas from an XML file). That's why
    // we start observing contents when someone asks for the eval string.
    if (observedContents == null)
	startObservingContents();

    return evalString;
}

/**
 * Sets the eval string.
 *
 * @param newEvalString the new eval string
 */
public void setEvalString(String newEvalString) {
    if (evalString != newEvalString
	&& (evalString == null || !evalString.equals(newEvalString)))
    {
	stopObservingContents();
	evalString = newEvalString;
	startObservingContents();

	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Starts observing all formulas and parameters referenced by this formula.
 */
protected void startObservingContents() {
    observedContents = new ArrayList();	// Even if evalString is null

    if (evalString == null || evalString.length() == 0)
	return;

    // Formulas
    int pos = 0;
    while ((pos = evalString.indexOf("{@", pos)) != -1) {
	int pos2 = evalString.indexOf("}", pos + 1);
	if (pos2 != -1) {
	    Formula f = Formula.find(evalString.substring(pos + 2, pos2));
	    observedContents.add(f);
	    f.addObserver(this);
	}
	pos = pos2 + 1;
    }

    // Parameters
    pos = 0;
    while ((pos = evalString.indexOf("{?", pos)) != -1) {
	int pos2 = evalString.indexOf("}", pos + 1);
	if (pos2 != -1) {
	    Parameter p = Parameter.find(evalString.substring(pos + 2, pos2));
	    observedContents.add(p);
	    p.addObserver(this);
	}
	pos = pos2 + 1;
    }
}

/**
 * Stops observing all formulas referenced by this formula.
 */
protected void stopObservingContents() {
    if (observedContents != null) {
	for (Iterator iter = observedContents.iterator(); iter.hasNext(); )
	    ((Observable)iter.next()).deleteObserver(this);
	observedContents = null;
    }
}

/**
 * Returns the eval string fit for human consumption. This mainly means
 * that we substitute formula and parameter numbers with names. Called from
 * a formula editor. This code assumes that curly braces are never nested.
 *
 * @return the eval string with formula and parameter id numbers replaced
 * with names
 * @see jimm.datavision.gui.FormulaWin
 */
public String getEditableEvalString() {
    return Utils.formulaToDisplay(getEvalString());
}

/**
 * Sets the eval string after replacing formula names with formula id
 * numbers. Called from a formula editor.
 * <p>
 * This method will throw an <code>IllegalArgumentException</code> if any
 * formula name or parameter name is not the name of some existing formula
 * or parameter.
 *
 * @param newEvalString the new eval string
 * @see jimm.datavision.gui.FormulaWin#save
 * @throws NullPointerException
 */
public void setEditableEvalString(String newEvalString) {
    setEvalString(Utils.displayToFormula(newEvalString));
}

public String dragString() {
    return "formula:" + getId();
}

public String designLabel() {
    return "{@" + getName() + "}";
}

public String formulaString() {
    return "{@" + getId() + "}";
}

/**
 * Returns <code>true</code> if this formula contains a reference to the
 * specified field.
 *
 * @param f a field
 * @return <code>true</code> if this field contains a reference to the
 * specified field
 * @see FormulaField
 * @see SubtotalField
 */
public boolean refersTo(Field f) {
    String str = getEvalString();
    return str.indexOf(f.formulaString()) != -1;
}

/**
 * Returns <code>true</code> if this formula contains a reference to the
 * specified formula.
 *
 * @param f a formula
 * @return <code>true</code> if this field contains a reference to the
 * specified formula
 * @see FormulaField
 * @see SubtotalField
 */
public boolean refersTo(Formula f) {
    String str = getEvalString();
    return this == f || str.indexOf(f.formulaString()) != -1;
}

/**
 * Returns <code>true</code> if this formula contains a reference to the
 * specified parameter.
 *
 * @param p a parameter
 * @return <code>true</code> if this field contains a reference to the
 * specified parameter
 * @see FormulaField
 * @see SubtotalField
 */
public boolean refersTo(Parameter p) {
    String str = getEvalString();
    return str.indexOf(p.formulaString()) != -1;
}

/**
 * Returns a collection of the columns used in the formula. This is used
 * by the report's query when it is figuring out what columns and tables
 * are used by the report. Called from <code>FormulaField.columnsUsed</code>.
 *
 * @return a possibly empty collection of database columns
 * @see FormulaField#columnsUsed
 * @see Query#findColumnsUsed
 */
public Collection columnsUsed() {
    ArrayList list = new ArrayList();
    int pos = 0;
    String str = getEvalString();
    while ((pos = str.indexOf("{", pos)) != -1) {
	int pos2 = str.indexOf("}", pos + 1);
	char c = str.charAt(pos + 1);
	if (c != '%' && c != '@' && c != '?') {
	    // Ignore special fields, formulas, and parameters
	    Column col = Column.find(str.substring(pos + 1, pos2));
	    if (col != null)	// Should never be null
		list.add(col);
	}
	pos = pos2 + 1;
    }
    return list;
}

/**
 * Evaluate this formula and return the result as either a
 * <code>String</code> or a <code>Double</code>. Before evaluation, special
 * fields, formulas, and column references are subtituted with their
 * values.
 *
 * @return the result of evaluating the formula as either a
 * <code>String</code> or a <code>Double</code>.
 * @see #rpnEval
 */
public Object eval() {
    String str = new String(getEvalString()); // We're gonna munge the string

    // Special values
    int pos, pos2;
    while ((pos = str.indexOf("{%")) != -1) {
	pos2 = str.indexOf("}", pos + 2);
	if (pos2 != -1) {
	    str = str.substring(0, pos)
		+ SpecialField.value(str.substring(pos + 1, pos2), report)
		+ str.substring(pos2 + 1);
	}
    }

    // Formula values
    while ((pos = str.indexOf("{@")) != -1) {
	pos2 = str.indexOf("}", pos + 2);
	if (pos2 != -1) {
	    String formulaIdAsString = str.substring(pos + 2, pos2);
	    str = str.substring(0, pos)
		+ Formula.find(formulaIdAsString).eval()
		+ str.substring(pos2 + 1);
	}
    }

    // Parameter values
    while ((pos = str.indexOf("{?")) != -1) {
	pos2 = str.indexOf("}", pos + 2);
	if (pos2 != -1) {
	    String parameterIdAsString = str.substring(pos + 2, pos2);
	    str = str.substring(0, pos)
		+ Parameter.find(parameterIdAsString).getValue()
		+ str.substring(pos2 + 1);
	}
    }

    // Column values
    while ((pos = str.indexOf("{")) != -1) {
	pos2 = str.indexOf("}", pos + 2);
	if (pos2 != -1) {
	    Column col = Column.find(str.substring(pos + 1, pos2));
	    switch (col.getType()) {
	    case java.sql.Types.CHAR:
	    case java.sql.Types.VARCHAR:
	    case java.sql.Types.DATE:
	    case java.sql.Types.TIME:
	    case java.sql.Types.TIMESTAMP:
		str = str.substring(0, pos) + quoted(report.columnValue(col))
		    + str.substring(pos2 + 1);
		break;
	    default:
		str = str.substring(0, pos) + report.columnValue(col)
		    + str.substring(pos2 + 1);
		break;
	    }
	}
    }

    try {
	return rpnEval(str);
    }
    catch (IOException e) {
	e.printStackTrace();
	// Report error back to user?
	return e.toString();
    }
}

/**
 * Returns the value of the specified string as either a
 * <code>String</code> or a <code>Double</code>, after evaluating it. The
 * string will not contain special, formula, or column values because they
 * have already been replaced in <code>eval</code>.
 * <p>
 * This method is public for testing reasons only. No reason not to make
 * it public, I suppose.
 *
 * @param str the string to evalue
 * @return either a <code>String</code> or a <code>Double</code>
 * @see #eval
 */
public Object rpnEval(String str) throws IOException {
    stack = new Stack();
    BufferedReader r = new BufferedReader(new StringReader(str));
    StreamTokenizer tokenizer = new StreamTokenizer(r);
    tokenizer.wordChars('\u0021', '\uffff');
    tokenizer.quoteChar('\'');
    tokenizer.quoteChar('"');

    int tokType;
    while ((tokType = tokenizer.nextToken()) != StreamTokenizer.TT_EOF) {
	if (tokType == StreamTokenizer.TT_NUMBER)
	    stack.push(new Double(tokenizer.nval));
	else if (tokType == StreamTokenizer.TT_WORD)
	    performOperator(tokenizer.sval);
	else if (tokenizer.ttype == '-') {
	    // Special case because the tokenizer apparently thinks this is
	    // a unary minus sign and reports it as a separate token, not
	    // as a word.
	    performOperator("-");
	}
	else if (tokenizer.ttype == '"')
	    stack.push(quoted(tokenizer.sval));
//  	else			// DEBUG
//  	    System.err.println("error: ttype = " + tokenizer.ttype
//  			       + "; sval = " + tokenizer.sval);
    }

    Object retval = stack.pop();
    if (!stack.empty())
	ErrorHandler.error("Formula.rpnEval: stack not empty after"
			   + " last token; top of stack = ["
			   + stack.pop() + "]");
    return retval;
}

/**
 * Given an operator character, perform the operation on the value(s)
 * at the top of the eval stack.
 *
 * @param op an operator character (for example, '+')
 */
//  protected void performOperator(char op) {
protected void performOperator(String op) {
    Object obj0, obj1;
    if (op.equals("+")) {
	obj1 = stack.pop();
	obj0 = stack.pop();
	if (obj0 instanceof Double && obj1 instanceof Double)
	    stack.push(new Double(((Double)obj0).doubleValue()
				 + ((Double)obj1).doubleValue()));
	else {			// Make "foo" "bar" into "foobar"
	    String s0 = (obj0 instanceof String)
		? ((String)obj0).substring(0, ((String)obj0).length() - 1)
		: obj0.toString();
	    String s1 = (obj1 instanceof String)
		? ((String)obj1).substring(1) : obj1.toString();
	    stack.push(s0 + s1);
	}
    }
    else if (op.equals("-")) {
	obj1 = stack.pop();
	obj0 = stack.pop();
	if (obj0 instanceof Double && obj1 instanceof Double)
	    stack.push(new Double(((Double)obj0).doubleValue()
				 - ((Double)obj1).doubleValue()));
	else
	    throw new UnsupportedOperationException("Operator -" +
						    " only works on numbers");
    }
    else if (op.equals("*")) {
	obj1 = stack.pop();
	obj0 = stack.pop();
	if (obj0 instanceof Double && obj1 instanceof Double)
	    stack.push(new Double(((Double)obj0).doubleValue()
				 * ((Double)obj1).doubleValue()));
	else if (obj1 instanceof Double) { // String * number
	    int times = ((Double)obj1).intValue();
	    String s0 = ((String)obj0).substring(1,
						 ((String)obj0).length() - 1);
	    StringBuffer buf = new StringBuffer();
	    for (int i = 0; i < times; ++i)
		buf.append(s0);
	    stack.push(buf.toString());
	}
	else
	    throw new UnsupportedOperationException("Operator * " +
						    " 2nd arg must be number");
    }
    else if (op.equals("/")) {
	obj1 = stack.pop();
	obj0 = stack.pop();
	if (obj0 instanceof Double && obj1 instanceof Double)
	    stack.push(new Double(((Double)obj0).doubleValue()
				 / ((Double)obj1).doubleValue()));
	else
	    throw new UnsupportedOperationException("Operator /" +
						    " only works on numbers");
    }
    else {
	throw new UnsupportedOperationException("Operator \"" + op
						+ "\" is illegal");
    }
}

/**
 * Returns a string representation of an object enclosed in double quotes.
 * Each double quote within the object's string representation is escaped
 * with a backslash character.
 *
 * @param obj any object
 * @return a double-quoted string representation of the object
 */
protected String quoted(Object obj) {
    String val = obj.toString();
    StringBuffer buf = new StringBuffer("\"");
    int len = val.length();
    for (int i = 0; i < len; ++i) {
	char c = val.charAt(i);
	switch (c) {
	case '"': case '\\':
	    buf.append('\\');
	    buf.append(c);
	    break;
	default:
	    buf.append(c);
	    break;
	}
    }
    buf.append('"');
    return buf.toString();
}

/**
 * Writes this formula as an XML tag.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<formula id=\"" + id + "\" name=\"" + Utils.escapeXML(name)
		+ "\"><![CDATA[" + getEvalString() + "]]></formula>");
}

}
