package jimm.datavision;
import java.util.*;
import jimm.datavision.sql.Column;

/**
 * A group uses a database column to define record grouping. A group may
 * contain multiple header and footer sections.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Group implements Dirtiable, Writeable {

/**
 * Sort records by the group's database column in ascending order (the
 * default).
 */
public static final int SORT_ASCENDING = 0;
/**
 * Sort records by the group's database column in descending order.
 */
public static final int SORT_DESCENDING = 1;

//  protected static HashMap groups = new HashMap();

protected String id;
protected ArrayList headerSections;
protected ArrayList footerSections;
protected Column column;
protected int sortOrder;
protected Object value;		// Not part of dirtiness; current row value
protected boolean newValue;
protected boolean firstValue;
protected boolean dirty;

//  /**
//   * Given a column, returns the group using that column or <code>null</code>
//   * if there is no such group.
//   *
//   * @param col a database column
//   * @return a group, or <code>null</code> if there is no group for this column
//   */
//  public static Group find(Column col) {
//      return (Group)groups.get(col);
//  }


/**
 * Creates a new group and gives it a header section containing a column
 * field and an empty footer section.
 *
 * @param report a report
 * @param col a database column
 * @return a new group with a header section containing a column field and
 * an empty footer section
 */
public static Group create(Report report, Column col) {
    Group group = new Group(col);

    // Create new header section to add to group.
    Section header = new Section(report, "Group Header", false);

    // Create field for selected column. Make it bold.
    Field f = Field.create(null, header, "column", col.getId());
    Format fmt = new Format();
    fmt.setBold(true);
    f.setFormat(fmt);

    // Add field to header section and add section to group.
    header.addField(f);
    group.addHeader(header);

    // Add empty footer section.
    group.addFooter(new Section(report, "Group Header", false));

    return group;
}

/**
 * Constructor.
 *
 * @param col database column
 */
public Group(Column col) {
    headerSections = new ArrayList();
    footerSections = new ArrayList();
    column = col;
    sortOrder = SORT_ASCENDING;
    setDirty(false);
}

/**
 * Returns <code>true</code> if this group or anything it contains has been
 * modified.
 *
 * @return <code>true</code> if this group or anything it contains has been
 * modified
 */
public boolean isDirty() {
    if (dirty) return true;
    for (Iterator iter = headers(); iter.hasNext(); )
	if (((Section)iter.next()).isDirty()) return true;
    for (Iterator iter = footers(); iter.hasNext(); )
	if (((Section)iter.next()).isDirty()) return true;
    return false;
}

/**
 * Sets or clears the dirty bit. If <i>dirty</i> is <code>false</code>,
 * clears the dirty bit for everything contained in this group.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) {
    this.dirty = dirty;
    if (!dirty) {
	for (Iterator iter = headers(); iter.hasNext(); )
	    ((Section)iter.next()).setDirty(false);
	for (Iterator iter = footers(); iter.hasNext(); )
	    ((Section)iter.next()).setDirty(false);
    }
}

//  /**
//   * Returns the id of this group.
//   *
//   * @return the unique id
//   */
//  public String getId() { return id; }

/**
 * Returns the database column used by this group.
 *
 * @return a database column
 */
public Column getColumn() { return column; }

/**
 * Sets the database column used by this group.
 *
 * @param newColumn the new column
 */
public void setColumn(Column newColumn) {
    if (column != newColumn) {
	column = newColumn;
	setDirty(true);
    }
}

/**
 * Returns the sort order (either <code>SORT_ASCENDING</code> or
 * <code>SORT_DESCENDING</code>).
 *
 * @return either <code>SORT_ASCENDING</code> or <code>SORT_DESCENDING</code>
 */
public int getSortOrder() { return sortOrder; }

/**
 * Sets the sort order.
 *
 * @param newSortOrder either <code>SORT_ASCENDING</code> or
 * <code>SORT_DESCENDING</code>
 */
public void setSortOrder(int newSortOrder) {
    if (sortOrder != newSortOrder) {
	sortOrder = newSortOrder;
	setDirty(true);
    }
}

/**
 * Returns the value of this group's database column. Only valid while
 * the report is running.
 *
 * @return the value of the database column; undefined when the report
 * is not running
 */
public Object getValue() { return value; }

/**
 * Sets the group value that is returned by <code>getValue</code>. This
 * method should only be called by the report while it is running, in
 * order to hand the current value of the row to this group.
 */
public void setValue(Object val) {
    if (value == null) {
	value = val;
	firstValue = true;
	newValue = true;
    }
    else if (value.equals(val)) {
	newValue = false;
	firstValue = false;
    }
    else {
	value = val;
	newValue = true;
	firstValue = false;
    }
}

/**
 * Returns <code>true</code> when the value of the database column has
 * changed.
 *
 * @return <code>true</code> when the value of the database column has
 * changed 
 */
public boolean isNewValue() { return newValue; }


/**
 * Returns <code>true</code> when this is the first value ever seen
 * by the group during a report run.
 *
 * @return <code>true</code> if this is the first value ever seen
 */
public boolean isFirstValue() { return firstValue; }

// ---------------- headers

/**
 * Adds a new header to the list of header sections.
 *
 * @param s a report section
 */
public void addHeader(Section s) {
    headerSections.add(s);
    setDirty(true);
}

/**
 * Creates, inserts, and returns a new section after the specified one.
 *
 * @param s a report section
 * @return a new section that has been inserted after the specified one
 */
public Section insertHeaderAfter(Section s) {
    int i = headerSections.indexOf(s);
    if (i == -1)
	return null;

    Report report = ((Section)headerSections.get(0)).report;
    Section newSection = new Section(report, "Group Header", false);
    headerSections.add(i + 1, newSection);
    setDirty(true);
    return newSection;
}

/**
 * Removes a section from this group.
 *
 * @param s a report section
 */
public void removeHeader(Section s) {
    headerSections.remove(s);
    setDirty(true);
}

/**
 * Returns an iterator over all header sections in the group.
 *
 * @return an iterator over the header sections
 */
public Iterator headers() { return headerSections.iterator(); }

// ---------------- footers

/**
 * Adds a new footer to the list of footer sections.
 *
 * @param s a report section
 */
public void addFooter(Section s) {
    footerSections.add(s);
    setDirty(true);
}

/**
 * Creates, inserts, and returns a new section after the specified one.
 *
 * @param s a report section
 * @return a new section that has been inserted after the specified one
 */
public Section insertFooterAfter(Section s) {
    int i = footerSections.indexOf(s);
    if (i == -1)
	return null;

    Report report = ((Section)footerSections.get(0)).report;
    Section newSection = new Section(report, "Group Footer", false);
    footerSections.add(i + 1, newSection);
    setDirty(true);
    return newSection;
}

/**
 * Removes a section from this group.
 *
 * @param s a report section
 */
public void removeFooter(Section s) {
    footerSections.remove(s);
    setDirty(true);
}
/**
 * Returns an iterator over all footer sections in the group.
 *
 * @return an iterator over the footer sections
 */
public Iterator footers() { return footerSections.iterator(); }

/**
 * Returns <code>true</code> if the specified section is inside this group,
 * either as a header or a footer.
 *
 * @param s a section
 * @return <code>true</code> if the section is within this group
 */
public boolean contains(Section s) {
    for (Iterator iter = headers(); iter.hasNext(); )
	if ((Section)iter.next() == s)
	    return true;
    for (Iterator iter = footers(); iter.hasNext(); )
	if ((Section)iter.next() == s)
	    return true;
    return false;
}

/**
 * Called by a report when it starts running, this method prepares the
 * group for use.
 */
public void reset() {
    value = null;
    newValue = firstValue = true;
}

/**
 * Writes this group as an XML tag. Asks each section to write itself
 * as well.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<group column=\"" + column.getId() + "\">");
    out.indent();

    out.println("<headers>");
    out.indent();
    for (Iterator iter = headerSections.iterator(); iter.hasNext(); )
	((Section)iter.next()).writeXML(out);
    out.outdent();
    out.println("</headers>");

    out.println("<footers>");
    out.indent();
    for (Iterator iter = footerSections.iterator(); iter.hasNext(); )
	((Section)iter.next()).writeXML(out);
    out.outdent();
    out.println("</footers>");

    out.outdent();
    out.println("</group>");
}

}
