package jimm.datavision;
import jimm.datavision.sql.Column;

/**
 * A join represents the relationship between two columns in the database.
 * It is used by a query to build the SQL string necessary to retrieve data.
 *
 * @see Query
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Join implements Dirtiable, Writeable, Cloneable {

public static final String[] RELATIONS = {
    "=", "!=", "<", "<=", ">", ">=", "like", "not like", // "null", "not null",
    "in", "not in"
};

protected Column from;
protected String relation;
protected Column to;
protected boolean dirty;

/**
 * Constructor.
 *
 * @param fromCol a database column
 * @param relation a string like "=" or "&lt;" used to join the two database
 * columns
 * @param toCol another database column
 * @see jimm.datavision.sql.Column
 */
public Join(Column fromCol, String relation, Column toCol) {
    from = fromCol;
    this.relation = relation;
    to = toCol;
}

public Object clone() {
    return new Join(from, relation, to);
}

/**
 * Returns <code>true</code> if this join has been modified.
 *
 * @return <code>true</code> if this join has been modified
 */
public boolean isDirty() { return dirty; }

/**
 * Sets or clears the dirty bit.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) { this.dirty = dirty; }

/**
 * Returns the "from" column.
 *
 * @return a column
 */
public Column getFrom() { return from; }

/**
 * Sets the "from" column.
 *
 * @param newColumn the new column
 */
public void setFrom(Column newFrom) {
    if (from != newFrom && (from == null || !from.equals(newFrom))) {
	from = newFrom;
	setDirty(true);
    }
}

/**
 * Returns the relation string (for example, "=" or "&lt;").
 *
 * @return a string used to define the relationship between the two
 * database columns
 */
public String getRelation() { return relation; }

/**
 * Sets the relation string (for example, "=" or "&lt;").
 *
 * @param newRelation the new string
 */
public void setRelation(String newRelation) {
    if (relation != newRelation
	&& (relation == null || !relation.equals(newRelation)))
    {
	relation = newRelation;
	setDirty(true);
    }
}

/**
 * Returns the "to" column.
 *
 * @return a column
 */
public Column getTo() { return to; }

/**
 * Sets the "to" column.
 *
 * @param newColumn the new column
 */
public void setTo(Column newTo) {
    if (to != newTo && (to == null || !to.equals(newTo))) {
	to = newTo;
	setDirty(true);
    }
}

/**
 * Returns a string representation of this join, useable as a where
 * clause in a SQL query.
 */
public String toString() {
    return from.fullName() + " " + relation + " " + to.fullName();
}

/**
 * Writes this join as an XML tag.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<join from=\"" + from.fullName() + "\" relation=\""
		+ relation + "\" to=\"" + to.fullName() + "\"/>");
}

}
