package jimm.datavision;
import java.io.PrintWriter;
import java.util.*;

/**
 * A layout engine is responsible for formatting and outputting report data.
 * <code>LayoutEngine</code> is an abstract class. The Template design
 * pattern is heavily used to provide a framework for concrete layout
 * engine subclasses.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public abstract class LayoutEngine {

/** The number of points per inch. */
public static final int POINTS_PER_INCH = 72;
/** Page width in points. */
public static final double PAGE_WIDTH_POINTS = 72 * 8;
/** Page height in points. */
public static final double PAGE_HEIGHT_POINTS = 72 * 10.25;

protected static final int SECT_REPORT_HEADER = 0;
protected static final int SECT_REPORT_FOOTER = 1;
protected static final int SECT_PAGE_HEADER = 2;
protected static final int SECT_PAGE_FOOTER = 3;
protected static final int SECT_GROUP_HEADER = 4;
protected static final int SECT_GROUP_FOOTER = 5;
protected static final int SECT_DETAIL = 6;

/** Set by report by it's <code>setLayoutEngine</code>. */
protected Report report;
protected int pageNumber;
protected PrintWriter out;
protected boolean newPage;
protected double pageHeightUsed;
protected Section currentSection;
protected int currentSectionType;

/**
 * Constructor.
 */
public LayoutEngine() {
    this(null);
}

/**
 * Constructor.
 *
 * @param out output print writer
 */
public LayoutEngine(PrintWriter out) {
    report = null;
    this.out = out;
}

/**
 * Returns the page width in points.
 *
 * @return the page width in points
 */
public double pageWidth() { return PAGE_WIDTH_POINTS; }

/**
 * Returns the page height in points.
 *
 * @return the page height in points
 */
public double pageHeight() { return PAGE_HEIGHT_POINTS; }

/**
 * Called by the report at the beginning of a report run.
 */
public void start() {
    newPage = true;
    pageHeightUsed = 0;
    pageNumber = 0;
    doStart();
    startPage();
}

/**
 * Called by <code>start</code> as a chance to insert behavior when the
 * report starts.
 */
protected void doStart() {}

/**
 * Called by the report at the end of a report run.
 */
public void end() {
    for (Iterator iter = report.reportFooters.iterator(); iter.hasNext(); )
	outputSection((Section)iter.next(), SECT_REPORT_FOOTER);
    endPage(true);
    doEnd();
    closeOutput();
}

/**
 * Called by <code>end</code> as a chance to insert behavior when the
 * report ends.
 */
protected void doEnd() {}


/**
 * Called by <code>end</code> to let this layout engine clean up a bit.
 */
protected void closeOutput() {
    if (out != null) {
	out.close();
	out = null;
    }
}

/**
 * Called by the report when group headers need to be output.
 *
 * @param isLastRow if <code>true</code>, this is the last row of the report
 */
public void groupHeaders(boolean isLastRow) {
    checkRemainingPageLength(isLastRow);
    for (Iterator iter = report.groups.iterator(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	if (g.isNewValue()) {
	    for (Iterator i2 = g.headerSections.iterator(); i2.hasNext(); )
		outputSection((Section)i2.next(), SECT_GROUP_HEADER);
	}
    }
}

/**
 * Called by the report when a single detail row needs to be output.
 *
 * @param isLastRow if <code>true</code>, this is the last row of the report
 */
public void detail(boolean isLastRow) {
    checkRemainingPageLength(isLastRow);
    for (Iterator iter = report.details.iterator(); iter.hasNext(); )
	outputSection((Section)iter.next(), SECT_DETAIL);
}

/**
 * Called by the report when group headers need to be output.
 *
 * @param isLastRow if <code>true</code>, this is the last row of the report
 */
public void groupFooters(boolean isLastRow) {
    checkRemainingPageLength(isLastRow);
    for (Iterator iter = report.groups.iterator(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	if (g.isNewValue() || isLastRow) {
	    for (Iterator i2 = g.footerSections.iterator(); i2.hasNext(); )
		outputSection((Section)i2.next(), SECT_GROUP_FOOTER);
	}
    }
}

/**
 * Checks remaining page length and outputs a new page if we are at the
 * bottom of the page.
 *
 * @param isLastRow if <code>true</code>, this is the last row of the report
 */
protected void checkRemainingPageLength(boolean isLastRow) {
    // NOTE: need to do this dynamically each row because each one
    // of the multiple detail sections may or may not be active for
    // this row.
    double detailHeight = calcDetailHeight();
    double footerHeight = calcPageFooterHeight();
    if (isLastRow) footerHeight += calcReportFooterHeight();

    if ((pageHeightUsed + footerHeight + detailHeight) > pageHeight())
	endPage(isLastRow);

    if (newPage) startPage();
}

/**
 * Returns the current page number.
 *
 * @return the current page number
 */
public int pageNumber() { return pageNumber; }

/**
 * Starts a new page.
 */
protected void startPage() {
    pageNumber += 1;
    pageHeightUsed = 0;
    newPage = false;

    doStartPage();
    if (pageNumber == 1) {
	for (Iterator iter = report.reportHeaders.iterator(); iter.hasNext(); )
	    outputSection((Section)iter.next(), SECT_REPORT_HEADER);
    }
    for (Iterator iter = report.pageHeaders.iterator(); iter.hasNext(); )
	outputSection((Section)iter.next(), SECT_PAGE_HEADER);
}

/**
 * Called by <code>startPage</code> as a chance to insert behavior when a
 * new page starts.
 */
protected void doStartPage() { }

/**
 * Ends a new page.
 *
 * @param isLastPage if <code>true</code>, this is the last page of the report
 */
protected void endPage(boolean isLastPage) {
    for (Iterator iter = report.pageFooters.iterator(); iter.hasNext(); )
	outputSection((Section)iter.next(), SECT_PAGE_FOOTER);

    newPage = true;
    doEndPage();
}

/**
 * Called by <code>endPage</code> as a chance to insert behavior when a
 * new page ends.
 */
protected void doEndPage() { }

/**
 * Outputs a section.
 *
 * @param sect the section to output
 * @param which the type of section (for example,
 * <code>SECT_PAGE_FOOTER</code>)
 */
protected void outputSection(Section sect, int which) {
    if (sect.isActiveGivenData(report.rset)) {
	currentSectionType = which;
	currentSection = sect;
	doOutputSection(currentSection);
	pageHeightUsed += sect.height;
    }
}

/**
 * Called by <code>outputSection</code> as a chance to insert behavior
 * when a section is output.
 *
 * @param sect a section
 */
protected void doOutputSection(Section sect) {
    // Output the fields in the section
    for (Iterator iter = sect.fields.iterator(); iter.hasNext(); ) {
	Field f = (Field)iter.next();
	if (f.isVisible()) outputField(f);
    }
    // Output the lines
    for (Iterator iter = sect.lines.iterator(); iter.hasNext(); ) {
	Line l = (Line)iter.next();
	if (l.isVisible()) outputLine(l);
    }
}

/**
 * Outputs a field.
 *
 * @param sect the field to output
 */
protected void outputField(Field field) {
    doOutputField(field);
}

/**
 * Called by <code>outputField</code> as a chance to insert behavior
 * when a field is output.
 *
 * @param field a field
 */
protected abstract void doOutputField(Field field);

/**
 * Outputs a line.
 *
 * @param sect the line to output
 */
protected void outputLine(Line line) {
    doOutputLine(line);
}

/**
 * Called by <code>outputLine</code> as a chance to insert behavior
 * when a line is output.
 *
 * @param line a line
 */
protected abstract void doOutputLine(Line line);

/**
 * Returns the total height of all sections in the specified list.
 *
 * @param list a collection of sections
 * @return the total height of all sections in the list
 */
protected double calcSectionHeights(Collection list) {
    double sum = 0;
    for (Iterator iter = list.iterator(); iter.hasNext(); ) {
	Section s = (Section)iter.next();
	if (s.isActiveGivenData(report.rset))
	    sum += s.height;
    }
    return sum;
}

/**
 * Returns the total height of all detail sections.
 *
 * @return the total height of all detail sections
 */
protected double calcDetailHeight() {
    return calcSectionHeights(report.details);
}

/**
 * Returns the total height of all page footer sections.
 *
 * @return the total height of all page footer sections
 */
protected double calcPageFooterHeight() {
    return calcSectionHeights(report.pageFooters);
}

/**
 * Returns the total height of all report footer sections.
 *
 * @return the total height of all report footer sections
 */
protected double calcReportFooterHeight() {
    return calcSectionHeights(report.reportFooters);
}

}
