package jimm.datavision;

/**
 * A line is a visual report element. Lines are used in field borders and
 * independently.
 * <p>
 * Note that currently, line thickness is ignored in the Java GUI (but not
 * in layout engines such as the LaTeXLE).
 *
 * @see Border
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Line extends Element {

protected double thickness;
protected Point[] points;

/**
 * Constructor.
 *
 * @param section the section containing this line
 */
public Line(Section section) { this(section, 1.0, null, null); }

/**
 * Constructor.
 *
 * @param section the section containing this line
 * @param thickness the line thickness
 */
public Line(Section section, double thickness) {
    this(section, thickness, null, null);
}

/**
 * Constructor.
 *
 * @param section the section containing this line
 * @param thickness the line thickness
 * @param p0 one end point of the line
 * @param p0 the other end point of the line
 */
public Line(Section section, double thickness, Point p0, Point p1) {
    super(section);
    this.thickness = thickness;
    points = new Point[2];
    points[0] = p0;
    points[1] = p1;
    setDirty(false);
}

/**
 * Adds an end point to the line. Used when constructing a line from XML,
 * where we don't see the point until after creating this line. Does not
 * set the dirty bit.
 *
 * @param x the x coordinate
 * @param y the y coordinate
 */
public void addEndPoint(double x, double y) {
    Point p = new Point(x, y);
    if (points[0] == null)
	points[0] = p;
    else
	points[1] = p;
    p.addObserver(this);
}

/**
 * Returns <code>true</code> if this line has been modified.
 *
 * @return <code>true</code> if this line has been modified
 */
public boolean isDirty() {
    if (super.isDirty()) return true;
    return points[0].isDirty() ||points[1].isDirty();
}

/**
 * Sets or clears the dirty bit.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) {
    this.dirty = dirty;
    if (!dirty) {
	points[0].setDirty(false);
	points[1].setDirty(false);
    }
}

/**
 * Returns the line thickness.
 *
 * @return the line thickness
 */
public double getThickness() { return thickness; }

/**
 * Sets the line thickness.
 *
 * @param newThickness the new line thickness
 */
public void setThickness(double newThickness) {
    if (thickness != newThickness) {
	thickness = newThickness;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns one of the two end points of the line.
 *
 * @param index either 0 or 1
 * @return a point
 */
public Point getPoint(int index) { return points[index]; }

/**
 * Sets one of the two end points.
 *
 * @param newPoint a point
 * @param index either 0 or 1
 */
public void setPoint(Point newPoint, int index) {
    if (points[index] != newPoint) {
	if (points[index] != null) points[index].deleteObserver(this);
	points[index] = newPoint;
	if (points[index] != null) points[index].addObserver(this);
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the length of the line.
 *
 * @return the distance between the two end points
 */
public double length() { return points[0].distanceTo(points[1]); }

/**
 * Returns a string representation of this line.
 *
 * @return a string representing this line
 */
public String toString() { return "(" + points[0] + ", " + points[0] + ")"; }

/**
 * Writes this line as an XML tag.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<line thickness=\"" + thickness + "\">");
    out.indent();
    points[0].writeXML(out);
    points[1].writeXML(out);
    out.outdent();
    out.println("</line>");
}

}
