package jimm.datavision;
import jimm.datavision.sql.Table;
import jimm.datavision.sql.Column;
import java.util.*;
import java.sql.PreparedStatement;
import java.sql.SQLException;

/**
 * Queries build SQL query strings. They contain tables, joins, and
 * where clauses.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Query implements Dirtiable, Writeable {

protected Report report;
protected ArrayList joins;
protected String whereClause;
protected HashMap sorts;
protected Set tables;
protected ArrayList columns;
protected boolean dirty;
protected ArrayList preparedStmtValues;

/**
 * Constructor.
 *
 * @param report the report for which this query will generate SQL
 */
public Query(Report report) {
    this.report = report;
    joins = new ArrayList();
    whereClause = null;
    sorts = new HashMap();
    tables = new HashSet();
    columns = new ArrayList();
}

/**
 * Returns <code>true</code> if this query or any of its joins have
 * been modified.
 *
 * @return <code>true</code> if this query has been modified
 */
public boolean isDirty() {
    if (dirty) return true;
    for (Iterator iter = joins.iterator(); iter.hasNext(); )
	if (((Join)iter.next()).isDirty()) return true;
    return false;
}

/**
 * Sets or clears the dirty bit.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) {
    this.dirty = dirty;
    if (!dirty) {
	for (Iterator iter = joins.iterator(); iter.hasNext(); )
	    ((Join)iter.next()).setDirty(false);
    }
}
/**
 * Returns <code>true</code> if the specified parameter exists within this
 * query's where clause.
 *
 * @param p a parameter
 * @return <code>true</code> if the specified parameter exists within
 * the where clause
 */
public boolean containsReferenceTo(Parameter p) {
    return whereClause != null && whereClause.indexOf(p.formulaString()) != -1;
}

/**
 * Adds a join to the list of joins used by this query.
 *
 * @param join a join
 */
public void addJoin(Join join) {
    joins.add(join);
    setDirty(true);
}

/**
 * Adds all joins in the collection to our list.
 *
 * @param coll a collection of joins
 */
public void addAllJoins(Collection coll) {
    joins.addAll(coll);
}

/**
 * Removes a join from the list of joins used by this query.
 *
 * @param join a join
 */
public void removeJoin(Join join) {
    joins.remove(join);
    setDirty(true);
}

/**
 * Removes all joins in the collection from our list.
 */
public void clearJoins() {
    joins.clear();
}

/**
 * Returns an iterator over all the joins used by this query.
 *
 * @return an iterator over all the joins
 */
public Iterator joins() { return joins.iterator(); }

/**
 * Returns the where clause fit for human consumption. This mainly means
 * that we substitute formula and parameter numbers with names. Called from
 * a where clause editor. This code assumes that curly braces are never
 * nested.
 *
 * @return the eval string with formula and parameter id numbers replaced
 * with names
 * @see jimm.datavision.gui.WhereClauseWin
 */
public String getEditableWhereClause() {
    return Utils.formulaToDisplay(getWhereClause());
}

/**
 * Returns the raw where clause string; may be <code>null</code>.
 *
 * @return the where clause string; may be <code>null</code>
 * @see #getWhereClauseForPreparedStatement
 */
public String getWhereClause() {
    return whereClause;
}

/**
 * Returns the where clause string; may be <code>null</code>. If there
 * are any parameter values, we return '?' in their place and save the
 * values for later use.
 *
 * @return the where clause string; may be <code>null</code>
 * @see #getWhereClause
 */
protected String getWhereClauseForPreparedStatement() {
   if (whereClause == null)
	return null;
    if (whereClause.indexOf("{?") == -1)
	return whereClause;

    StringBuffer buf = new StringBuffer();
    int len = whereClause.length();
    for (int i = 0; i < len; ++i) {
	char c = whereClause.charAt(i);
	if (c == '{' && (i + 1) < len) {
	    int nameStart, nameEnd;
	    switch (whereClause.charAt(i + 1)) {
	    case '@':		// Formula
		nameStart = i + 2;
		nameEnd = whereClause.indexOf("}", nameStart);
		if (nameEnd != -1) {
		    String formulaIdAsString =
			whereClause.substring(nameStart, nameEnd);
		    preparedStmtValues.add(Formula.find(formulaIdAsString).eval());
		    buf.append('?'); // For prepared statement
		    i = nameEnd;
		}
		break;
	    case '?':		// Parameter
		nameStart = i + 2;
		nameEnd = whereClause.indexOf("}", nameStart);
		if (nameEnd != -1) {
		    String idAsString =
			whereClause.substring(nameStart, nameEnd);
		    // Ask report for parameter value so the report can
		    // ask the user to fill in the parameter's value.
		    preparedStmtValues.add(report.getParameter(idAsString));
		    buf.append('?'); // For prepared statement
		    i = nameEnd;
		}
		break;
	    default:
		buf.append(c);
		break;
	    }
	}
	else
	    buf.append(c);
    }
    return buf.toString();
}

/**
 * Sets the where clause (may be <code>null</code>). The string passed
 * to us contains parameter and formula display strings, not their
 * "real" <code>formulaString</code> representations. Translate the latter
 * into the former before saving this string.
 *
 * @param newWhereClause a where clause string; may be <code>null</code>
 */
public void setEditableWhereClause(String newWhereClause) {
    setWhereClause(Utils.displayToFormula(newWhereClause));
}

/**
 * Sets the where clause (may be <code>null</code>).
 *
 * @param newWhereClause a where clause string; may be <code>null</code>
 */
public void setWhereClause(String newWhereClause) {
    if (whereClause != newWhereClause
	&& (whereClause == null || !whereClause.equals(newWhereClause)))
    {
	whereClause = newWhereClause;
	setDirty(true);
    }
}

/**
 * Adds a column with the default sort order (descending).
 *
 * @param col a database column
 */
public void addSort(Column col) {
    addSort(col, "desc");
}

/**
 * Adds a sort order for the specified column. The first character
 * of the <i>order</i> string is inspected. If it is a 'd' or 'D',
 * then it's taken to mean "descending". Anything else will result
 * in an "ascending" sort order.
 *
 * @param col a database column
 * @param order a string
 */
public void addSort(Column col, String order) {
    sorts.put(col, order.toLowerCase().charAt(0) == 'd' ? "desc" : "asc");
    setDirty(true);
}

/**
 * Removes a sorting from the list.
 *
 * @param clause a where clause
 */
public void removeSort(Column col) {
    if (sorts.containsKey(col)) {
	sorts.remove(col);
	setDirty(true);
    }
}

/**
 * Removes all sorts from our list.
 */
public void clearSorts() {
    sorts = new HashMap();
}

/**
 * Returns an iterator over all the sorted columns used by this query.
 *
 * @return an iterator
 */
public Iterator sortedColumns() { return sorts.keySet().iterator(); }

/**
 * Returns the sort order string ("asc" or "desc") of the specified column.
 * If the column is not sorted, returns <code>null</code>
 *
 * @param col a database column
 * @return the string "asc" or "desc"; <code>null</code> if the column is
 * not on the list of sorted columns.
 */
public String sortOrderOf(Column col) { return (String)sorts.get(col); }

/**
 * Returns the index of the specified column in the select statement.
 *
 * @param column a database column
 */
public int indexOfColumn(Column column) { return columns.indexOf(column); }

/**
 * Builds collections of the report tables and column fields actually used
 * in the report.
 */
public void findColumnsUsed() {
    columns.clear();
    tables.clear();
    report.withFieldsDo(new FieldWalker() {
	public void step(Field f) {
	    if (f instanceof ColumnField) {
		Column col = ((ColumnField)f).getColumn();
		if (!columns.contains(col)) columns.add(col);
		tables.add(col.getTable());
	    }
	    else if (f instanceof FormulaField) {
		FormulaField ff = (FormulaField)f;
		for (Iterator iter = ff.columnsUsed().iterator();
		     iter.hasNext(); )
		{
		    Column col = (Column)iter.next();
		    if (!columns.contains(col)) columns.add(col);
		    tables.add(col.getTable());
		}
	    }
	}
	});

    // Add all columns and tables used in joins.
    for (Iterator iter = joins.iterator(); iter.hasNext(); ) {
	Join join = (Join)iter.next();
	if (!columns.contains(join.from)) columns.add(join.from);
	if (!columns.contains(join.to)) columns.add(join.to);
	tables.add(join.from.getTable());
	tables.add(join.to.getTable());
    }

    // Add all columns and tables used in where clauses. This is difficult
    // since the where clause is a string. We assume that anything of the
    // form "foo.bar" is a table.column name.
// FIX

    // Add all columns and tables used in sorts.
    for (Iterator iter = sorts.keySet().iterator(); iter.hasNext(); ) {
	Column col = (Column)iter.next();
	if (!columns.contains(col)) columns.add(col);
	tables.add(col.getTable());
    }
}

/**
 * Returns a collection containing the tables used by this query.
 *
 * @return the collection of tables used by this query
 */
public Collection getTablesUsed() {
    findColumnsUsed();
    return tables;
}

/**
 * Returns the query as a human-readable SQL statement, including parameter
 * and formula display strings.
 *
 * @return a SQL query string
 */
public String toString() {
    return queryAsString(true);
}

/**
 * Returns the query as a SQL string suitable for building a prepared
 * statement.
 *
 * @return a SQL query string
 */
public String toPreparedStatementString() {
    preparedStmtValues = new ArrayList();
    return queryAsString(false);
}

/**
 * Returns the query as either a human-readable SQL statement or a SQL
 * string suitable for building a prepared statement.
 *
 * @param forDisplay if <code>true</code> return a human-readable string,
 * else return a SQL string suitable for building a prepared statement
 * @return a SQL string
 */
protected String queryAsString(boolean forDisplay) {
    // Rebuild collections of tables and columns every time (not just first
    // time) since the list of columns we want to use may have changed
    // since last time.
    findColumnsUsed();

    if (tables.size() == 0 || columns.size() == 0)
	return "";

    StringBuffer str = new StringBuffer("select ");
    str.append(Utils.join(columns, ", "));

    str.append(" from ");
    boolean first = true;
    for (Iterator iter = tables.iterator(); iter.hasNext(); ) {
	if (first) first = false;
	else str.append(", ");
	str.append(((Table)iter.next()).getName());
    }

    if (!joins.isEmpty() || (whereClause != null && whereClause.length() > 0))
    {
	str.append(" where ");
	if (!joins.isEmpty()) {
	    str.append("(");
	    str.append(Utils.join(joins, ") and ("));
	    str.append(")");
	}
	if (whereClause != null && whereClause.length() > 0) {
	    if (!joins.isEmpty())
		str.append(" and ");

	    str.append("(");
	    if (forDisplay)
		str.append(getEditableWhereClause());
	    else {
		// Call getWhereClauseForPreparedStatement so parameter
		// values are substituted and saved.
		str.append(getWhereClauseForPreparedStatement());
	    }
	    str.append(")");
	}
    }

    if (!report.groups.isEmpty() || !sorts.isEmpty()) {
	str.append(" order by ");
	ArrayList orders = new ArrayList();
	for (Iterator iter = report.groups.iterator(); iter.hasNext(); ) {
	    Group g = (Group)iter.next();
	    StringBuffer buf = new StringBuffer(g.column.fullName());
	    switch (g.sortOrder) {
	    case Group.SORT_ASCENDING: buf.append(" asc"); break;
	    case Group.SORT_DESCENDING: buf.append(" desc"); break;
	    }
	    orders.add(buf.toString());
	}
	for (Iterator iter = sortedColumns(); iter.hasNext(); ) {
	    Column col = (Column)iter.next();
	    orders.add(col.fullName() + " " + sorts.get(col));
	}

	str.append(Utils.join(orders, ", "));
    }

    return str.toString();
}

/**
 * Given a prepared statement created with the text returned by
 * <code>toPreparedStatementString</code>, plug in all the parameter
 * and formula values.
 *
 * @see #toPreparedStatementString
 */
public void setParameters(PreparedStatement stmt) throws SQLException {
    int i = 1;
    for (Iterator iter = preparedStmtValues.iterator(); iter.hasNext(); ++i)
	stmt.setObject(i, iter.next());
}

/**
 * Writes this query as an XML tag. Writes all joins, where clauses,
 * and sorts as well.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<query>");
    out.indent();

    for (Iterator iter = joins(); iter.hasNext(); )
	((Join)iter.next()).writeXML(out);

    if (whereClause != null && whereClause.length() > 0)
	out.println("<where><![CDATA[" + whereClause + "]]></where>");

    for (Iterator iter = sortedColumns(); iter.hasNext(); ) {
	Column col = (Column)iter.next();
	out.println("<sort order=\"" + sorts.get(col) + "\" column=\""
		    + col.getId() + "\" />");
    }

    out.outdent();
    out.println("</query>");
}

}
