package jimm.datavision;
import jimm.datavision.sql.*;
import jimm.datavision.gui.DbPasswordDialog;
import jimm.datavision.gui.DesignWin;
import jimm.datavision.gui.StatusDialog;
import jimm.datavision.gui.parameter.ParamAskWin;
import java.io.*;
import java.util.*;
import java.sql.*;
import javax.swing.JOptionPane;
import org.xml.sax.SAXException;

/**
 * A report holds database information, accepts parameters from the user,
 * runs a query, and uses a layout engine to format the output. It may
 * contain many different sections, each of which can contain logic for
 * surpressing itself.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Report implements Dirtiable, Writeable {

protected static final double DEFAULT_DTD_VERSION = 0.2;

protected String name;
protected String title;
protected String author;
protected String description;
protected Database database;
protected ArrayList formulas;
protected ArrayList parameters;
protected ArrayList reportHeaders;
protected ArrayList reportFooters;
protected ArrayList pageHeaders;
protected ArrayList pageFooters;
protected ArrayList groups;
protected ArrayList details;
protected ResultSet rset;
protected Query query;
protected LayoutEngine layoutEngine;
protected Collection subtotalFields;
protected boolean dirty;
protected String databasePassword;
protected double dtdVersion;
protected boolean askedForParameters;
protected boolean parametersHaveValues;
protected String paramXMLFileName;

/**
 * Constructs an empty report.
 */
public Report() {
    init();
    initEmptyReport();
    setDirty(false);
}

/**
 * Constructs a report by reading the specified XML file. If the user was
 * prompted for a database password and that dialog box was cancelled,
 * we set our database object to <code>null</code>. This is checked in the
 * <code>jimm.datavision.gui.DesignWin</code> constructor.
 *
 * @param xmlFileName file path to report XML file
 * @see jimm.datavision.gui.DesignWin#DesignWin
 * @see jimm.datavision.gui.DbPasswordDialog
 */
public Report(String xmlFileName) {
    init();
    if (xmlFileName != null) {
	try {
	    readFile(xmlFileName);
	}
	catch (UserCancellationException uce) {
	    database = null;
	}
	catch (Exception e) {
	    ErrorHandler.error(e);
	    initEmptyReport();
	}
    }
    else
	initEmptyReport();
    setDirty(false);
}

/**
 * Initializes the report. Called from constructors.
 */
private void init() {
    query = new Query(this);
    formulas = new ArrayList();
    parameters = new ArrayList();
    reportHeaders = new ArrayList();
    reportFooters = new ArrayList();
    pageHeaders = new ArrayList();
    pageFooters = new ArrayList();
    groups = new ArrayList();
    details = new ArrayList();
    name = "Unnamed";
    title = "Untitled";
    askedForParameters = false;
    parametersHaveValues = false;
    dtdVersion = DEFAULT_DTD_VERSION;
}

/**
 * Initialize an empty report. Called from constructors.
 */
protected void initEmptyReport() {
    reportHeaders.add(new Section(this, "Report Header", false));
    pageHeaders.add(new Section(this, "Page Header", false));
    pageFooters.add(new Section(this, "Page Footer", false));
    reportFooters.add(new Section(this, "Report Footer", false));
    details.add(new Section(this, "Detail", true));
}

/**
 * Sets the layout engine to use.
 *
 * @param layoutEngine a layout engine
 */
public void setLayoutEngine(LayoutEngine layoutEngine) {
    this.layoutEngine = layoutEngine;
    this.layoutEngine.report = this;
}

/**
 * Returns <code>true</code> if this report or anything inside it has been
 * modified.
 *
 * @return <code>true</code> if this report or anything inside it has been
 * modified
 */
public boolean isDirty() {
    if (dirty) return true;
    if (database != null && database.isDirty()) return true;
    if (query != null && query.isDirty()) return true;
    Iterator iter;
    for (iter = parameters.iterator(); iter.hasNext(); )
	if (((Parameter)iter.next()).isDirty()) return true;
    for (iter = formulas.iterator(); iter.hasNext(); )
	if (((Formula)iter.next()).isDirty()) return true;
    for (iter = groups.iterator(); iter.hasNext(); )
	if (((Group)iter.next()).isDirty()) return true;

    // Can't modify a final, but *can* modify the contents of a final.
    final boolean[] someSectionIsDirty = new boolean[1];
    someSectionIsDirty[0] = false;
    withSectionsDo(new SectionWalker() {
	public void step(Section s) {
	    if (s.isDirty()) someSectionIsDirty[0] = true;
	}
	});
    return someSectionIsDirty[0];
}

/**
 * Sets or clears the dirty bit. If <i>dirty</i> is <code>false</code>,
 * clears the dirty bit for everything inside the report as well.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) {
    this.dirty = dirty;
    if (!dirty) {
	if (database != null) database.setDirty(false);
	if (query != null) query.setDirty(false);
	Iterator iter;
	for (iter = parameters.iterator(); iter.hasNext(); )
	    ((Parameter)iter.next()).setDirty(false);
	for (iter = formulas.iterator(); iter.hasNext(); )
	    ((Formula)iter.next()).setDirty(false);
	for (iter = groups.iterator(); iter.hasNext(); )
	    ((Group)iter.next()).setDirty(false);
	withSectionsDo(new SectionWalker() {
	    public void step(Section s) { s.setDirty(false); }
	    });
    }
}

public String getName() { return name; }
public void setName(String newName) {
    if (name != newName && (name == null || !name.equals(newName))) {
	name = newName;
	setDirty(true);
    }
}

public String getTitle() { return title; }
public void setTitle(String newTitle) {
    if (title != newTitle && (title == null || !title.equals(newTitle))) {
	title = newTitle;
	setDirty(true);
    }
}

public String getAuthor() { return author; }
public void setAuthor(String newAuthor) {
    if (author != newAuthor && (author == null || !author.equals(newAuthor))) {
	author = newAuthor;
	setDirty(true);
    }
}

public String getDescription() { return description; }
public void setDescription(String newDescription) {
    if (description != newDescription
	&& (description == null || !description.equals(newDescription)))
    {
	description = newDescription;
	setDirty(true);
    }
}

public Database getDatabase() { return database; }
public void setDatabase(Database newDatabase) {
    if (database != newDatabase) {
	database = newDatabase;
	setDirty(true);
    }
}

public Query getQuery() { return query; }
public void setQuery(Query newQuery) {
    if (query != newQuery) {
	query = newQuery;
	setDirty(true);
    }
}

// ---------------- parameters

public void addParameter(Parameter p) {
    parameters.add(p);
    setDirty(true);
}
public void removeParameter(Parameter p) {
    parameters.remove(p);
    setDirty(true);
}
public Iterator parameters() { return parameters.iterator(); }

// ---------------- formulas

public void addFormula(Formula f) {
    formulas.add(f);
    setDirty(true);
}
public void removeFormula(Formula f) {
    formulas.remove(f);
    setDirty(true);
}
public Iterator formulas() { return formulas.iterator(); }

// ---------------- headers

public void addHeader(Section s) {
    reportHeaders.add(s);
    setDirty(true);
}
public Section insertHeaderAfter(Section s) {
    int i = reportHeaders.indexOf(s);
    Section newSection = new Section(this, "Report Header", false);
    reportHeaders.add(i + 1, newSection);
    setDirty(true);
    return newSection;
}
public void removeHeader(Section s) {
    reportHeaders.remove(s);
    setDirty(true);
}
public Iterator headers() { return reportHeaders.iterator(); }

// ---------------- footers

public void addFooter(Section s) {
    reportFooters.add(s);
    setDirty(true);
}
public Section insertFooterAfter(Section s) {
    int i = reportFooters.indexOf(s);
    Section newSection = new Section(this, "Report Footer", false);
    reportFooters.add(i + 1, newSection);
    setDirty(true);
    return newSection;
}
public void removeFooter(Section s) {
    reportFooters.remove(s);
    setDirty(true);
}
public Iterator footers() { return reportFooters.iterator(); }

// ---------------- page headers

public void addPageHeader(Section s) {
    pageHeaders.add(s);
    setDirty(true);
}
public Section insertPageHeaderAfter(Section s) {
    int i = pageHeaders.indexOf(s);
    Section newSection = new Section(this, "Page Header", false);
    pageHeaders.add(i + 1, newSection);
    setDirty(true);
    return newSection;
}
public void removePageHeader(Section s) {
    pageHeaders.remove(s);
    setDirty(true);
}
public Iterator pageHeaders() { return pageHeaders.iterator(); }

// ---------------- page footers

public void addPageFooter(Section s) {
    pageFooters.add(s);
    setDirty(true);
}
public Section insertPageFooterAfter(Section s) {
    int i = pageFooters.indexOf(s);
    Section newSection = new Section(this, "Page Footer", false);
    pageFooters.add(i + 1, newSection);
    setDirty(true);
    return newSection;
}
public void removePageFooter(Section s) {
    pageFooters.remove(s);
    setDirty(true);
}
public Iterator pageFooters() { return pageFooters.iterator(); }

// ---------------- groups

public void addGroup(Group g) {
    groups.add(g);
    query.removeSort(g.getColumn());
    setDirty(true);
}
public void removeGroup(Group g) { groups.remove(g); setDirty(true); }
public void removeAllGroups() {
    if (groups.size() > 0) {
	groups.clear();
	setDirty(true);
    }
}
public Iterator groups() { return groups.iterator(); }

/**
 * Returns an iterator over the groups in reverse order. Useful when
 * displaying group footers.
 *
 * @return an iterator over the groups, in reverse order
 */
public Iterator groupsReversed() {
    ArrayList reversed = (ArrayList)groups.clone();
    Collections.reverse(reversed);
    return reversed.iterator();
}

// ---------------- details

public void addDetail(Section s) { details.add(s); setDirty(true); }
public Section insertDetailAfter(Section s) {
    int i = details.indexOf(s);
    Section newSection = new Section(this, "Detail", true);
    reportHeaders.add(i + 1, newSection);
    setDirty(true);
    return newSection;
}
public void removeDetail(Section s) { details.remove(s); setDirty(true); }
public Iterator details() { return details.iterator(); }

public void removeField(Field f) {
    Section s = sectionContaining(f);
    if (s != null)
	s.removeField(f);
}

public double getDTDVersion() { return dtdVersion; }
public void setDTDVersion(double v) { dtdVersion = v; }

/**
 * Returns <code>true</code> if the specified field exists within this
 * report.
 *
 * @param f a field
 * @return <code>true</code> if the specified field exists within this
 * report
 */
public boolean contains(Field f) {
    return sectionContaining(f) != null;
}

/**
 * Returns the section containing the specified field, or <code>null</code>
 * if the field is not in the report.
 *
 * @param f a field
 * @return the section containing the specified field, or <code>null</code>
 * if the field is not in the report
 */
public Section sectionContaining(Field f) {
    Section s;
    Iterator iter, iter2;
    for (iter = reportHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.contains(f)) return s;
    }
    for (iter = pageHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.contains(f)) return s;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.contains(f)) return s;
	}
    }
    for (iter = details.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.contains(f)) return s;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.contains(f)) return s;
	}
    }
    for (iter = reportFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.contains(f)) return s;
    }
    for (iter = pageFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.contains(f)) return s;
    }

    return null;
}

/**
 * Returns <code>true</code> if the specified field exists within this
 * report either directly (as a field) or indirectly (as a formula used
 * by a subtotal or by another formula).
 *
 * @param f a field
 * @return <code>true</code> if the specified field exists within this
 * report
 */
public boolean containsReferenceTo(Field f) {
    Section s;
    Iterator iter, iter2;
    for (iter = reportHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = pageHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(f)) return true;
	}
    }
    for (iter = details.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(f)) return true;
	}
    }
    for (iter = reportFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = pageFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }

    return false;
}

/**
 * Returns <code>true</code> if the specified formula exists within this
 * report either directly (as a formula field) or indirectly (as a formula
 * used by a subtotal or by another formula). This is not the same as
 * answering the question, "Does this report contain this formula?" because
 * we want to know if the visual portion of the report or some other formula
 * accesses the specified formula.
 *
 * @param f a formula
 * @return <code>true</code> if the specified parameter exists within some
 * visual element of the report or within some formula
 */
public boolean containsReferenceTo(Formula f) {
    Section s;
    Iterator iter, iter2;
    for (iter = reportHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = pageHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(f)) return true;
	}
    }
    for (iter = details.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(f)) return true;
	}
    }
    for (iter = reportFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = pageFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }

    return false;
}

/**
 * Returns <code>true</code> if the specified parameter exists within this
 * report either directly (as a parameter field) or indirectly (as a
 * parameter used by a subtotal or by another parameter or in the query's
 * where clause). This is not the same as answering the question, "Does
 * this report contain this parameter?" because we want to know if the
 * visual portion of the report or some formula accesses the specified
 * parameter.
 *
 * @param p a parameter
 * @return <code>true</code> if the specified parameter exists within some
 * visual element of the report or within some formula or within the
 * query's where clause
 */
public boolean containsReferenceTo(Parameter p) {
    if (query.containsReferenceTo(p))
	return true;

    Section s;
    Iterator iter, iter2;

    for (iter = reportHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(p)) return true;
    }
    for (iter = pageHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(p)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(p)) return true;
	}
    }
    for (iter = details.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(p)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(p)) return true;
	}
    }
    for (iter = reportFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(p)) return true;
    }
    for (iter = pageFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(p)) return true;
    }

    return false;
}

/**
 * Returns a list of all the parameters actually used in the
 * report. Calls <code>containsReferenceTo</code>.
 *
 * @return a list of parameters
 * @see #containsReferenceTo
 */
protected List collectUsedParameters() {
    ArrayList list = new ArrayList();
    for (Iterator iter = parameters.iterator(); iter.hasNext(); ) {
	Parameter p = (Parameter)iter.next();
	if (containsReferenceTo(p))
	    list.add(p);
    }
    return list;
}

/**
 * Returns the group associated with the specified column, or
 * <code>null</code> if there isn't one.
 *
 * @param col a database column
 * @return the group associated with this column, or <code>null</code>
 * if there isn't one
 */
public Group findGroup(Column col) {
    for (Iterator iter = groups(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	if (g.getColumn() == col)
	    return g;
    }
    return null;
}

/**
 * Returns the group associated with the specified section, or
 * <code>null</code> if there isn't one.
 *
 * @param section a section
 * @return the group associated with this section, or <code>null</code>
 * if there isn't one
 */
public Group findGroup(Section section) {
    for (Iterator iter = groups(); iter.hasNext(); ) {
	Group group = (Group)iter.next();
	if (group.contains(section))
	    return group;
    }
    return null;
}

/**
 * Returns <code>true</code> if the specified section is contained in
 * any group. The section may be either a header or a footer section.
 *
 * @param section a section
 * @return <code>true</code> if the section is within some group
 */
public boolean isInsideGroup(Section section) {
    return findGroup(section) != null;
}

/**
 * Returns <code>true</code> if the specified database column is a
 * group column.
 *
 * @return <code>true</code> if the specified database column is a
 * group column
 */
public boolean isGroupColumn(Column col) {
    return findGroup(col) != null;
}

/**
 * Creates and returns a new section below the specified one.
 *
 * @param s a section
 * @return a new section
 */
public Section insertSectionBelow(Section s) {
    if (reportHeaders.contains(s))
	return insertHeaderAfter(s);
    if (reportFooters.contains(s))
	return insertFooterAfter(s);
    if (pageHeaders.contains(s))
	return insertPageHeaderAfter(s);
    if (pageFooters.contains(s))
	return insertPageFooterAfter(s);
    if (details.contains(s))
	return insertDetailAfter(s);

    for (Iterator iter = groups(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	if (g.headerSections.contains(s))
	    return g.insertHeaderAfter(s);
	if (g.footerSections.contains(s))
	    return g.insertFooterAfter(s);
    }

    return null;		// Should not happen
}

/**
 * Removes the specified section.
 *
 * @param s a section
 */
public void removeSection(Section s) {
    if (reportHeaders.contains(s))
	removeHeader(s);
    else if (reportFooters.contains(s))
	removeFooter(s);
    else if (pageHeaders.contains(s))
	removePageHeader(s);
    else if (pageFooters.contains(s))
	removePageFooter(s);
    else if (details.contains(s))
	removeDetail(s);
    else {
	for (Iterator iter = groups(); iter.hasNext(); ) {
	    Group g = (Group)iter.next();
	    if (g.headerSections.contains(s)) {
		g.removeHeader(s);
		return;
	    }
	    if (g.footerSections.contains(s)) {
		g.removeFooter(s);
		return;
	    }
	}
    }
}

/**
 * Returns <code>true</code> if this report contains some field anywhere.
 * Useful when the GUI is enabling menu items, for example.
 *
 * @return <code>true</code> if this report contains some field anywhere
 */
public boolean hasFields() {
    return countTablesUsed() > 0;
}

/**
 * Returns the number of tables used by the report. Useful when the GUI
 * is enabling menu items, for example.
 *
 * @return the number of tables used by this report (by the query)
 */
public int countTablesUsed() {
    return query.getTablesUsed().size();
}

/**
 * Returns <code>true</code> if the specified section is the only section
 * of its kind; that is, the only section in the collection in which it
 * is contained.
 *
 * @param s a report section
 * @return <code>true</code> if this section is a loner
 */
public boolean isOneOfAKind(Section s) {
    if (reportHeaders.contains(s) && reportHeaders.size() == 1)
	return true;
    if (reportFooters.contains(s) && reportFooters.size() == 1)
	return true;
    if (pageHeaders.contains(s) && pageHeaders.size() == 1)
	return true;
    if (pageFooters.contains(s) && pageFooters.size() == 1)
	return true;
    if (details.contains(s) && details.size() == 1)
	return true;

    for (Iterator iter = groups(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	if (g.headerSections.contains(s) && g.headerSections.size() == 1)
	    return true;
	if (g.footerSections.contains(s) && g.footerSections.size() == 1)
	    return true;
    }

    return false;
}

/**
 * Sets the database password.
 */
public void setDatabasePassword(String pwd) { databasePassword = pwd; }

/**
 * Sets a database's user name and password. If we already have the
 * password, give it to the database. If we don't, ask the user for
 * both the user name (supplied to us) and the password and give them
 * to the database.
 *
 * @param db the database
 */
public void askForPassword(Database db) {
    if (databasePassword != null) {
	db.setPassword(databasePassword);
	return;
    }

    if (ErrorHandler.usingGUI()) {
	DbPasswordDialog dialog =
	    new DbPasswordDialog(DesignWin.findWindowFor(this), db.getName(),
				 db.getUserName());
	// This dialog is modal, so when we return the values have been
	// filled.
	db.setUserName(dialog.getUserName());
	db.setPassword(dialog.getPassword());
    }
    else {
	System.out.print("User name [" + db.getUserName() + "]: ");
	System.out.flush();
	try {
	    BufferedReader in =
		new BufferedReader(new InputStreamReader(System.in));
	    String username = in.readLine();
	    if (username.length() > 0)
		db.setUserName(username);
	}
	catch (IOException e) {
	    ErrorHandler.error("Error reading database user name");
	}

	System.out.print("Password for user " + db.getUserName() + ": ");
	System.out.flush();
	try {
	    BufferedReader in =
		new BufferedReader(new InputStreamReader(System.in));
	    databasePassword = in.readLine();
	    db.setPassword(databasePassword);
	}
	catch (IOException e) {
	    ErrorHandler.error("Error reading database password");
	}
    }
}

/**
 * Returns the value of the specified parameter. First time at the start
 * of each report run, asks user for parameter values.
 *
 * @param paramId a parameter id
 */
public Object getParameter(Object paramId) {
    if (!askedForParameters)
	askForParameters();

    return Parameter.find(paramId).getValue();
}

/**
 * Asks the user for parameter values. If the user cancels, we throw a
 * <code>UserCancellationException</code>.
 */
protected void askForParameters() {
    List usedParameters = null;
    if (askedForParameters
	|| (usedParameters = collectUsedParameters()).size() == 0)
    {
	askedForParameters = true;
	return;
    }

    if (ErrorHandler.usingGUI()) {
	DesignWin win = DesignWin.findWindowFor(this);
	if (parametersHaveValues) {
	    // Ask user about re-using the previous values
	    String msg = "Use previous parameter values?";
	    if (JOptionPane.showConfirmDialog(win, msg,
					      "Report parameters",
					      JOptionPane.YES_NO_OPTION)
		== JOptionPane.YES_OPTION)
		{
		    askedForParameters = true;
		    parametersHaveValues = true;
		    return;
		}
	}

	// This dialog is modal. By the time it returns, the parameters
	// have their new values.
	    
	if (new ParamAskWin(win, usedParameters).userCancelled())
	    throw new UserCancellationException("Operation cancelled by user");
    }
    else {
	if (paramXMLFileName == null) {
	    ErrorHandler.error("Parameter XML file name is missing; please"
			       + " specify one on the command line. Report"
			       + " run has been cancelled.");
	    throw new UserCancellationException("Missing parameter XML file");
	}
	try {
	    new ParameterReader(this).readFile(paramXMLFileName);
	}
	catch (Exception ex) {
	    ErrorHandler.error("Error while reading parameter XML file "
			       + paramXMLFileName
			       + "; report run has been cancelled.", ex);
	    throw new UserCancellationException("Missing parameter XML file");
	}
    }

    askedForParameters = true;
    parametersHaveValues = true;
}

/**
 * Spawns a new thread that runs the report, using the layout engine
 * to generate output.
 */
public void run() {
    new Thread(new Runnable() {
	public void run() { Report.this.runReport(); }
	}).start();
}

protected void runReport() {
    askedForParameters = false;	// Ask user again

    for (Iterator iter = groups.iterator(); iter.hasNext(); )
	((Group)iter.next()).reset();

    collectSubtotalFields();

    Connection conn = null;
    PreparedStatement stmt = null;
    rset = null;
    StatusDialog statusDialog = null;
    try {
	if (ErrorHandler.usingGUI()) {
	    DesignWin win = DesignWin.findWindowFor(this);
	    statusDialog = new StatusDialog(win, "Report Run",
					    "Running report");
	}

	conn = database.getConnection();

	// Suggested by Konstantin. Though it works for his Oracle driver,
	// it doesn't work for my PostgreSQL driver. These args are also
	// legal for prepared statements.
//  	stmt = conn.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE,
//  				    ResultSet.CONCUR_READ_ONLY);

	stmt = conn.prepareStatement(query.toPreparedStatementString());
	query.setParameters(stmt);
	rset = stmt.executeQuery();

	boolean layoutStarted = false;
	while (rset.next()) {
	    if (statusDialog != null)
		statusDialog.update("Processing row " + rowNumber());

	    if (!layoutStarted) {
		layoutEngine.start();
		layoutStarted = true;
	    }
	    processResultRow();
	}

	rset.last();		// Recall last row so we can access the data
	if (!layoutStarted) {	// No rows in report
	    layoutEngine.start();
	    layoutEngine.end();
	}
	else {			// Output group footers and end of report
	    layoutEngine.groupFooters(true);
	    layoutEngine.end();
	}
    }
    catch (UserCancellationException uce) {}
    catch (Exception e) {
	ErrorHandler.error(e);
    }
    finally {
	try {
	    if (rset != null) rset.close();
	    if (stmt != null) stmt.close();
	}
	catch (SQLException sqle2) {}

	subtotalFields = null;
	if (statusDialog != null)
	    statusDialog.dispose();
    }
}

/**
 * Processes a single database row. Note that the <code>next</code>
 * method of the result set has already been called.
 */
protected void processResultRow() throws java.sql.SQLException {
    updateGroups();

    // To output footers, bring back the previous row of data
    if (!rset.isFirst()) {
	rset.previous();
	layoutEngine.groupFooters(false);
	rset.next();
    }

    updateSubtotals();

    boolean isLastRow = rset.isLast();
    layoutEngine.groupHeaders(isLastRow);
    layoutEngine.detail(isLastRow);
}

/**
 * Returns the current value of the specified column. Only defined
 * when running a report.
 *
 * @return the string or double value of the column
 */
public Object columnValue(Column column) {
    try {
	// Ask query for field number
	return rset.getObject(query.indexOfColumn(column) + 1);
    }
    catch (SQLException e) {
	ErrorHandler.error(e);
	return null;
    }
}

/**
 * Returns the current page number. Asks the layout engine. Only defined
 * when running a report.
 *
 * @return a page number
 */
public int pageNumber() {
    return layoutEngine.pageNumber;
}

/**
 * Returns the current data row number. Only defined when running a report.
 */
public int rowNumber() {
    try {
	return rset.getRow();
    }
    catch (SQLException e) {
	return 0;
    }
}

/**
 * Iterates over all sections in the report, passing the section to the
 * specified <code>SectionWalker</code>. The sections are visited in the
 * order in which they will be displayed.
 *
 * @param s a section walker
 */
public void withSectionsDo(SectionWalker s) {
    for (Iterator iter = reportHeaders.iterator(); iter.hasNext(); )
	s.step((Section)iter.next());
    for (Iterator iter = pageHeaders.iterator(); iter.hasNext(); )
	s.step((Section)iter.next());
    for (Iterator iter = groups.iterator(); iter.hasNext(); )
	for (Iterator iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); )
	    s.step((Section)iter2.next());
    for (Iterator iter = details.iterator(); iter.hasNext(); )
	s.step((Section)iter.next());
    for (Iterator iter = groups.iterator(); iter.hasNext(); )
	for (Iterator iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); )
	    s.step((Section)iter2.next());
    for (Iterator iter = reportFooters.iterator(); iter.hasNext(); )
	s.step((Section)iter.next());
    for (Iterator iter = pageFooters.iterator(); iter.hasNext(); )
	s.step((Section)iter.next());
}

/**
 * Iterates over all fields in the report, passing the section to the
 * specified <code>FieldWalker</code>.
 *
 * @param f a field walker
 */
public void withFieldsDo(final FieldWalker f) {
    withSectionsDo(new SectionWalker() {
	public void step(Section s) {
	    for (Iterator iter = s.fields.iterator(); iter.hasNext(); )
		f.step((Field)iter.next());
	}
	});
}

/**
 * Collects all subtotal fields and lets each one initialize itself.
 * Used once at the beginning of each run.
 */
protected void collectSubtotalFields() {
    subtotalFields = new ArrayList();
    withFieldsDo(new FieldWalker() {
	public void step(Field f) {
	    if (f instanceof SubtotalField) {
		((SubtotalField)f).initialize();
		subtotalFields.add(f);
	    }
	}
	});
}

/**
 * Collects all subtotal fields that are subtotaling the specified field.
 * Used by the report design GUI.
 * 
 * @param field a field
 * @return a list of subtotal fields
 */
public AbstractList getSubtotalFieldsFor(final Field field) {
    final ArrayList subs = new ArrayList();
    withFieldsDo(new FieldWalker() {
	public void step(Field f) {
	    if (f instanceof SubtotalField
		&& ((SubtotalField)f).getField() == field)
	    {
		subs.add(f);
	    }
	}
	});
    return subs;
}

/**
 * Updates each subtotal field.
 */
protected void updateSubtotals() {
    for (Iterator iter = subtotalFields.iterator(); iter.hasNext(); )
	((SubtotalField)iter.next()).updateSubtotal();
}

/**
 * Updates each group's value based on the current value of the column
 * each group uses.
 */
protected void updateGroups() {
    for (Iterator iter = groups.iterator(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	g.setValue(columnValue(g.column));
    }
}

/**
 * Remembers name of parameter XML file.
 *
 * @param fileName an XML file name
 */
public void setParameterXMLFile(String fileName)
{
    paramXMLFileName = fileName;
}

/**
 * Read an XML file and build the contents of this report. Uses a
 * <code>ReportReader</code>.
 *
 * @param fileName a file name string
 */
public void readFile(String fileName)
    throws FileNotFoundException, IOException, SQLException,
	   ClassNotFoundException, InstantiationException,
	   IllegalAccessException, SAXException, UserCancellationException
{
    new ReportReader(this).readFile(fileName);
}

/**
 * Writes the contents of this report as an XML file.
 *
 * @param fileName a file name string
 */
public void writeFile(String fileName) {
    IndentWriter out = null;
    try {
	out = new IndentWriter(new FileWriter(fileName));
	writeXML(out);
	out.close();
    }
    catch (IOException ioe) {
	ErrorHandler.error("Error writing to file " + fileName, ioe,
			   "File error");
    }
    finally {
	if (out != null) out.close();
    }
}

/**
 * Writes the contents of this report as an XML file.
 *
 * @param out an indent writer
 */
public void writeXML(IndentWriter out) {
	out.println("<?xml version=\"1.0\"?>");

	out.println("<report dtd-version=\"" + dtdVersion + "\" name=\""
		    + Utils.escapeXML(name) + "\" title=\""
		    + Utils.escapeXML(title) + "\"");
	out.indent();
	out.indent();
	out.println("author=\"" + Utils.escapeXML(author) + "\">");
	out.outdent();

	out.println();
	out.println("<description><![CDATA["
		    + (description == null ? "" : description)
		    + "]]></description>");

	out.println();
	database.writeXML(out);

	out.println();
	query.writeXML(out);

	Iterator iter;

	out.println();
	out.println("<parameters>");
	out.indent();
	for (iter = parameters.iterator(); iter.hasNext(); )
	     ((Parameter)iter.next()).writeXML(out);
	out.outdent();
	out.println("</parameters>");

	out.println();
	out.println("<formulas>");
	out.indent();
	for (iter = formulas.iterator(); iter.hasNext(); )
	     ((Formula)iter.next()).writeXML(out);
	out.outdent();
	out.println("</formulas>");

	out.println();
	ListWriter.writeList(out, reportHeaders, "headers");
	out.println();
	ListWriter.writeList(out, reportFooters, "footers");

	out.println();
	out.println("<page>");
	out.indent();
	ListWriter.writeList(out, pageHeaders, "headers");
	ListWriter.writeList(out, pageFooters, "footers");
	out.outdent();
	out.println("</page>");

	out.println();
	out.println("<groups>");
	out.indent();
	for (iter = groups.iterator(); iter.hasNext(); ) {
	    out.println();
	    ((Group)iter.next()).writeXML(out);
	}
	out.outdent();
	out.println("</groups>");

	out.println();
	ListWriter.writeList(out, details, "details");

	out.outdent();
	out.println();
	out.println("</report>");
}

}
