package jimm.datavision;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Observer;
import java.util.Observable;

/**
 * A section of a report contains a group of fields and suppression
 * information. Sections include page headers and footers, report headers
 * and footers, group headers and footers, and details. Sections contain
 * elements (fields and lines).
 *
 * @see Report
 * @see Field
 * @see Line
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Section
    extends Observable
    implements Dirtiable, Writeable, Observer
{

protected static final double DEFAULT_HEIGHT = 20;

protected Report report;
protected String name;
protected double height;
protected ArrayList fields;
protected ArrayList lines;
protected boolean suppressed;
protected SuppressionProc suppressionProc;
protected boolean isDetail;
protected boolean dirty;

/**
 * Constructor. Uses <code>DEFAULT_HEIGHT</code>.
 *
 * @param report the report containing this section
 * @param name the name of the section (for example, "Group Header")
 * @param isDetail if <code>true<code> this section is a report detail
 * section
 */
public Section(Report r, String name, boolean isDetail) {
    this(r, name, DEFAULT_HEIGHT, isDetail);
}

/**
 * Constructor.
 *
 * @param report the report containing this section
 * @param name the name of the section (for example, "Group Header")
 * @param h the height
 * @param isDetail if <code>true<code> this section is a report detail
 * section
 */
public Section(Report r, String name, double h, boolean isDetail) {
    report = r;
    this.name = name;
    height = h;
    this.isDetail = isDetail;
    fields = new ArrayList();
    lines = new ArrayList();
    suppressed = false;
}

public void update(Observable o, Object arg) {
    setChanged();
    notifyObservers(arg);
}

/**
 * Returns <code>true</code> if this section or any of its elements (fields
 * and lines) have been modified.
 *
 * @return <code>true</code> if this section or its contents have been
 * modified
 */
public boolean isDirty() {
    if (dirty) return true;
    for (Iterator iter = fields(); iter.hasNext(); )
	if (((Field)iter.next()).isDirty()) return true;
    for (Iterator iter = lines(); iter.hasNext(); )
	if (((Line)iter.next()).isDirty()) return true;
    return false;
}

/**
 * Sets or clears the dirty bit. If <i>dirty</i> is <code>false</code>,
 * clears the dirty bit for each element (field or line) in the section.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) {
    this.dirty = dirty;
    if (!dirty) {
	for (Iterator iter = fields(); iter.hasNext(); )
	    ((Field)iter.next()).setDirty(false);
	for (Iterator iter = lines(); iter.hasNext(); )
	    ((Line)iter.next()).setDirty(false);
    }
}

/**
 * Returns the report containing this section.
 */
public Report getReport() { return report; }

/**
 * Returns the height of this section.
 *
 * @return the height
 */
public double getHeight() { return height; }

/**
 * Sets the height.
 *
 * @param newHeight the new height
 */
public void setHeight(double newHeight) {
    if (height != newHeight) {
	height = newHeight;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the name of this section.
 *
 * @return the section name
 */
public String getName() { return name; }
public void setName(String newName) {
    if (name != newName && (name == null || !name.equals(newName))) {
	name = newName;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Adds a field to this section.
 *
 * @param f a field
 */
public void addField(Field f) {
    fields.add(f);
    f.addObserver(this);
    f.setSection(this);
    setDirty(true);
    setChanged();
    notifyObservers();
}

/**
 * Removes a field from this section.
 *
 * @param f field
 */
public void removeField(Field f) {
    fields.remove(f);
    f.deleteObserver(this);
    f.setSection(null);
    setDirty(true);
    setChanged();
    notifyObservers();
}

/**
 * Returns an iterator over all fields in this section.
 *
 * @return an iterator
 */
public Iterator fields() { return fields.iterator(); }

/**
 * Adds a line to this section.
 *
 * @param f a line
 */
public void addLine(Line l) {
    lines.add(l);
    setDirty(true);
    setChanged();
    notifyObservers();
}

/**
 * Removes a line from this section.
 *
 * @param f line
 */
public void removeLine(Line f) {
    lines.remove(f);
    setDirty(true);
    setChanged();
    notifyObservers();
}

/**
 * Returns an iterator over all lines in this section.
 *
 * @return an iterator
 */
public Iterator lines() { return lines.iterator(); }

public boolean isSuppressed() { return suppressed; }
public void setSuppressed(boolean newSuppressed) {
    if (suppressed != newSuppressed) {
	suppressed = newSuppressed;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns <code>true</code> if this is a report detail section.
 *
 * @return <code>true</code> if this is a report detail section
 */
public boolean isDetail() { return isDetail; }

/**
 * Tells this section if it is a detail section. Since sections can't be
 * moved around, this is only called by report creation code such as
 * <code>ReportReader.section</code>
 *
 * @param val new boolean value
 */
public void setIsDetail(boolean val) {
    if (isDetail != val) {
	isDetail = val;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the supression proc this section uses
 *
 * @return a suppression proc
 */
public SuppressionProc getSuppressionProc() { return suppressionProc; }

/**
 * Sets the suppression proc to use.
 *
 * @param newSuppressionProc a new suppression proc
 */
public void setSuppressionProc(SuppressionProc newSuppressionProc) {
    if (suppressionProc != newSuppressionProc) {
	suppressionProc = newSuppressionProc;
	setDirty(true);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns <code>true</code> if the specified field is inside this section.
 *
 * @param f a field
 * @return <code>true</code> if the field is within this section
 */
public boolean contains(Field f) {
    for (Iterator iter = fields.iterator(); iter.hasNext(); )
	if ((Field)iter.next() == f)
	    return true;
    return false;
}

/**
 * Returns <code>true</code> if the specified field exists within this
 * section either directly (as a field) or indirectly (as a formula used
 * by a subtotal or by another formula).
 *
 * @param f a field
 * @return <code>true</code> if the specified field is referenced within
 * this section
 */
public boolean containsReferenceTo(Field f) {
    for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
	Field field = (Field)iter.next();
	if (field == f || field.refersTo(f))
	    return true;
    }
    return false;
}

/**
 * Returns <code>true</code> if the specified formula exists within this
 * secion either directly (as a formula field) or indirectly (as a formula
 * used by a subtotal or by another formula).
 *
 * @param f a formula
 * @return <code>true</code> if the specified formula is referenced within
 * this section
 */
public boolean containsReferenceTo(Formula f) {
    for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
	Field field = (Field)iter.next();
	if (field.refersTo(f))
	    return true;
    }
    return false;
}

/**
 * Returns <code>true</code> if the specified parameter exists within this
 * secion either directly (as a parameter field) or indirectly (as a parameter
 * used by a subtotal or by a formula).
 *
 * @param p a parameter
 * @return <code>true</code> if the specified formula is referenced within
 * this section
 */
public boolean containsReferenceTo(Parameter p) {
    for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
	Field field = (Field)iter.next();
	if (field.refersTo(p))
	    return true;
    }
    return false;
}

/**
 * Return true if this section should be printed, given this particular
 * row of data and our supressed state and suppression proc.
 *
 * @param row a row of data
 * @return a boolean specifying if the data should be displayed or not
 */
public boolean isActiveGivenData(ResultSet row) {
    if (suppressed) return false;
    if (suppressionProc == null) return true;
    return suppressionProc.suppressGivenData(row);
}

/**
 * Writes this section and all it contains as an XML tag.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.print("<section height=\"" + height + "\"");
    if (suppressed) out.print(" suppressed=\"true\"");
    out.println(">");
    out.indent();

    for (Iterator iter = fields.iterator(); iter.hasNext(); )
	((Field)iter.next()).writeXML(out);
    for (Iterator iter = lines.iterator(); iter.hasNext(); )
	((Line)iter.next()).writeXML(out);
// FIX
// Suppression proc

    out.outdent();
    out.println("</section>");
}

}
