package jimm.datavision;
import java.util.HashMap;

/**
 * A SpecialField represents a special value such as the report name or
 * current page number. The value of a SpecialField is a string that
 * identifies which value to display.
 * <ul>
 * <li>report.name</li>
 * <li>report.title</li>
 * <li>report.author</li>
 * <li>report.description</li>
 * <li>report.date</li>
 * <li>report.row</li>
 * <li>page.number</li>
 * </ul>
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class SpecialField extends Field {

public static final String TYPE_STRING = "special";
protected static final HashMap SPECIAL_FIELD_NAMES = new HashMap();

/**
 * Returns an array of all of the special field names.
 *
 * @return an array strings containing all of the special field names
 */
public static HashMap specialFieldNames() {
    if (SPECIAL_FIELD_NAMES.get("report.title") == null) {
	SPECIAL_FIELD_NAMES.put("report.title", "Report Title");
	SPECIAL_FIELD_NAMES.put("report.name", "Report Name");
	SPECIAL_FIELD_NAMES.put("report.author", "Author's Name");
	SPECIAL_FIELD_NAMES.put("report.description", "Report Description");
	SPECIAL_FIELD_NAMES.put("report.date", "Report Date");
	SPECIAL_FIELD_NAMES.put("report.row", "Report Row");
	SPECIAL_FIELD_NAMES.put("page.number", "Page Number");
    }
    return SPECIAL_FIELD_NAMES;
}

/**
 * Returns the value associated with the given special field name string.
 *
 * @param str the special field name string
 * @param report the report that is running
 * @return a string
 */
public static String value(String str, Report report) {
    if ("report.title".equals(str))
	return report.title;
    if ("report.name".equals(str))
	return report.name;
    if ("report.author".equals(str))
	return report.author;
    if ("report.description".equals(str))
	return report.description;
    else if ("report.date".equals(str))
	return new java.util.Date().toString();
    else if ("report.row".equals(str))
	return "" + report.rowNumber();
    else if ("page.number".equals(str))
	return "Page " + report.pageNumber();
    else
	return "<unknown special field>";
}

/**
 * Constructs a special field with the specified id in the specified report
 * section whose special value is <i>value</i>.
 *
 * @param id the new field's id
 * @param section the report section in which the field resides
 * @param value the magic string
 */
public SpecialField(Long id, Section section, Object value) {
    super(id, section, value);
}

public String dragString() {
    return typeString() + ":" + value;
}

public String typeString() { return TYPE_STRING; }

public String designLabel() { return "{" + value + "}"; }

public String formulaString() { return "{%" + value + "}"; }

/**
 * Returns the value of this field.
 *
 * @return a string
 */
public Object getValue() {
    return SpecialField.value((String)value, section.report);
}
}
