package jimm.datavision.gui;
import jimm.datavision.Report;
import jimm.datavision.ErrorHandler;
import jimm.datavision.sql.Database;
import java.io.IOException;
import java.awt.GridLayout;
import java.awt.BorderLayout;
import java.awt.event.*;
import java.util.ArrayList;
import javax.swing.*;
import org.xml.sax.SAXException;

/**
 * A database connection editing dialog box. The user can either enter
 * separate values or copy values from an existing report XML file.
 *
 * @see DbConnReader
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class DbConnWin extends EditWin implements ActionListener {

class RevertInfo {
boolean reportDirty;
Database database;
boolean dirty;
String driverClassName;
String connInfo;
String dbName;
String userName;
String password;
RevertInfo(boolean reportDirty, Database db, boolean dirty, String className,
	   String connInfo, String dbName, String user, String password)
{
    this.reportDirty = reportDirty;
    database = db;
    this.dirty = dirty;
    driverClassName = className;
    this.connInfo = connInfo;
    this.dbName = dbName;
    userName = user;
    this.password = password;
}
}

protected static final int TEXT_FIELD_COLS = 32;

protected DesignWin designWin;
protected Report report;
protected JTextField driverClassNameField;
protected JTextField connInfoField;
protected JTextField dbNameField;
protected JTextField userNameField;
protected JPasswordField passwordField;

/**
 * Constructor.
 *
 * @param designWin the window to which this dialog belongs
 * @param database report database connection
 */
public DbConnWin(DesignWin designWin, Report report, boolean modal) {
    super(designWin, "Database Connection", modal);

    this.designWin = designWin;
    this.report = report;

    buildWindow();
    pack();
    setVisible(true);
}

/**
 * Builds the window contents.
 */
protected void buildWindow() {
    JPanel editorPanel = buildEditor();

    // Ok, Apply, Revert, and Cancel Buttons
    JPanel buttonPanel = closeButtonPanel();

    // Add values and buttons to window
    getContentPane().add(editorPanel, BorderLayout.CENTER);
    getContentPane().add(buttonPanel, BorderLayout.SOUTH);
}

protected JPanel buildEditor() {
    // Labels
    JPanel labelPanel = new JPanel();
    labelPanel.setLayout(new GridLayout(0, 1));
    JLabel label;
    labelPanel.add(label = new JLabel("Driver Class Name:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Connection Info:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Database Name:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("User Name:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Password:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);

    Database db = report.getDatabase();

    // Fields
    JPanel fieldPanel = new JPanel();
    fieldPanel.setLayout(new GridLayout(0, 1));

    driverClassNameField =
	new JTextField(db == null ? "" : db.getDriverClassName(),
		       TEXT_FIELD_COLS);
    fieldPanel.add(driverClassNameField);

    connInfoField =
	new JTextField(db == null ? "" : db.getConnectionInfo(),
		       TEXT_FIELD_COLS);
    fieldPanel.add(connInfoField);

    dbNameField = new JTextField(db == null ? "" : db.getName(),
				 TEXT_FIELD_COLS);
    fieldPanel.add(dbNameField);

    userNameField = new JTextField(db == null ? "" : db.getUserName(),
				   TEXT_FIELD_COLS);
    fieldPanel.add(userNameField);

    String password = (db == null ? "" : db.getPassword());
    if (password == null) password = "";
    passwordField = new JPasswordField(password, TEXT_FIELD_COLS);
    fieldPanel.add(passwordField);

    JPanel panel = new JPanel();
    panel.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
    panel.setLayout(new BorderLayout());
    panel.add(labelPanel, BorderLayout.CENTER);
    panel.add(fieldPanel, BorderLayout.EAST);

    // Click to copy info from another report
    JButton copyButton = new JButton("Copy Settings...");
    copyButton.addActionListener(this);
    JPanel copyPanel = new JPanel();
    copyPanel.add(copyButton);
    panel.add(copyPanel, BorderLayout.SOUTH);

    return panel;
}

protected void fillEditFields(String driverClassName, String connInfo,
			      String dbName, String username)
{
    driverClassNameField.setText(driverClassName);
    connInfoField.setText(connInfo);
    dbNameField.setText(dbName);
    userNameField.setText(username);
    passwordField.setText("");
}

/**
 * Handles the "Copy Settings..." button.
 *
 * @param e action event
 */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();
    if ("Copy Settings...".equals(cmd)) {
	JFileChooser chooser = designWin.getChooser();
	int returnVal = chooser.showOpenDialog(this);
	if (returnVal == JFileChooser.APPROVE_OPTION) {
	    String path = chooser.getSelectedFile().getPath();
	    DbConnReader reader = new DbConnReader();
	    try {
		reader.readFile(path);
		driverClassNameField.setText(reader.getDriverClassName());
		connInfoField.setText(reader.getConnectionInfo());
		dbNameField.setText(reader.getDbName());
		userNameField.setText(reader.getUserName());
	    }
	    catch (IOException ioe) {
		ErrorHandler.error("Error copying settings", ioe);
	    }
	    catch (SAXException saxe) {
		ErrorHandler.error("Error copying settings", saxe);
	    }
	}
    }
    else
	super.actionPerformed(e);
}

protected void doSave(boolean saveRevertInfo) {
    try {
	Database db = report.getDatabase();
	if (db == null) {
	    if (saveRevertInfo && !revertInfoList.containsKey(this)) {
		RevertInfo ri =
		    new RevertInfo(report.isDirty(), null, false, null, null,
				   null, null, null);
		revertInfoList.put(this, ri);
	    }

	    db = new Database(driverClassNameField.getText(),
			      connInfoField.getText(), report,
			      dbNameField.getText(), userNameField.getText(),
			      new String(passwordField.getPassword()));
	    report.setDatabase(db);
	}
	else {
	    if (saveRevertInfo && !revertInfoList.containsKey(this)) {
		RevertInfo ri =
		    new RevertInfo(report.isDirty(), db, db.isDirty(),
				   db.getDriverClassName(),
				   db.getConnectionInfo(), db.getName(),
				   db.getUserName(), db.getPassword());
		revertInfoList.put(this, ri);
	    }

	    db.reset(driverClassNameField.getText(), connInfoField.getText(),
		     dbNameField.getText(), userNameField.getText(),
		     new String(passwordField.getPassword()));
	}
    }
    catch (Exception e) {
	ErrorHandler.error(e, "Error connecting to database");
    }
}

protected void doRevert(Object revertInfo) {
    RevertInfo ri = (RevertInfo)revertInfo;
    Database db = ri.database;
    if (db == null)
	report.setDatabase(null);
    else {
	try {
	    db.reset(ri.driverClassName, ri.connInfo, ri.dbName, ri.userName,
		     ri.password);
	    db.setDirty(ri.dirty);
	}
	catch (Exception e) {
	    ErrorHandler.error("An error occurred while trying to revert\n"
			       + "the database information. The error will\n"
			       + "be ignored, but the error message was:\n",
			       e, "Error reverting info");
	}
    }
    report.setDirty(ri.reportDirty);

    // Revert GUI
    fillEditFields(ri.driverClassName, ri.connInfo, ri.dbName, ri.userName);
}

}
