package jimm.datavision.gui;
import jimm.datavision.*;
import jimm.datavision.layout.SwingLE;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.MouseInputAdapter;

/**
 * The main GUI report design window.
 *
 * @see Report
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class DesignWin extends JFrame implements ActionListener, Observer {

public static final int GRID_SIZE = 8;

public static final int ALIGN_TOP = 0;
public static final int ALIGN_MIDDLE = 1;
public static final int ALIGN_BOTTOM = 2;
public static final int ALIGN_LEFT = 3;
public static final int ALIGN_CENTER = 4;
public static final int ALIGN_RIGHT = 5;
public static final int ALIGN_SNAP_TO_GRID = 6;

public static final int SIZE_SAME_WIDTH = 0;
public static final int SIZE_SAME_HEIGHT = 1;
public static final int SIZE_SAME_SIZE = 2;

protected static final String REPORT_TEMP_DIR = "datavision";
protected static ArrayList designWindows = new ArrayList();

protected Report report;
protected ArrayList sectionWidgets;
protected String reportFileName;
protected JFileChooser chooser;
protected JLayeredPane sectionContainer;
protected ArrayList selectedFields;
protected JScrollPane scroller;
protected JMenuItem undoItem, cutItem, copyItem, pasteItem, delSelectionItem,
    delGroupItem, delSectionItem, subtotItem, groupItem, sectItem, formatItem,
    borderItem, boundsItem;
protected JMenu alignSubmenu, sizeSubmenu;
protected boolean placingNewTextField;
protected boolean ignoreKeys;

/**
 * Adds a window to the list of open design windows. Also notifies the
 * error handler class that we it should use the GUI to display error
 * messages.
 *
 * @param win a design window
 */
public static void addWindow(DesignWin win) {
    designWindows.add(win);
}

/**
 * Returns the design window associated with the specified report, or
 * <code>null</code> if one is not found.
 *
 * @param r a report
 * @return a design window or <code>null</code> if one is not found
 */
public static DesignWin findWindowFor(Report r) {
    for (Iterator iter = designWindows.iterator(); iter.hasNext(); ) {
	DesignWin win = (DesignWin)iter.next();
	if (win.report == r)
	    return win;
    }
    return null;
}

/**
 * Deletes a window from the list of windows. If there are no more design
 * windows left, exits the application.
 *
 * @param win a design window
 */
protected static void deleteWindow(DesignWin win) {
    designWindows.remove(win);
    if (designWindows.isEmpty())
	System.exit(0);
}

/**
 * Closes each open design window (unless user cancels). If all are closed,
 * quits the application.
 */
protected static void maybeQuit() {
    // Can't use iterator over original list 'cause we're deleting items
    // from the list.
    ArrayList copy = (ArrayList)designWindows.clone();
    for (Iterator iter = copy.iterator(); iter.hasNext(); )
	((DesignWin)iter.next()).maybeClose();
}

/**
 * Constructor. Reads the named report file or, if it's <code>null</code>,
 * creates a new, empty report.
 *
 * @param fileName an XML report file name; may be <code>null</code>
 */
public DesignWin(String fileName) {
    this(fileName, null);
}

/**
 * Constructor. Reads the named report file or, if it's <code>null</code>,
 * creates a new, empty report.
 *
 * @param fileName an XML report file name; may be <code>null</code>
 * @param databasePassword string to give to report; OK if it's
 * <code>null</code>
 */
public DesignWin(String fileName, String databasePassword) {
    super("DataVision Report Designer");
    selectedFields = new ArrayList();

    ErrorHandler.useGUI(true);	// Must set before opening report
    boolean askForDbConnInfo = readReport(fileName, databasePassword);

    if (report == null || (report.getDatabase() == null && !askForDbConnInfo))
    {
	// User cancelled when asked for password. Close this window and
	// quit if this is the only report open.
	dispose();
	DesignWin.deleteWindow(this); // Will quit if no other windows
	return;
    }

    if (askForDbConnInfo) {
	openDbConnWin(true);	// Modally ask for database connection info
	if (report.getDatabase() == null) {
	    // User cancelled. Close this window and quit if this is the
	    // only report open.
	    dispose();
	    DesignWin.deleteWindow(this); // Will quit if no other windows
	    return;
	}
    }

    setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
    buildWindow();
    enableMenuItems();

    // Make sure we close the window when the user asks to close the window.
    addWindowListener(new WindowAdapter() {
	public void windowClosing(WindowEvent e) {
	    maybeClose();
	}
    });

    // Might be using this code for dragging to select multiple fields.
//      SelectionHandler h = new SelectionHandler();
//      addMouseListener(h);
//      addMouseMotionListener(h);

    pack();
    show();
    DesignWin.addWindow(this);
}

public void update(Observable o, Object arg) {
    enableMenuItems();
}

/**
 * Reads the named report file or, if it's <code>null</code>, creates
 * a new, empty report. Returns <code>true</code> if we need to ask
 * the user for connection info because this is a new report.
 *
 * @param fileName
 * @param databasePassword string to give to report; OK if it's
 * <code>null</code>
 * @return <code>true</code> if we need to ask the user for connection info
 */
protected boolean readReport(String fileName, String databasePassword) {
    if (fileName != null && fileName.length() > 0) {
	reportFileName = fileName;
	report = new Report();
	report.setDatabasePassword(databasePassword);
	try {
	    report.readFile(fileName);
	    return false;
	}
	catch (UserCancellationException uce) {
	    report = null;	// Signal report open cancelled
	    return true;	// Doesn't matter what we return
	}
	catch (Exception e) {
	    ErrorHandler.error(e);
	    report = new Report();
	    return true;
	}
    }
    else {
	report = new Report();
	reportFileName = null;
	return true;
    }
}

void setIgnoreKeys(boolean ignore) { ignoreKeys = ignore; }

/**
 * Builds the window components.
 */
protected void buildWindow() {
    buildMenuBar();
    buildSections();
    // Set up key handling (also put inside section canvas?)
    addKeyListener(new KeyAdapter() {
	public void keyPressed(KeyEvent e) {
	    if (!ignoreKeys) {
		int code = e.getKeyCode();
		if (code == KeyEvent.VK_BACK_SPACE
		    || code == KeyEvent.VK_DELETE)
		    deleteSelectedFields();
	    }
	}
	});
}

/**
 * Builds the window menu bar.
 */
protected void buildMenuBar() {
    JMenuBar bar = new JMenuBar();
    bar.add(buildFileMenu());
    bar.add(buildEditMenu());
    bar.add(buildInsertMenu());
    bar.add(buildFormatMenu());
    bar.add(buildDatabaseMenu());
    bar.add(buildReportMenu());
    bar.add(Box.createHorizontalGlue());
    bar.add(buildHelpMenu());
    setJMenuBar(bar);
}

/**
 * Builds and returns the "File" menu.
 *
 * @return a menu
 */
protected JMenu buildFileMenu() {
    JMenu menu = new JMenu("File");
    menu.setMnemonic(KeyEvent.VK_F);

    MenuUtils.addToMenu(this, menu, "New", "new", KeyEvent.VK_N);
    MenuUtils.addToMenu(this, menu, "Open", "open", KeyEvent.VK_O);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, "Save", "save", KeyEvent.VK_S);
    MenuUtils.addToMenu(this, menu, "Save As...", "save_as", KeyEvent.VK_A, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, "Close", "close", KeyEvent.VK_C,
			KeyEvent.VK_W);
    MenuUtils.addToMenu(this, menu, "Quit", "quit", KeyEvent.VK_Q);

    return menu;
}

/**
 * Builds and returns the "Edit" menu.
 *
 * @return a menu
 */
protected JMenu buildEditMenu() {
    JMenu menu = new JMenu("Edit");
    menu.setMnemonic(KeyEvent.VK_E);

    undoItem = MenuUtils.addToMenu(this, menu, "Undo", "undo", KeyEvent.VK_U,
				   KeyEvent.VK_Z);
    undoItem.setEnabled(false);
    menu.addSeparator();
    cutItem = MenuUtils.addToMenu(this, menu, "Cut", "cut", KeyEvent.VK_C,
				  KeyEvent.VK_X);
    copyItem = MenuUtils.addToMenu(this, menu, "Copy", "copy", KeyEvent.VK_O,
				   KeyEvent.VK_C);
    copyItem.setEnabled(false);
    pasteItem = MenuUtils.addToMenu(this, menu, "Paste", "paste",
				    KeyEvent.VK_P, KeyEvent.VK_V);
    pasteItem.setEnabled(false);
    menu.addSeparator();
    delSelectionItem = MenuUtils.addToMenu(this, menu,
					   "Delete Selected Fields",
					   "del_selection", KeyEvent.VK_F,
					   KeyEvent.VK_DELETE, 0);
    menu.addSeparator();
    delGroupItem = MenuUtils.addToMenu(this, menu, "Delete Group", "del_group",
				       KeyEvent.VK_G, 0);
    delSectionItem = MenuUtils.addToMenu(this, menu, "Delete Section",
					 "del_section", KeyEvent.VK_S, 0);

    return menu;
}

/**
 * Builds and returns the "Insert" menu.
 *
 * @return a menu
 */
protected JMenu buildInsertMenu() {
    JMenu menu = new JMenu("Insert");
    menu.setMnemonic(KeyEvent.VK_I);

    MenuUtils.addToMenu(this, menu, "Database Field...", "database_field",
			KeyEvent.VK_D, 0);
    MenuUtils.addToMenu(this, menu, "Text", "text", KeyEvent.VK_T, 0);
    MenuUtils.addToMenu(this, menu, "Formula Field...", "formula_field",
			KeyEvent.VK_F, 0);
    MenuUtils.addToMenu(this, menu, "Parameter Field...", "parameter_field",
			KeyEvent.VK_P, 0);
    MenuUtils.addToMenu(this, menu, "Special Field...", "special_field",
			KeyEvent.VK_S, 0);
    subtotItem = MenuUtils.addToMenu(this, menu, "Running Total Field...",
				     "running_total_field", KeyEvent.VK_R, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, "Line", "line", KeyEvent.VK_L, 0)
	.setEnabled(false);
    menu.addSeparator();
    groupItem = MenuUtils.addToMenu(this, menu, "Group...", "add_group",
				    KeyEvent.VK_G, 0);
    sectItem = MenuUtils.addToMenu(this, menu, "Section", "ins_section",
				   KeyEvent.VK_E, 0);

    return menu;
}

/**
 * Builds and returns the "Format" menu.
 *
 * @return a menu
 */
protected JMenu buildFormatMenu() {
    JMenu menu = new JMenu("Format");
    menu.setMnemonic(KeyEvent.VK_O);

    formatItem = MenuUtils.addToMenu(this, menu, "Format Field...",
				     "format_field", KeyEvent.VK_F, 0);
    borderItem = MenuUtils.addToMenu(this, menu, "Border...", "border",
				     KeyEvent.VK_B, 0);
    menu.addSeparator();
    boundsItem = MenuUtils.addToMenu(this, menu, "Position and Size...",
				     "position", KeyEvent.VK_P, 0);
    menu.add(alignSubmenu = MenuUtils.buildAlignMenu(this, true, null));
    menu.add(sizeSubmenu = MenuUtils.buildSizeMenu(this, true, null));

    return menu;
}

/**
 * Builds and returns the "Database" menu.
 *
 * @return a menu
 */
protected JMenu buildDatabaseMenu() {
    JMenu menu = new JMenu("Database");
    menu.setMnemonic(KeyEvent.VK_D);

    MenuUtils.addToMenu(this, menu, "Table Linker...", "visual_table_linker",
			KeyEvent.VK_V, 0);
    menu.addSeparator();
//      MenuUtils.addToMenu(this, menu, "Join Tables...", "join_tables",
//  			KeyEvent.VK_J, 0)
//  	.setEnabled(false);
//      menu.addSeparator();
    MenuUtils.addToMenu(this, menu, "SQL Query Text...", "sql_query_text",
			KeyEvent.VK_S, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, "Connection...", "db_connection",
			KeyEvent.VK_C, 0);

    return menu;
}

/**
 * Builds and returns the "Report" menu.
 *
 * @return a menu
 */
protected JMenu buildReportMenu() {
    JMenu menu = new JMenu("Report");
    menu.setMnemonic(KeyEvent.VK_R);

    MenuUtils.addToMenu(this, menu, "Run", "run", KeyEvent.VK_R);
    MenuUtils.addToMenu(this, menu, "Export...", "export", KeyEvent.VK_E, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, "Select Records...", "select_records",
			KeyEvent.VK_S, 0);
    MenuUtils.addToMenu(this, menu, "Sort By...", "sort_by", KeyEvent.VK_O, 0);
    MenuUtils.addToMenu(this, menu, "Group By...", "group_by",
			KeyEvent.VK_G, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, "Summary...", "summary", KeyEvent.VK_U, 0);

    return menu;
}

/**
 * Builds and returns the "Help" menu.
 *
 * @return a menu
 */
protected JMenu buildHelpMenu() {
    JMenu menu = new JMenu("Help");
    menu.setMnemonic(KeyEvent.VK_H);

    MenuUtils.addToMenu(this, menu, "Help", "help", KeyEvent.VK_H, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, "About DataVision...", "about",
			KeyEvent.VK_A, 0);

    return menu;
}

/**
 * Builds the sections and adds them to the section container.
 */
protected void buildSections() {
    getContentPane().setLayout(new BorderLayout());

    sectionWidgets = new ArrayList();
    sectionContainer = new JLayeredPane();
    sectionContainer.setLayout(new DesignWinLayout());

    // Can't just call report.each_section because we want to append
    // '(a)', '(b)', etc. if there is more than one section in the group.
    buildSectionsInList(report.headers());
    buildSectionsInList(report.pageHeaders());

    // Add headers for each group.
    for (Iterator iter = report.groups(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	buildSectionsInList(g.headers());
    }

    buildSectionsInList(report.details());

    // Add footers for each group, where the groups are in reverse order.
    for (Iterator iter = report.groupsReversed(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	buildSectionsInList(g.footers());
    }

    buildSectionsInList(report.footers());
    buildSectionsInList(report.pageFooters());

    renameSectionWidgets();	// Muck with section names (e.g., add " (a)")

    // Put sectionContainer within scroller and add scroller to content
    // pane.
    int width = SectionWidget.LHS_WIDTH + SectionWidget.PAGE_WIDTH;
    scroller = new JScrollPane(sectionContainer);
    scroller.setPreferredSize(new Dimension(width + 4, 400));
    getContentPane().add(scroller, BorderLayout.CENTER);
}

/**
 * Builds the section widgets for the sections in a collection and adds
 * them to the section container. We name the widgets with a separate
 * call to <code>renameSectionWidgets</code>.
 * <p>
 * We also start observing each section. Though our caller is called
 * more than once, it is OK to call addObserver() on an Observable
 * multiple times.
 *
 * @param iter an iterator over a collection of sections
 * @see #renameSectionWidgets
 */
protected void buildSectionsInList(Iterator iter)
{
    while (iter.hasNext()) {
	Section sect = (Section)iter.next();
	sect.addObserver(this);	// Start observing changes

	SectionWidget sw = new SectionWidget(this, sect, "", "");
	sectionWidgets.add(sw);

	// Add to palette layer (one above default layer) because we want
	// to move field widgets below section widgets temporarily at the
	// end of a drag (see putDown()).
	sectionContainer.add(sw, JLayeredPane.PALETTE_LAYER);
    }
}

/**
 * Recalculates section names for all sections in the report. Calls
 * <code>renameSectionWidgetsIn</code> for each group of sections.
 *
 * @see #renameSectionWidgetsIn
 */
protected void renameSectionWidgets() {
    renameSectionWidgetsIn(report.headers(), "Report Header", null);
    renameSectionWidgetsIn(report.pageHeaders(), "Page Header", null);
    int i = 1;
    for (Iterator iter = report.groups(); iter.hasNext(); ++i) {
	Group g = (Group)iter.next();
	renameSectionWidgetsIn(g.headers(), "Group #" + i + " Header", g);
    }
    renameSectionWidgetsIn(report.details(), "Detail", null);
    i = 1;
    for (Iterator iter = report.groups(); iter.hasNext(); ++i) {
	Group g = (Group)iter.next();
	renameSectionWidgetsIn(g.footers(), "Group #" + i + " Footer", g);
    }
    renameSectionWidgetsIn(report.footers(), "Report Footer", null);
    renameSectionWidgetsIn(report.pageFooters(), "Page Footer", null);
}

/**
 * Recalculates section names for a collection of sections. Called
 * from <code>renameSectionWidgets</code>. Also sets the section's popup
 * menu's first item text.
 *
 * @param iter an iterator over a collection of sections
 * @param prefix prepended to all section names in the collection
 * @param group the group containing this section list; may be
 * <code>null</code>
 * @see #renameSectionWidgets
 */
protected void renameSectionWidgetsIn(Iterator iter, String prefix,
				      Group group)
{
    SectionWidget firstSectionWidget = null;
    boolean firstSectionFixed = false;
    for (int i = 0; iter.hasNext(); ++i) {
	Section s = (Section)iter.next();
	SectionWidget sw = findSectionWidgetFor(s);
	if (firstSectionWidget == null) {
	    sw.setDisplayName(prefix);
	    firstSectionWidget = sw;
	}
	else {
	    if (!firstSectionFixed) {
		firstSectionWidget.setDisplayName(prefix + " (a)");
		firstSectionFixed = true;
	    }
	    sw.setDisplayName(prefix + " (" + (char)('a' + i) + ")");
	}
	sw.setPopupName(group == null ? prefix : group.getColumn().fullName());
    }
}

/**
 * Enables or disables popup menu items based on field and window state.
 */
protected void enableMenuItems() {
    int numSelected = countSelectedFields();
    boolean someFieldSelected = numSelected > 0;
    boolean multipleFieldsSelected = numSelected > 1;
    FieldWidget first = someFieldSelected ? (FieldWidget)selectedFields.get(0)
	: null;

    // Edit menu
    Section s = (first == null) ? null : first.getSectionWidget().getSection();

    cutItem.setEnabled(someFieldSelected);
    delSelectionItem.setEnabled(someFieldSelected);
    delGroupItem.setEnabled(someFieldSelected && report.isInsideGroup(s));
    delSectionItem.setEnabled(someFieldSelected && !report.isOneOfAKind(s));

    // Insert menu
    if (numSelected == 1) {	// One field is selected
	// Only enable running subtotals if the selected field is a field
	// for which subtotals make sense.
	subtotItem.setEnabled(SubtotalField.canSubtotal(first.getField()));
    }
    else
	subtotItem.setEnabled(false);
    groupItem.setEnabled(report.hasFields());
    sectItem.setEnabled(someFieldSelected);

    // Format menu
    formatItem.setEnabled(someFieldSelected);
    borderItem.setEnabled(someFieldSelected);
    boundsItem.setEnabled(someFieldSelected);
    alignSubmenu.setEnabled(multipleFieldsSelected);
    sizeSubmenu.setEnabled(multipleFieldsSelected);
}

/**
 * Handles user actions.
 */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();

    // File menu
    if ("new".equals(cmd)) newReport();
    else if ("open".equals(cmd)) openReport();
    else if ("save".equals(cmd)) saveReport();
    else if ("save_as".equals(cmd)) saveReportAs();
    else if ("export".equals(cmd)) exportReport();
    else if ("close".equals(cmd)) maybeClose();
    else if ("quit".equals(cmd)) maybeQuit();

    // Edit menu
    else if ("cut".equals(cmd)) deleteSelectedFields();
    else if ("del_selection".equals(cmd)) deleteSelectedFields();
    else if ("del_group".equals(cmd)) {
	FieldWidget fw = (FieldWidget)selectedFields.get(0);
	deleteGroupContaining(fw.getSectionWidget().getSection());
    }
    else if ("del_selection".equals(cmd)) {
	FieldWidget fw = (FieldWidget)selectedFields.get(0);
	deleteSection(fw.getSectionWidget().getSection());
    }

    // Insert menu
    else if ("database_field".equals(cmd))
	openFieldPickerWin(FieldPickerWin.REPORT_DATABASE_FIELDS);
    else if ("text".equals(cmd)) placeNewTextField();
    else if ("formula_field".equals(cmd))
	openFieldPickerWin(FieldPickerWin.FORMULAS);
    else if ("parameter_field".equals(cmd))
	openFieldPickerWin(FieldPickerWin.PARAMETERS);
    else if ("running_total_field".equals(cmd)) openSubtotalWin();
    else if ("special_field".equals(cmd))
	openFieldPickerWin(FieldPickerWin.SPECIAL_FIELDS);
    else if ("add_group".equals(cmd)) openNewGroupWin();
    else if ("ins_section".equals(cmd)) insertSection();

    // Format menu
    else if ("format_field".equals(cmd)) openFormatWin(0);
    else if ("border".equals(cmd)) openFormatWin(1);
    else if ("position".equals(cmd)) openBoundsWin();

    // Format/Align submenu
    else if ("align_top".equals(cmd)) align(ALIGN_TOP);
    else if ("align_middle".equals(cmd)) align(ALIGN_MIDDLE);
    else if ("align_bottom".equals(cmd)) align(ALIGN_BOTTOM);
    else if ("align_left".equals(cmd)) align(ALIGN_LEFT);
    else if ("align_center".equals(cmd)) align(ALIGN_CENTER);
    else if ("align_right".equals(cmd)) align(ALIGN_RIGHT);
    else if ("snap_to_grid".equals(cmd)) align(ALIGN_SNAP_TO_GRID);

    // Format/Size submenu
    else if ("size_width".equals(cmd)) size(SIZE_SAME_WIDTH);
    else if ("size_height".equals(cmd)) size(SIZE_SAME_HEIGHT);
    else if ("size_size".equals(cmd)) size(SIZE_SAME_SIZE);

    // Database menu
    else if ("visual_table_linker".equals(cmd)) openVisTableWin();
    else if ("sql_query_text".equals(cmd)) showSQL();
    else if ("db_connection".equals(cmd)) openDbConnWin(false);

    // Report menu
    else if ("run".equals(cmd)) runReport();
    else if ("select_records".equals(cmd)) openWhereClauseEditor();
    else if ("sort_by".equals(cmd)) openSortWin();
    else if ("group_by".equals(cmd)) openGroupWin();
    else if ("summary".equals(cmd)) openDescripWin();

    // Help menu
    else if ("help".equals(cmd)) help();
    else if ("about".equals(cmd)) about();
}

JFileChooser getChooser() {
    if (chooser == null)
	chooser = new JFileChooser();
    return chooser;
}

/**
 * Creates a new report in a new design window.
 */
protected void newReport() {
    new DesignWin(null);	// Open a new window
}

/**
 * Opens an existing report in a new design window.
 */
protected void openReport() {
    int returnVal = getChooser().showOpenDialog(this);
    if (returnVal == JFileChooser.APPROVE_OPTION)
	new DesignWin(getChooser().getSelectedFile().getPath());
}

/**
 * Saves the current report.
 */
protected void saveReport() {
    if (reportFileName == null)
	saveReportAs();
    else
	writeReportFile(reportFileName);
}

/**
 * Saves the current report in a different file.
 */
protected void saveReportAs() {
    int returnVal = getChooser().showSaveDialog(this);
    if (returnVal == JFileChooser.APPROVE_OPTION) {
	reportFileName = getChooser().getSelectedFile().getPath();
	writeReportFile(reportFileName);
    }
}

/**
 * Writes the current report to the specified file.
 *
 * @param fileName a file name
 */
protected void writeReportFile(String fileName) {
    report.writeFile(fileName);
    report.setDirty(false);
}

/**
 * Exports the report output using one of the layout engines.
 */
protected void exportReport() {
    new ExportWin(this, report);
}

/**
 * Saves the report if it is dirty and closes the current design window.
 * If there are no more open design windows, exist the application.
 */
protected void maybeClose() {
    if (report.isDirty()) {
	String str = "This report has been modified. Save it?";
	switch (JOptionPane.showConfirmDialog(this, str)) {
	case JOptionPane.YES_OPTION:
	    saveReport();
	case JOptionPane.NO_OPTION:
	    break;
	case JOptionPane.CANCEL_OPTION:
	    return;		// Don't close window
	}
    }
    dispose();
    DesignWin.deleteWindow(this);
}

/**
 * Runs and displays the report.
 */
protected void runReport() {
    new ReportRunner(report).run(new SwingLE());
}

/**
 * Returns the number of selected fields.
 *
 * @return the number of selected fields
 */
int countSelectedFields() {
    return selectedFields.size();
}

/**
 * Hands each field widget to the specified <code>FieldWidgetWalker</code>.
 *
 * @param perambulator a field walker
 */
void withWidgetsDo(FieldWidgetWalker perambulator) {
    for (Iterator iter = sectionWidgets.iterator(); iter.hasNext(); ) {
	SectionWidget sw = (SectionWidget)iter.next();
	Component[] kids = sw.fieldPanel.getComponents();
	for (int i = 0; i < kids.length; ++i)
	    if (kids[i] instanceof FieldWidget)
		perambulator.step((FieldWidget)kids[i]);
    }
}

/**
 * Hands each selected field widget to the specified
 * <code>FieldWidgetWalker</code>.
 *
 * @param perambulator a field walker
 */
void withSelectedWidgetsDo(FieldWidgetWalker perambulator) {
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	perambulator.step((FieldWidget)iter.next());
}

/**
 * Deletes the group that contains the specified section.
 *
 * @param section a report section
 */
void deleteGroupContaining(Section section) {
    Group group = report.findGroup(section);
    if (group != null) {
	report.removeGroup(group);
	rebuildGroups();
    }
}

/**
 * Adds a new group.
 *
 * @param group the new group
 */
void addGroup(Group group) {
    report.addGroup(group);
    rebuildGroups();
}

/**
 * Inserts a new section below the section containing the first selected
 * field widget. Only called when at least one widget is selected.
 *
 * @param section a report section
 */
void insertSection() {
    Field f = firstSelectedField().getField();
    if (f != null)
	insertSectionBelow(f.getSection());
}

/**
 * Inserts a new section below the specified section.
 *
 * @param s a report section
 */
void insertSectionBelow(Section s) {
    Section newSection = report.insertSectionBelow(s);
    newSection.addObserver(this);
    SectionWidget sw = new SectionWidget(this, newSection, "", "");

    SectionWidget swBefore = findSectionWidgetFor(s);
    int insertIndex = sectionWidgets.indexOf(swBefore) + 1;
    sectionWidgets.add(insertIndex, sw);

    // Add to palette layer (one above default layer) because
    // we want to move field widgets below section widgets
    // temporarily at the end of a drag (see putDown()).
    sectionContainer.add(sw, JLayeredPane.PALETTE_LAYER, insertIndex);

    renameSectionWidgets();	// Muck with names (e.g., add " (a)")
}

/**
 * Deletes the specified section. Does nothing if this section is the only
 * one of its kind. (This method should only be called if the section is
 * not one-of-a-kind.)
 *
 * @param s a report section
 */
void deleteSection(Section s) {
    if (report.isOneOfAKind(s))	// Should not happen
	return;

    s.deleteObserver(this);
    report.removeSection(s);
    SectionWidget sw = findSectionWidgetFor(s);
    sectionContainer.remove(sw);
    renameSectionWidgets();
}

/**
 * Rebuilds the group sections and redisplays the report.
 */
void rebuildGroups() {
    deselectAll();
    remove(scroller);
    buildSections();
    pack();
}


/**
 * Creates and adds a new text field to the first section of the page header.
 * If there is no page header, does nothing.
 *
 * @param x where to place the title
 * @param width how wide it should be
 * @param title the string to display
 */
void addTitleField(int x, int width, String title) {
    Section s = (Section)report.pageHeaders().next();
    if (s == null)
	return;

    // Create the field.
    Field f = Field.create(null, s, "text", title);
    jimm.datavision.Rectangle b = f.getBounds();
    b.setBounds(x, 0, width, Field.DEFAULT_HEIGHT);

    // Make the field underlined and bold.
    Format fmt = new Format();
    fmt.setBold(true);
    fmt.setUnderline(true);
    f.setFormat(fmt);

    // Add the field to the report section.
    s.addField(f);

    // Create widget and add to section widget.
    TextFieldWidget tfw = new TextFieldWidget(null, f);
    tfw.moveToSection(findSectionWidgetFor(s));
}

/**
 * Aligns the selected fields to the first selected field (chronologically
 * speaking).
 *
 * @param which alignment constant.
 */
protected void align(int which) {
    if (sectionWidgets.isEmpty())
	return;

    // Align the widgets based on the position of the first selected field.
    FieldWidget first = firstSelectedField();
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	((FieldWidget)iter.next()).align(which, first);
}

/**
 * Resize the selected fields based on the first selected field
 * (chronologically speaking) and the speicified resize (width, height, both).
 *
 * @param which size constant
 */
protected void size(int which) {
    if (sectionWidgets.isEmpty())
	return;

    // Resize the widgets based on the position of the first selected field.
    FieldWidget first = firstSelectedField();
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	((FieldWidget)iter.next()).size(which, first);
}

/**
 * Returns first selected field widget (chronologically speaking). May
 * return <code>null</code> if no fields are selected
 *
 * @return field widget selected the earliest, or <code>null</code> if
 * no fields are selected
 */
FieldWidget firstSelectedField() {
    return selectedFields.isEmpty()
	? null : (FieldWidget)selectedFields.get(0);
}

/**
 * Selects a field widget, possibly deselecting all others everywhere.
 * Called from section widget.
 *
 * @param FieldWidget a field widget
 * @param makeSelected new selection state
 * @param deselectAll if <code>true</code>, all other fields in all sections
 * are deselected first
 */
void select(FieldWidget fieldWidget, boolean makeSelected,
	    boolean deselectOthers)
{
    if (deselectOthers)
	deselectAll();

    fieldWidget.doSelect(makeSelected);
    if (makeSelected) {
	if (!selectedFields.contains(fieldWidget)) // Don't add it twice
	    selectedFields.add(fieldWidget);
    }
    else {
	selectedFields.remove(fieldWidget);
    }

    enableMenuItems();
}

/**
 * Deselect all fields. Called from SectionWidget#deselectAll.
 *
 * @see SectionWidget#deselectAll
 */
protected void deselectAll() {
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	((FieldWidget)iter.next()).doSelect(false);
    selectedFields.clear();

    enableMenuItems();
}

/**
 * Delete selected fields.
 */
protected void deleteSelectedFields() {
    deleteSelectedFieldsAnd(null);
}

/**
 * Delete specified field and all selected fields.
 *
 * @param oneMore an additional field to delete; may be <code>null</code>
 */
protected void deleteSelectedFieldsAnd(FieldWidget oneMore) {
    HashSet affectedSections = new HashSet();

    // Delete fields from selectedFields before we iterate through them.
    // That way we won't try to operate on them elsewhere (for example,
    // when updating menu items in response to an update from a field or
    // section).
    ArrayList sfCopy = (ArrayList)selectedFields.clone();
    selectedFields.clear();

    for (Iterator iter = sfCopy.iterator(); iter.hasNext(); ) {
	FieldWidget fw = (FieldWidget)iter.next();
	fw.doDelete();		// Widget deletes itself and field from report
	affectedSections.add(fw.getSectionWidget());
    }

    if (oneMore != null && !sfCopy.contains(oneMore)) {
	oneMore.doDelete();
	affectedSections.add(oneMore.getSectionWidget());
    }

    for (Iterator iter = affectedSections.iterator(); iter.hasNext(); ) {
	SectionWidget sw = (SectionWidget)iter.next();
	sw.repaint();
    }

    enableMenuItems();
}

/**
 * Toggles the visibility of all selected fields. Called by a section widget.
 *
 * @see FieldWidget#doSetVisibility
 */
void setFieldVisibility(boolean newVisiblity) {
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	((FieldWidget)iter.next()).doSetVisibility(newVisiblity);
}

/**
 * Picks up the field widget because field dragging is starting.
 * Called from SectionWidget#pickUp.
 *
 * @param f field widget
 * @see SectionWidget#pickUp
 */
void pickUp(FieldWidget f) {
    Dimension size = sectionContainer.getBounds().getSize();
    sectionContainer.setPreferredSize(size);
    sectionContainer.setMinimumSize(size);

    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); ) {
	FieldWidget fw = (FieldWidget)iter.next();
	int y = fw.getSectionWidget().getBounds().y;

	// Remove from section view widget, but not section model in report.
	fw.getSectionWidget().removeField(fw);

	// Add to the drag layer of our section container.
	sectionContainer.add(fw, JLayeredPane.DRAG_LAYER);

	// Recalculate bounds relative to the section container.
	jimm.datavision.Rectangle b = fw.getField().getBounds();
	b.setBounds(b.x + SectionWidget.LHS_WIDTH, b.y + y, b.width, b.height);
    }
}

/**
 * Puts the dragged field widgets down inside the sections they are
 * floating above. Called from SectionWidget#putDown.
 *
 * @param f the field widget being dragged; all other selected fields
 * have been dragged along with it
 * @param origScreenPos the original location of the field in screen
 * coordinates
 * @param mouseScreenPos the current mouse position in screen coordinates;
 * note 
 * @see SectionWidget#putDown
 */
void putDown(FieldWidget f, java.awt.Point origScreenPos,
	     java.awt.Point mouseScreenPos)
{
    // Move all dragged fields under everything else so getComponentAt()
    // will not return this field.
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	sectionContainer.setLayer((FieldWidget)iter.next(),
				  JLayeredPane.DEFAULT_LAYER.intValue());

    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); ) {
	FieldWidget fw = (FieldWidget)iter.next();
	java.awt.Point pos = fw.getLocationOnScreen();

	// Move to new section. Each field may be dropped into a different
	// section.
	SectionWidget sw = getSectionWidgetAt(pos);
	if (sw == null)		// Field snaps back to orig pos if sw is null
	    fw.snapBack();
	else {
	    // Recalculate bounds
	    jimm.datavision.Rectangle b = fw.getField().getBounds();
	    b.setBounds(b.x - SectionWidget.LHS_WIDTH, b.y - sw.getBounds().y,
			b.width, b.height);

	    // Move model and view to new section
	    fw.moveToSection(sw);
	}
    }
}

/**
 * Returns the section widget under the mouse.
 *
 * @param fwScreenPos a widget position in screen coordinates
 * @return the SectionWidget under the mouse
 */
protected SectionWidget getSectionWidgetAt(java.awt.Point fwScreenPos) {
    // Translate fwScreenPos to sectionContainer coords
    java.awt.Point scScreenPos = sectionContainer.getLocationOnScreen();
    fwScreenPos.translate(-scScreenPos.x, -scScreenPos.y);

    // We've either landed on a section or a field in the section.
    // Crawl up the parent hierarchy until we find the section.
    Component sw = sectionContainer.getComponentAt(fwScreenPos);
    if (sw == null) return null;
    while (!(sw instanceof SectionWidget)) {
	sw = sw.getParent();
	if (sw == null || sw instanceof SectionNameLabel)
	    return null;
    }
    return (SectionWidget)sw;
}

/**
 * Drags the selected field widgets a few pixels. Called from section that
 * contains field being dragged.
 *
 * @param action a <code>FieldWidget.ACTION_*</code> constant
 * @param dx delta x
 * @param dy delta y
*/
protected void dragSelectedWidgets(int action, int dx, int dy) {
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	((FieldWidget)iter.next()).doDrag(action, dx, dy);
}

/**
 * Opens a new or existing field picker window.
 *
 * @param startingType the index of the starting type to display
 */
protected void openFieldPickerWin(int startingType) {
    new FieldPickerWin(this, report, startingType);
}

/**
 * Opens a new or existing field running subtotal window. We should
 * only get here if there is exactly one selected field.
 */
protected void openSubtotalWin() {
    new SubtotalsWin(this, (FieldWidget)selectedFields.get(0));
}

/**
 * Opens a new or existing new group window.
 */
protected void openNewGroupWin() {
    new NewGroupWin(this, report);
}

/**
 * Sets the flag that tells everyone else that the user wants to place
 * a new text field.
 */
protected void placeNewTextField() {
    placingNewTextField = true;
    setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
}

/**
 * Returns <code>true</code> if the user is trying to place a new text field.
 *
 * @return <code>true</code> if the user is trying to place a new text field
 */
boolean isPlacingNewTextField() {
    return placingNewTextField;
}

/**
 * The caller has accepted the new text field.
 */
void acceptNewTextField() {
    placingNewTextField = false;
    setCursor(null);
}

/**
 * The caller has rejected the new text field.
 */
void rejectNewTextField() {
    acceptNewTextField();
}

/**
 * Opens a new or existing field format window.
 *
 * @param whichTab the index of the tab to display when opened
 */
protected void openFormatWin(int whichTab) {
    FieldWidget fw = firstSelectedField();
    if (fw != null)
	new FormatWin(this, whichTab);
}

/**
 * Opens a new or existing bounds editor window.
 */
protected void openBoundsWin() {
    FieldWidget fw = firstSelectedField();
    if (fw != null)
	new BoundsWin(this);
}

/**
 * Opens a new or existing visible table joiner window.
 */
protected void openVisTableWin() {
    new VisTableWin(this, report);
}

/**
 * Opens a new or existing where clause editor.
 */
protected void openWhereClauseEditor() {
    new WhereClauseWin(this, report);
}

/**
 * Opens a new or existing sort order window.
 */
protected void openSortWin() {
    new SortWin(this, report);
}

/**
 * Opens a new or existing group order window.
 */
protected void openGroupWin() {
    new GroupWin(this, report);
}

/**
 * Opens a new or existing sort order window.
 */
protected void openDescripWin() {
    new DescripWin(this, report);
}

/**
 * Opens the help window.
 */
protected void help() {
    HelpWin helpWin = HelpWin.instance();
    helpWin.setState(Frame.NORMAL); // De-iconify. Why is this necessary?
    helpWin.setVisible(true);
    helpWin.toFront();
}

/**
 * Opens the about box.
 */
protected void about() {
    String msg = "DataVision v" + info.Version + "\n"
	+ "by Jim Menard (jimm@io.com).\n"
	+ info.URL + "\n\n"
	+ info.Copyright + ".\n\n"
	+ "This product includes MinML, the XML\n"
	+ "parser developed by John Wilson.";
    JOptionPane.showMessageDialog(null, msg, "About DataVision",
				  JOptionPane.PLAIN_MESSAGE);
}

/**
 * Opens a new or existing database connection info window.
 *
 * @param modal passed on to dialog constructor
 */
protected void openDbConnWin(boolean modal) {
    new DbConnWin(this, report, modal);
}

/**
 * Opens a new window containing the SQL query text.
 */
protected void showSQL() {
    new SQLQueryWin(this, report.getQuery().toString());

}

/**
 * Returns the section widget containing the specified section.
 *
 * @param s section
 * @return the section widget containing the section
 */
SectionWidget findSectionWidgetFor(Section s) {
    for (Iterator iter = sectionWidgets.iterator(); iter.hasNext(); ) {
	SectionWidget sw = (SectionWidget)iter.next();
	if (sw.section == s)
	    return sw;
    }
    return null;
}


/**
 * Snaps the rectangle to the grid.
 *
 * @param r a rectangle
 */
public void snapToGrid(jimm.datavision.Rectangle r) {
    int x = (int)r.x;
    int mod = x % GRID_SIZE;
    if (mod != 0) {
	if (mod <= GRID_SIZE / 2)
	    r.setX(x - mod);
	else
	    r.setX(x + GRID_SIZE - mod);
    }

    int y = (int)r.y;
    mod = y % GRID_SIZE;
    if (mod != 0) {
	if (mod < GRID_SIZE / 2)
	    r.setY(y - mod);
	else
	    r.setY(y + GRID_SIZE - mod);
    }
}

// Might be using this code for dragging to select multiple fields. It
// should eventually live in a separate file.

//  class SelectionHandler extends MouseInputAdapter {

//  protected java.awt.Rectangle rect;
//  protected HashSet fields;
//  protected JPanel selectionOutline;

//  public void mousePressed(MouseEvent e) {
//      int x = e.getX() - SectionWidget.LHS_WIDTH;
//      if (x < 0)
//  	return;

//      System.out.println("starting rect drag");
//      deselectAll();

//      int y = e.getY();
//      rect = new java.awt.Rectangle(x, y, 0, 0);
//      selectionOutline = new JPanel();
//      selectionOutline.setBounds(rect);
//      getContentPane().add(selectionOutline, 0); // Add to top of visual stack
//  }

//  public void mouseDragged(MouseEvent e) {
//      System.out.println("rect drag");
//      updateSize(e);
//  }

//  public void mouseReleased(MouseEvent e) {
//      System.out.println("rect drag released");
//      updateSize(e);
//      remove(selectionOutline);
//  }

//  /* 
//   * Updates the size of the current rectangle, changes selection list, and
//   * calls <code>repaint</code>. Because rect always has the same origin,
//   * translate it if the width or height is negative.
//   * 
//   * For efficiency, specify the painting region using arguments to the
//   * <code>repaint</code> call.
//   */
//  void updateSize(MouseEvent e) {
//      if (selectionOutline == null)
//  	return;

//      int x = e.getX();
//      int y = e.getY();
//      rect.setSize(x - rect.x, y - rect.y);
//      selectionOutline.setSize(x - rect.x, y - rect.y);

//  //      selectFieldsWithin(rect);

//  //      java.awt.Rectangle totalRepaint = rectToDraw.union(previousRectDrawn);
//  //      repaint(totalRepaint.x, totalRepaint.y,
//  //  	    totalRepaint.width, totalRepaint.height);
//  }

//  }

}
