package jimm.datavision.gui;
//import java.awt.Frame;
import java.awt.event.*;
import java.util.HashMap;
import java.util.Iterator;
import javax.swing.*;

/**
 * The abstract parent of all edit windows except the main design window.
 * Handles common behavior.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public abstract class EditWin extends JDialog implements ActionListener {

protected HashMap revertInfoList;
protected boolean saved;
protected JButton revertButton;

/**
 * Constructor for a non-modal dialog.
 *
 * @param owner the window to which this dialog belongs
 * @param whichTab the index of the tab to display when opened
 */
public EditWin(JFrame owner, String title) {
    this(owner, title, false);
}

/**
 * Constructor.
 *
 * @param designWin the window to which this dialog belongs
 * @param whichTab the index of the tab to display when opened
 * @param modal passed on to superclass
 */
public EditWin(JFrame owner, String title, boolean modal) {
    super(owner, title, modal);
    revertInfoList = new HashMap();
    saved = false;
}

/**
 * Builds and returns a panel containing the bottom four buttons.
 *
 * @return the four buttons Ok, Apply, Revert, and Cancel
 */
protected JPanel closeButtonPanel() {
    JPanel buttonPanel = new JPanel();
    JButton button;

    buttonPanel.add(button = new JButton("Ok"));
    button.addActionListener(this);
    button.setDefaultCapable(true);

    buttonPanel.add(button = new JButton("Apply"));
    button.addActionListener(this);

    buttonPanel.add(revertButton = new JButton("Revert"));
    revertButton.addActionListener(this);
    revertButton.setEnabled(false);

    buttonPanel.add(button = new JButton("Cancel"));
    button.addActionListener(this);

    return buttonPanel;
}

/**
 * Handles the four buttons.
 *
 * @param e action event
 */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();
    if ("Ok".equals(cmd)) {
	save(false);
	dispose();
    }
    else if ("Apply".equals(cmd)) {
	save(true);
    }
    else if ("Revert".equals(cmd)) {
	revert();
    }
    else if ("Cancel".equals(cmd)) {
	revert();
	dispose();
    }
}

/**
 * Saves all data. If needed, save revert info.
 *
 * @param saveRevertInfo passed through to <code>doSave</code>
 * @see #doSave
 */
protected void save(boolean saveRevertInfo) {
    doSave(saveRevertInfo);
    revertButton.setEnabled(true);
    saved = true;
}

/**
 * Saves all data, rembering original values if required.
 *
 * @param saveRevertInfo if <code>true</code>, must save reversion info
 */
protected abstract void doSave(boolean saveRevertInfo);

/**
 * Reverts all state information. Walks <code>revertInfoList</code>, calling
 * <code>doRevert</code> for each one.
 *
 * @see #doRevert
 */
protected void revert() {
    if (saved && !revertInfoList.isEmpty()) {
	for (Iterator iter = revertInfoList.keySet().iterator();
	     iter.hasNext();)
	{
	    doRevert(revertInfoList.get(iter.next()));
	}
    }
    revertButton.setEnabled(false);
    saved = false;
}

/**
 * Reverts all data, including dirty flag states, for a single object. This
 * method is called by <code>revert</code> once for each revert info object
 * in <code>revertInfoList</code>.
 *
 * @param revertInfo the revert info to use; may be <code>null</code>
 * @see #revert
 */
protected abstract void doRevert(Object revertInfo);

}
