package jimm.datavision.gui;
import jimm.datavision.Report;
import jimm.datavision.ErrorHandler;
import jimm.datavision.LayoutEngine;
import jimm.datavision.IndentWriter;
import jimm.datavision.layout.*;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.awt.BorderLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.*;

/**
 * This dialog lets the user export report results using one of the
 * layout engines.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class ExportWin extends JDialog implements ActionListener {

protected static final String[] LAYOUT_NAMES = {
    "Comma-Separated",
    "Tab-Separated",
    "DocBook",
    "HTML",
    "LaTeX",
    "XML"
};

protected Report report;
protected DesignWin designWin;
protected JComboBox combo;

/**
 * Constructor.
 *
 * @param designWin the window to which this dialog belongs
 * @param initialText the initial text to edit
 * @param title the window title
 */
public ExportWin(DesignWin designWin, Report report) {
    super(designWin, "Export Report");
    this.report = report;
    this.designWin = designWin;
    buildWindow();
    pack();
    setVisible(true);
}

/**
 * Builds the window contents.
 *
 * @param whichTab the index of the tab to display when opened
 */
protected void buildWindow() {
    // Panel containing list of layout engines
    combo = new JComboBox(LAYOUT_NAMES);
    JPanel comboPanel = new JPanel();
    comboPanel.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
    comboPanel.add(combo);

    // Ok and Cancel Buttons
    JPanel buttonPanel = buildButtonPanel();

    // Add list, cards, and buttons to window
    getContentPane().setLayout(new BorderLayout());
    getContentPane().add(combo, BorderLayout.CENTER);
    getContentPane().add(buttonPanel, BorderLayout.SOUTH);
}

/**
 * Builds and returns a panel containing the OK and Cancel
 *
 * @return a panel
 */
protected JPanel buildButtonPanel() {
    JPanel buttonPanel = new JPanel();
    JButton button;

    buttonPanel.add(button = new JButton("Ok"));
    button.addActionListener(this);
    button.setDefaultCapable(true);

    buttonPanel.add(button = new JButton("Cancel"));
    button.addActionListener(this);

    return buttonPanel;
}

/**
 * Handles the OK and Cancel buttons.
 *
 * @param e action event
 */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();
    if ("Ok".equals(cmd)) {
	String choice = (String)combo.getSelectedItem();
	dispose();
	runReport(choice);
    }
    else if ("Cancel".equals(cmd)) {
	dispose();
    }
}


/**
 * Given the user's choice of layout engine, ask user for output file
 * and run the report.
 *
 * @param choice the combo box string that the user selected
 */
protected void runReport(String choice) {
    String extension = null;
    if ("Comma-Separated".equals(choice))
	extension = "";
    else if ("Tab-Separated".equals(choice))
	extension = "";
    else if ("DocBook".equals(choice))
	extension = "";
    else if ("HTML".equals(choice))
	extension = "";
    else if ("LaTeX".equals(choice))
	extension = "";
    else if ("XML".equals(choice))
	extension = "";

    PrintWriter out = openOutput(extension);
    if (out == null)
	return;

    LayoutEngine le = null;
    if ("Comma-Separated".equals(choice))
	le = new CharSepLE(out, ',');
    else if ("Tab-Separated".equals(choice))
	le = new CharSepLE(out, '\t');
    else if ("DocBook".equals(choice))
	le = new DocBookLE(out);
    else if ("HTML".equals(choice))
	le = new HTMLLE(out);
    else if ("LaTeX".equals(choice))
	le = new LaTeXLE(out);
    else if ("XML".equals(choice))
	le = new XMLLE(new IndentWriter(out));

    new ReportRunner(report).run(le);
    out.flush();
    out.close();
}

/**
 * Lets user select output file. I'd like to be able to give a default
 * file extension (thus the argument), but that will have to come later.
 *
 * @param extension default file name extension (unused)
 */
protected PrintWriter openOutput(String extension) {
    JFileChooser chooser = designWin.getChooser();
    int returnVal = chooser.showSaveDialog(this.getOwner());
    if (returnVal == JFileChooser.APPROVE_OPTION) {
	String path = chooser.getSelectedFile().getPath();
	PrintWriter out = null;
	try {
	    out = new PrintWriter(new FileWriter(path));
	}
	catch (IOException e) {
	    out = null;
	    ErrorHandler.error("There was an error while trying\n"
			       + " to open the output file.\n",
			       e, "Error selecting output");
	}
	finally {
	    return out;
	}
    }
    else
	return null;
}

}
