package jimm.datavision.gui;
import jimm.datavision.*;
import jimm.datavision.sql.*;
import jimm.datavision.gui.parameter.ParamEditWin;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.*;

/**
 * A window that lets the user drag any field available onto the report.
 * Uses a field picker tree.
 * <p>
 * The classes used to store information about leaf nodes are in
 * FPLeafInfo.java.
 *
 * @see FieldPickerTree
 * @see FPLeafInfo
 * @see ColumnInfo
 * @see FormulaInfo
 * @see ParameterInfo
 * @see SpecialInfo
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
class FieldPickerWin
    extends JDialog
    implements ActionListener, TreeSelectionListener, Observer
{

public static final int REPORT_DATABASE_FIELDS = 0;
public static final int FORMULAS = 1;
public static final int PARAMETERS = 2;
public static final int SPECIAL_FIELDS = 3;
public static final int ALL_DATABASE_FIELDS = 4;
//  public static final int SQL_EXPRESSION_FIELDS = XXXX;
//  public static final int RUNNING_TOTAL_FIELDS = XXXX;
//  public static final int GROUP_NAME_FIELDS = XXXX;

protected Report report;
protected JMenuItem cutItem, editFormulaItem, renameFormulaItem,
    editParameterItem, renameParameterItem;
protected FieldPickerTree tree;
protected DefaultTreeModel treeModel;
protected DefaultMutableTreeNode formulaCategoryNode;
protected DefaultMutableTreeNode parameterCategoryNode;
protected FPLeafInfo selectedInfo;

/**
 * Constructor.
 *
 * @param frame the window to which this dialog belongs
 * @param report the report
 * @param startingType the index of the starting type to display
 */
FieldPickerWin(JFrame frame, Report report, int startingType) {
    super(frame, "Fields");
    this.report = report;
    buildWindow(startingType);

    pack();
    show();
}

public void update(Observable o, Object arg) {
    // Primitive and overkill, but it works.
    tree.repaint();
}

/**
 * Builds the contents of the window.
 *
 * @param startingType the index of the starting type to display
 */
protected void buildWindow(int startingType) {
    buildMenuBar();

    DefaultMutableTreeNode top = new DefaultMutableTreeNode();
    createNodes(top);
    treeModel = new DefaultTreeModel(top);

    tree = new FieldPickerTree(treeModel);
    tree.setRootVisible(false);
    tree.getSelectionModel().setSelectionMode
	(TreeSelectionModel.SINGLE_TREE_SELECTION);

    // Make specified starting type visible and expanded
    tree.expandRow(startingType);

    // Listen for selection changes
    tree.addTreeSelectionListener(this);

    // Create the scroll pane and add the tree to it. 
    JScrollPane treeView = new JScrollPane(tree);
    treeView.setMinimumSize(new Dimension(100, 100));
    treeView.setPreferredSize(new Dimension(200, 300));

    getContentPane().add(treeView, BorderLayout.CENTER);
}

/**
 * Builds the window menu bar.
 */
protected void buildMenuBar() {
    JMenuBar bar = new JMenuBar();
    bar.add(buildFileMenu());
    bar.add(buildEditMenu());
    bar.add(buildFieldMenu());
    setJMenuBar(bar);
}

/**
 * Builds and returns the "File" menu.
 *
 * @return a menu
 */
protected JMenu buildFileMenu() {
    JMenu menu = new JMenu("File");
    menu.setMnemonic(KeyEvent.VK_F);

    MenuUtils.addToMenu(this, menu, "Close", "close", KeyEvent.VK_C,
			KeyEvent.VK_W);

    return menu;
}

/**
 * Builds and returns the "Edit" menu.
 *
 * @return a menu
 */
protected JMenu buildEditMenu() {
    JMenu menu = new JMenu("Edit");
    menu.setMnemonic(KeyEvent.VK_E);

    MenuUtils.addToMenu(this, menu, "Undo", "undo", KeyEvent.VK_U,
			KeyEvent.VK_Z)
	.setEnabled(false);
    menu.addSeparator();
    cutItem = MenuUtils.addToMenu(this, menu, "Cut", "cut", KeyEvent.VK_C,
				  KeyEvent.VK_X);
    cutItem.setEnabled(false);
    MenuUtils.addToMenu(this, menu, "Copy", "copy", KeyEvent.VK_O,
			KeyEvent.VK_C)
	.setEnabled(false);
    MenuUtils.addToMenu(this, menu, "Paste", "paste", KeyEvent.VK_P,
			KeyEvent.VK_V)
	.setEnabled(false);

    return menu;
}

/**
 * Builds and returns the "Field" menu.
 *
 * @return a menu
 */
protected JMenu buildFieldMenu() {
    JMenu menu = new JMenu("Field");
    menu.setMnemonic(KeyEvent.VK_I);

    MenuUtils.addToMenu(this, menu, "New Formula...", "new_formula",
			KeyEvent.VK_N, KeyEvent.VK_N);
    editFormulaItem = MenuUtils.addToMenu(this, menu, "Edit Formula...",
					  "edit_formula", KeyEvent.VK_E,
					  KeyEvent.VK_E);
    renameFormulaItem = MenuUtils.addToMenu(this, menu, "Rename Formula...",
					  "rename_formula", KeyEvent.VK_R,
					  KeyEvent.VK_R);
    editFormulaItem.setEnabled(false);
    renameFormulaItem.setEnabled(false);

    menu.addSeparator();

    MenuUtils.addToMenu(this, menu, "New Parameter...", "new_parameter",
			KeyEvent.VK_P, KeyEvent.VK_P);
    editParameterItem = MenuUtils.addToMenu(this, menu, "Edit Parameter...",
					  "edit_parameter", KeyEvent.VK_D,
					  KeyEvent.VK_D);
    renameParameterItem = MenuUtils.addToMenu(this, menu,
					      "Rename Parameter...",
					      "rename_parameter",
					      KeyEvent.VK_A, KeyEvent.VK_A);
    editParameterItem.setEnabled(false);
    renameParameterItem.setEnabled(false);

    return menu;
}

/**
 * Creates tree nodes.
 *
 * @param top top-level tree node
 */
protected void createNodes(DefaultMutableTreeNode top) {
    createUsedDatabaseTables(top);
    createFormulas(top);
//      createSQLExpressionFields(top);
    createParameters(top);
//      createSubtotalFields(top);
//      createGroupNameFields(top);
    createSpecialFields(top);
    createAllDatabaseTables(top);
}

/**
 * Creates nodes representing tables and columns for columns used by report.
 *
 * @param top top-level tree node
 */
protected void createUsedDatabaseTables(DefaultMutableTreeNode top) {
    DefaultMutableTreeNode categoryNode =
	new DefaultMutableTreeNode("Database Fields");
    top.add(categoryNode);

    // Build list of tables actually used by the report.
    final Set tables = new HashSet();
    report.withFieldsDo(new FieldWalker() {
	public void step(Field f) {
	    if (f instanceof ColumnField) {
		Column col = ((ColumnField)f).getColumn();
		tables.add(col.getTable());
	    }
	}
	});

    for (Iterator iter = tables.iterator(); iter.hasNext(); ) {
	Table table = (Table)iter.next();
	DefaultMutableTreeNode tableNode =
	    new DefaultMutableTreeNode(table.getName());
	categoryNode.add(tableNode);
	for (Iterator c_iter = table.columns(); c_iter.hasNext(); ) {
	    Column column = (Column)c_iter.next();
	    ColumnInfo info = new ColumnInfo(column, (JFrame)this.getOwner());
	    tableNode.add(new DefaultMutableTreeNode(info, false));
	}
    }
}

/**
 * Creates nodes representing formula fields.
 *
 * @param top top-level tree node
 */
protected void createFormulas(DefaultMutableTreeNode top) {
    formulaCategoryNode = new DefaultMutableTreeNode("Formulas");
    top.add(formulaCategoryNode);

    for (Iterator iter = report.formulas(); iter.hasNext(); ) {
	Formula f = (Formula)iter.next();
	FormulaInfo info = new FormulaInfo(report, f, (JFrame)this.getOwner());
	formulaCategoryNode.add(new DefaultMutableTreeNode(info));
	f.addObserver(this);
    }
}

//  /**
//   * Creates nodes representing SQL expressions.
//   *
//   * @param top top-level tree node
//   */
//  protected void createSQLExpressionFields(DefaultMutableTreeNode top) {
//      DefaultMutableTreeNode categoryNode =
//  	new DefaultMutableTreeNode("SQL Expression Fields");
//      top.add(categoryNode);
//  }

/**
 * Creates nodes representing parameter fields.
 *
 * @param top top-level tree node
 */
protected void createParameters(DefaultMutableTreeNode top) {
    parameterCategoryNode = new DefaultMutableTreeNode("Parameters");
    top.add(parameterCategoryNode);

    for (Iterator iter = report.parameters(); iter.hasNext(); ) {
	Parameter p = (Parameter)iter.next();
	ParameterInfo info = new ParameterInfo(report, p,
					       (JFrame)this.getOwner());
	parameterCategoryNode.add(new DefaultMutableTreeNode(info));
	p.addObserver(this);
    }
}

//  /**
//   * Creates nodes representing subtotal fields.
//   *
//   * @param top top-level tree node
//   */
//  protected void createSubtotalFields(DefaultMutableTreeNode top) {
//      final DefaultMutableTreeNode categoryNode =
//  	new DefaultMutableTreeNode("Running Total Fields");
//      top.add(categoryNode);

//      report.withFieldsDo(new FieldWalker() {
//  	public void step(Field f) {
//  	    if (f instanceof SubtotalField) {
//  		SubtotalInfo sub = new SubtotalInfo((SubtotalField)f,
//  						    (JFrame)this.getOwner());
//  		categoryNode.add(new DefaultMutableTreeNode(sub));
//  	    }
//  	}
//  	});
//  }

//  /**
//   * Creates node represting group name fields.
//   *
//   * @param top top-level tree node
//   */
//  protected void createGroupNameFields(DefaultMutableTreeNode top) {
//      DefaultMutableTreeNode categoryNode =
//  	new DefaultMutableTreeNode("Group Name Fields");
//      top.add(categoryNode);

//  }

/**
 * Creates nodes representing each possible special field.
 *
 * @param top top-level tree node
 */
protected void createSpecialFields(DefaultMutableTreeNode top) {
    DefaultMutableTreeNode categoryNode =
	new DefaultMutableTreeNode("Special Fields");
    top.add(categoryNode);

    HashMap strs = SpecialField.specialFieldNames();
    for (Iterator iter = strs.keySet().iterator(); iter.hasNext(); ) {
	String key = (String)iter.next();
	String val = (String)strs.get(key);
	key = SpecialField.TYPE_STRING + ':' + key;
	SpecialInfo info = new SpecialInfo(val, key, (JFrame)this.getOwner());
					   
	DefaultMutableTreeNode node = new DefaultMutableTreeNode(info);
	categoryNode.add(node);
    }
}

/**
 * Creates notes representing tables and columns for all tables in the
 * database.
 *
 * @param top top-level tree node
 */
protected void createAllDatabaseTables(DefaultMutableTreeNode top) {
    DefaultMutableTreeNode categoryNode =
	new DefaultMutableTreeNode("All Database Fields");
    top.add(categoryNode);

    for (Iterator iter = report.getDatabase().tables(); iter.hasNext(); ) {
	Table table = (Table)iter.next();
	DefaultMutableTreeNode tableNode =
	    new DefaultMutableTreeNode(table.getName());
	categoryNode.add(tableNode);
	for (Iterator c_iter = table.columns(); c_iter.hasNext(); ) {
	    Column column = (Column)c_iter.next();
	    ColumnInfo info = new ColumnInfo(column, (JFrame)this.getOwner());
	    tableNode.add(new DefaultMutableTreeNode(info));
	}
    }
}

/**
 * Deletes the currently selected field from the report and this window.
 * This method is only called if the field is legally deletable (for
 * example, it is not used in the report).
 */
protected void deleteField() {
    Object obj = selectedInfo.getLeaf();
    if (obj instanceof Field)
	report.removeField((Field)obj);
    else if (obj instanceof Formula)
	report.removeFormula((Formula)obj);
    else if (obj instanceof Parameter)
	report.removeParameter((Parameter)obj);

    tree.removeCurrentNode();
}

/**
 * Creates a new formula, adds it to the report and the tree, and opens
 * the formula editor.
 */
protected void newFormula() {
    // Create formula and let user enter name
    Formula f = new Formula(null, report, "", "");
    if (renameFormula(f) == false)
	return;			// User cancelled

    // Add to report
    report.addFormula(f);

    // Add to tree
    FormulaInfo info = new FormulaInfo(report, f, (JFrame)this.getOwner());
    DefaultMutableTreeNode node = new DefaultMutableTreeNode(info);
    treeModel.insertNodeInto(node, formulaCategoryNode,
			     formulaCategoryNode.getChildCount());

    // Make formula visible and make it the current selection
    TreePath path = new TreePath(node.getPath());
    tree.scrollPathToVisible(path);
    tree.setSelectionPath(path);

    // Open editor
    editFormula();
}

/**
 * Opens a name editor to rename the currently select formula. Returns
 * <code>true</code> if the user clicked OK, <code>false</code> if the
 * user clicked Cancel.
 *
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean renameFormula() {
    return renameFormula((Formula)selectedInfo.getLeaf());
}

/**
 * Opens a name editor to (re)name a formula. Returns <code>true</code>
 * if the user clicked OK, <code>false</code> if the user clicked Cancel.
 *
 * @param f a formula
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean renameFormula(Formula f) {
    String name = f.getName();
    String title = (name == null || name.length() == 0)
	? "New Formula Name" : "Edit Formula Name";
    name = new AskStringDialog((JFrame)this.getOwner(), title, "Formula Name",
			       name)
	.getString();

    if (name == null)		// User cancelled
	return false;

    if (name.length() == 0)
	name = "Unnamed Formula";

    f.setName(name);
    return true;
}

/**
 * Opens a formula editor on the currently selected formula.
 */
protected void editFormula() {
    selectedInfo.openEditor();
}

/**
 * Creates a new parameter, adds it to the report and the tree, and opens
 * the parameter editor.
 */
protected void newParameter() {
    // Create parameter and let user edit it.
    Parameter p = new Parameter(null, report);
    if (renameParameter(p) == false)
	return;			// User cancelled

    // Add to report
    report.addParameter(p);

    // Add to tree
    ParameterInfo info = new ParameterInfo(report, p, (JFrame)this.getOwner());
    DefaultMutableTreeNode node = new DefaultMutableTreeNode(info);
    treeModel.insertNodeInto(node, parameterCategoryNode,
			     parameterCategoryNode.getChildCount());

    // Make parameter visible and make it the current selection
    TreePath path = new TreePath(node.getPath());
    tree.scrollPathToVisible(path);
    tree.setSelectionPath(path);

    // Open editor
    editParameter();
}

/**
 * Opens a name editor to rename the currently select parameter. Returns
 * <code>true</code> if the user clicked OK, <code>false</code> if the
 * user clicked Cancel.
 *
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean renameParameter() {
    return renameParameter((Parameter)selectedInfo.getLeaf());
}

/**
 * Opens a name editor to (re)name a parameter. Returns <code>true</code>
 * if the user clicked OK, <code>false</code> if the user clicked Cancel.
 *
 * @param f a parameter
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean renameParameter(Parameter p) {
    String name = p.getName();
    String title = (name == null || name.length() == 0)
	? "New Parameter Name" : "Edit Parameter Name";
    name = new AskStringDialog((JFrame)this.getOwner(), title,
			       "Parameter Name", name)
	.getString();

    if (name == null)		// User cancelled
	return false;

    if (name.length() == 0)
	name = "Unnamed Parameter";

    p.setName(name);
    return true;
}

/**
 * Opens a parameter editor on the currently selected parameter.
 */
protected void editParameter() {
    selectedInfo.openEditor();
}

/**
 * Handles user actions. Actions are only allowed when legal. For example,
 * the "Cut" menu item will only be enabled when a delete operation is
 * possible.
 */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();

    // File menu
    if ("close".equals(cmd)) dispose();

    // Edit menu
    else if ("cut".equals(cmd)) deleteField();

    // Field menu
    else if ("new_formula".equals(cmd)) newFormula();
    else if ("edit_formula".equals(cmd)) editFormula();
    else if ("rename_formula".equals(cmd)) renameFormula();
    else if ("new_parameter".equals(cmd)) newParameter();
    else if ("edit_parameter".equals(cmd)) editParameter();
    else if ("rename_parameter".equals(cmd)) renameParameter();
}

/**
 * Modifies the menu in response to a change in the tree's selection.
 */
public void valueChanged(TreeSelectionEvent e) {
    if (e.isAddedPath()) {
	DefaultMutableTreeNode typeNode
	    = (DefaultMutableTreeNode)e.getPath().getLastPathComponent();
	Object obj = typeNode.getUserObject();
	if (obj instanceof FPLeafInfo)
	    selectedInfo = (FPLeafInfo)obj;
	else
	    selectedInfo = null;

	if (selectedInfo == null) {
	    editFormulaItem.setEnabled(false);
	    renameFormulaItem.setEnabled(false);
	    editParameterItem.setEnabled(false);
	    renameParameterItem.setEnabled(false);
	    cutItem.setEnabled(false);
	}
	else {
	    boolean isFormula = selectedInfo instanceof FormulaInfo;
	    boolean isParameter = selectedInfo instanceof ParameterInfo;
	    editFormulaItem.setEnabled(isFormula);
	    renameFormulaItem.setEnabled(isFormula);
	    editParameterItem.setEnabled(isParameter);
	    renameParameterItem.setEnabled(isParameter);
	    cutItem.setEnabled(selectedInfo.isDeletable());
	}

    }
    else {
	selectedInfo = null;
	cutItem.setEnabled(false);
	editFormulaItem.setEnabled(false);
	renameFormulaItem.setEnabled(false);
	editParameterItem.setEnabled(false);
	renameParameterItem.setEnabled(false);
    }
}

}
