package jimm.datavision.gui;
import jimm.datavision.Format;
import jimm.datavision.Field;
import jimm.datavision.Border;
import jimm.datavision.BorderEdge;
import java.awt.Insets;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.HashMap;
import javax.swing.*;

/**
 * A field format editing dialog box. There are tabs for text format and
 * borders. The initial format and border values are retrieved from one of
 * two places: either a field passed in to the constructor or the first
 * selected field in the desing window. The format and border are applied to
 * either a single field or all selected fields.
 * <p>
 * <i>Warning</i>: this code depends upon the fact that the strings in
 * <code>EDGE_COUNT_CHOICES</code> equals the integer value of the string
 * (zero, one, two, etc.) and that the strings in
 * <code>EDGE_STYLE_CHOICES</code> correspond to the numeric values of the
 * <code>BorderEdge.STYLE_*</code> constants.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class FormatWin extends EditWin implements FieldWidgetWalker {

// ================================================================
class RevertInfo {

FieldWidget fw;
boolean fieldDirty;
Format format;
Border border;

RevertInfo(FieldWidget widget, Format fmt, Border b)
{
    fw = widget;
    fieldDirty = fw.getField().isDirty();

    if (fmt == null)
	format = null;
    else {
	format = (Format)fmt.clone();
	format.setDirty(fmt.isDirty());
    }

    if (b == null)
	border = null;
    else {
	border = (Border)b.clone();
	border.setDirty(b.isDirty());
    }
}
}
// ================================================================

// ================================================================
/**
 * Holds a border edge and the widgets used for editing it.
 */
class EdgeWidgets {

protected BorderEdge edge;
protected int index;
protected JComboBox numberComboBox;
protected JComboBox styleComboBox;
protected JTextArea thicknessText;

EdgeWidgets(BorderEdge e, int index) {
    edge = e;
    this.index = index;
}

String getName() { return EDGE_NAMES[index]; }

}
// ================================================================

/**
 * List of pre-approved font size choices. Don't use this directly; instead
 * call <code>sizeChoices</code>.
 * @see #sizeChoices
 */
protected static Integer[] SIZE_CHOICES;

/**
 * List of field text alignment choices. In <code>addAlignment</code>,
 * we assume format align value == index in ALIGN_CHOICES.
 * @see #addAlignment
 */
protected static final String[] ALIGN_CHOICES = { "Left", "Center", "Right" };

protected static final int TOP = 0;
protected static final int LEFT = 1;
protected static final int RIGHT = 2;
protected static final int BOTTOM = 3;

/**
 * List of edge names.
 */
protected static final String[] EDGE_NAMES = {
    "Top", "Left", "Right", "Bottom"
};

/**
 * List of border edge count choices.
 * @see #buildBorderEdge
 */
protected static final String[] EDGE_COUNT_CHOICES = {
    "None", "One", "Two", "Three"
};

/**
 * List of border edge style choices.
 * @see #buildBorderEdge
 */
protected static final String[] EDGE_STYLE_CHOICES = {
    "Line", "Dashed", "Dotted"
};

protected DesignWin designWin;
protected FieldWidget fieldWidget;
protected Format format;
protected Border border;
protected boolean saveRevertInfo;
protected JComboBox size;
protected JCheckBox bold;
protected JCheckBox italic;
protected JCheckBox underline;
protected JCheckBox wrap;
protected JComboBox align;
protected JTextField formatText;
protected EdgeWidgets[] edgeWidgets;

/**
 * Constructor; uses format of first selected field.
 *
 * @param designWin the window to which this dialog belongs
 * @param whichTab the index of the tab to display when opened
 */
public FormatWin(DesignWin designWin, int whichTab) {
    this(designWin, designWin.firstSelectedField(), whichTab);
}

/**
 * Constructor.
 *
 * @param designWin the window to which this dialog belongs
 * @param fw a field widget from which we will take the format
 * @param whichTab the index of the tab to display when opened
 */
public FormatWin(DesignWin designWin, FieldWidget fw, int whichTab) {
    super(designWin, "Field Format");

    this.designWin = designWin;
    fieldWidget = fw;

    Field f = fieldWidget.getField();
    copyFormatAndBorder(f.getFormat(), f.getBorder());

    buildWindow(whichTab);
    pack();
    setVisible(true);
}

/**
 * Saves copies of format and border into the objects that we really edit.
 * Either may be <code>null</code>. Called from constructor and
 * <code>doRevert</code>.
 *
 * @param origFormat the format we are copying; not necessarily that of
 * the field
 * @param origBorder the border we are copying; not necessarily that of
 * the field
 * @see #doRevert
 */
protected void copyFormatAndBorder(Format origFormat, Border origBorder) {
    if (origFormat == null)
	format = new Format();
    else
	format = (Format)origFormat.clone();

    if (origBorder == null)
	border = new Border(fieldWidget.getField());
    else
	border = (Border)origBorder.clone();
    if (border.getTop() == null)
	border.setTop(new BorderEdge(BorderEdge.STYLE_LINE, 1, 0));
    if (border.getBottom() == null)
	border.setBottom(new BorderEdge(BorderEdge.STYLE_LINE, 1, 0));
    if (border.getLeft() == null)
	border.setLeft(new BorderEdge(BorderEdge.STYLE_LINE, 1, 0));
    if (border.getRight() == null)
	border.setRight(new BorderEdge(BorderEdge.STYLE_LINE, 1, 0));
}

/**
 * Builds the window contents.
 *
 * @param whichTab the index of the tab to display when opened
 */
protected void buildWindow(int whichTab) {
    JTabbedPane tabbedPane = new JTabbedPane();
    tabbedPane.addTab("Format", buildFormatTab());
    tabbedPane.addTab("Border", buildBorderTab());
    tabbedPane.setSelectedIndex(whichTab);

    // Ok, Apply, Revert, and Cancel Buttons
    JPanel buttonPanel = closeButtonPanel();

    // Add edit panes and buttons to window
    getContentPane().add(tabbedPane, BorderLayout.CENTER);
    getContentPane().add(buttonPanel, BorderLayout.SOUTH);

    fillFormatTab();
    fillBorderTab();
}

/**
 * Builds the format tab.
 */
protected JPanel buildFormatTab() {
    // Labels
    JPanel labelPanel = new JPanel();
    labelPanel.setLayout(new GridLayout(0, 1));
    JLabel label;
    labelPanel.add(label = new JLabel("Size:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Bold:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Italic:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Underline:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Wrap:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Align:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Format:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);

    // Edit values
    JPanel fieldPanel = new JPanel();
    fieldPanel.setLayout(new GridLayout(0, 1));

    addSize(fieldPanel);

    bold = addBool(fieldPanel, "Bold", KeyEvent.VK_B);
    italic = addBool(fieldPanel, "Italic", KeyEvent.VK_I);
    underline = addBool(fieldPanel, "Underline", KeyEvent.VK_U);
    wrap = addBool(fieldPanel, "Wrap", KeyEvent.VK_W);

    addAlignment(fieldPanel);
    addFormatString(fieldPanel);

    // Create a border layout panel. Put the labels in the center and the
    // fields in the east.
    JPanel panel = new JPanel();
    panel.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
    panel.setLayout(new BorderLayout());
    panel.add(labelPanel, BorderLayout.CENTER);
    panel.add(fieldPanel, BorderLayout.EAST);

    // We put a wrapper around that panel so everything is centered in the
    // tab.
    JPanel outerPanel = new JPanel();
    outerPanel.add(panel);
    return outerPanel;
}

/**
 * Adds a font size combo box to the specified panel.
 *
 * @param parent panel that should contain the size combo box
 */
protected void addSize(JPanel parent) {
    size = new JComboBox(sizeChoices());
    size.setEditable(true);
    parent.add(size);
}

/**
 * Returns the list of font size choices, lazily instantiating it if
 * necessary.
 *
 * @return an array of <code>Integer</code> size choices
 */
protected Integer[] sizeChoices() {
    if (SIZE_CHOICES == null) {
	ArrayList list = new ArrayList();

	for (int size = 6; size <= 12; ++size)
	    list.add(new Integer(size));
	for (int size = 14; size <= 36; size += 2)
	    list.add(new Integer(size));

	SIZE_CHOICES = new Integer[list.size()];
	list.toArray(SIZE_CHOICES);
    }
    return SIZE_CHOICES;
}

/**
 * Returns the index of the specified size value in the SIZE_CHOICES list.
 * Returns -1 if not found.
 *
 * @return an index, or -1 if not found
 */
protected int sizeIndexOf(int size) {
    Integer[] sizeChoices = sizeChoices();
    for (int i = 0; i < sizeChoices.length; ++i)
	if (sizeChoices[i].intValue() == size)
	    return i;
    return -1;
}

/**
 * Adds a boolean check box.
 *
 * @param parent the panel that will contain the check box
 * @param labelText the label of the check box
 * @param key shortcut (mnemonic) key
 */
protected JCheckBox addBool(JPanel parent, String labelText, int key)
{
    JCheckBox checkBox = new JCheckBox(labelText);
    checkBox.setMnemonic(key);
    parent.add(checkBox);
    return checkBox;
}

/**
 * Adds the alignment combo box.
 *
 * @param parent the panel that will contain the check box
 */
protected void addAlignment(JPanel parent) {
    align = new JComboBox(ALIGN_CHOICES);
    parent.add(align);
}

/**
 * Adds the format string text field.
 *
 * @param parent the panel that will contain the check box
 */
protected void addFormatString(JPanel parent) {
    formatText = new JTextField(format.getFormat());
    parent.add(formatText);
}

/**
 * Builds the border tab.
 */
protected JPanel buildBorderTab() {
    GridBagLayout bag = new GridBagLayout();
    GridBagConstraints c = new GridBagConstraints();
    c.insets = new Insets(6, 6, 6, 6);

    JPanel panel = new JPanel();
    panel.setLayout(bag);

    edgeWidgets = new EdgeWidgets[4];

    panel.add(buildBorderEdge(TOP, border.getTop(), bag, c, 0, 1));
    panel.add(buildBorderEdge(LEFT, border.getLeft(), bag, c, 1, 0));
    panel.add(buildBorderEdge(RIGHT, border.getRight(), bag, c, 1, 2));
    panel.add(buildBorderEdge(BOTTOM, border.getBottom(), bag, c, 2, 1));

    return panel;
}

/**
 * Builds one of the edges of the border.
 * <p>
 * <i>Warning</i>: this code depends upon the fact that the strings in
 * <code>EDGE_COUNT_CHOICES</code> equals the integer value of the string
 * (zero, one, two, etc.) and that the strings in
 * <code>EDGE_STYLE_CHOICES</code> correspond to the numeric values of the
 * <code>BorderEdge.STYLE_*</code> constants.
 *
 * @param edgeIndex one of <code>TOP</code>, <code>LEFT</code>, etc.
 * @param edge the edge we are representing visually
 * @param bag a grid bag layout
 * @param c grid bag constriaints
 * @param row the grid bag row
 * @param col the grid bag col
 */
protected JPanel buildBorderEdge(int edgeIndex, final BorderEdge edge,
				 GridBagLayout bag, GridBagConstraints c,
				 int row, int col)
{
    // Create container for all widgets associated with this edge
    EdgeWidgets ew = new EdgeWidgets(edge, edgeIndex);
    edgeWidgets[edgeIndex] = ew;

    // Labels
    JPanel labelPanel = new JPanel();
    labelPanel.setLayout(new GridLayout(0, 1));
    JLabel label;
    labelPanel.add(label = new JLabel("Count:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Style:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);
    labelPanel.add(label = new JLabel("Thickness:"));
    label.setHorizontalAlignment(SwingConstants.RIGHT);

    // Edit values
    JPanel fieldPanel = new JPanel();
    fieldPanel.setLayout(new GridLayout(0, 1));

    // Line count combo box
    final JComboBox edgeBox = new JComboBox(EDGE_COUNT_CHOICES);
    ew.numberComboBox = edgeBox;
    edgeBox.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	    String countStr = (String)edgeBox.getSelectedItem();
	    for (int i = 0; i < EDGE_COUNT_CHOICES.length; ++i)
		if (EDGE_COUNT_CHOICES[i].equals(countStr)) {
		    edge.setNumber(i);
		    break;
		}
	}
	});
    fieldPanel.add(edgeBox);

    // Line style combo box
    final JComboBox styleBox = new JComboBox(EDGE_STYLE_CHOICES);
    ew.styleComboBox = styleBox;
    styleBox.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	    String countStr = (String)styleBox.getSelectedItem();
	    for (int i = 0; i < EDGE_STYLE_CHOICES.length; ++i)
		if (EDGE_STYLE_CHOICES[i].equals(countStr)) {
		    edge.setStyle(i);
		    break;
		}
	}
	});
    fieldPanel.add(styleBox);

    // Line thickness text box
    JTextArea thicknessText = new JTextArea();
    ew.thicknessText = thicknessText;
    fieldPanel.add(thicknessText);

    // Parent panel, which we will return
    JPanel panel = new JPanel();
    panel.setLayout(new BorderLayout());
    panel.add(new JLabel(ew.getName()), BorderLayout.NORTH);
    panel.add(labelPanel, BorderLayout.CENTER);
    panel.add(fieldPanel, BorderLayout.EAST);

    c.gridx = col;
    c.gridy = row;
    bag.setConstraints(panel, c);

    return panel;
}

/**
 * Fills the format tab edit fields with values of format.
 */
protected void fillFormatTab() {
    int initialSizeIndex = sizeIndexOf((int)format.getSize());
    if (initialSizeIndex != -1)
	size.setSelectedIndex(initialSizeIndex);

    bold.setSelected(format.isBold());
    italic.setSelected(format.isItalic());
    underline.setSelected(format.isUnderline());
    wrap.setSelected(format.isWrap());

    // Assumes format align value == index in ALIGN_CHOICES
    align.setSelectedIndex(format.getAlign());

    formatText.setText(format.getFormat());
}

/**
 * Fills the border tab edit fields with values of border.
 */
protected void fillBorderTab() {
    for (int i = 0; i < 4; ++i) {
	EdgeWidgets ew = edgeWidgets[i];

	// Fill count. Assumes edge number (line count) value == index in
	// EDGE_COUNT_CHOICES.
	ew.numberComboBox.setSelectedIndex(ew.edge.getNumber());

	// Fill styles. Assumes edge style value == index in
	// EDGE_STYLE_CHOICES.
	ew.styleComboBox.setSelectedIndex(ew.edge.getStyle());

	// Fill thicknesses.
	ew.thicknessText.setText("" + ew.edge.getThickness());
    }
}

protected void doSave(boolean saveRevertInfo) {
    this.saveRevertInfo = saveRevertInfo;

    double fontSize = Double.parseDouble(size.getSelectedItem().toString());
    format.setSize(fontSize);
    format.setBold(bold.isSelected());
    format.setItalic(italic.isSelected());
    format.setUnderline(underline.isSelected());
    format.setWrap(wrap.isSelected());
    String alignText = (String)align.getSelectedItem();
    if (alignText.equals("Left"))
	format.setAlign(Format.ALIGN_LEFT);
    else if (alignText.equals("Center"))
	format.setAlign(Format.ALIGN_CENTER);
    else			// Right
	format.setAlign(Format.ALIGN_RIGHT);
    format.setFormat(formatText.getText());

    // The border count (number) and style save themselves, but we need to
    // read and set the thickness.
    for (int i = 0; i < 4; ++i) {
	EdgeWidgets ew = edgeWidgets[i];
	ew.edge.setThickness(Double.parseDouble(ew.thicknessText.getText()));
    }

    if (designWin.countSelectedFields() == 0)
	step(fieldWidget);
    else			// Call step() for all selected fields
	designWin.withSelectedWidgetsDo(this);
}

/**
 * Gives the format and borders to the specified field widget's field. If
 * needed, remembers the original settings by creating a revert info object.
 *
 * @param fw the field widget
 */
public void step(FieldWidget fw) {
    if (saveRevertInfo && !revertInfoList.containsKey(fw)) {
	Field f = fw.getField();
	RevertInfo ri = new RevertInfo(fw, f.getFormat(), f.getBorder());
	revertInfoList.put(fw, ri);
    }

    fw.applyNewFormatAndBorder(format, border);
}

protected void doRevert(Object revertInfo) {
    RevertInfo ri = (RevertInfo)revertInfo;
    Field f = ri.fw.getField();

    ri.fw.applyNewFormatAndBorder(ri.format, ri.border);

    if (ri.format != null)
	f.getFormat().setDirty(ri.format.isDirty());
    if (ri.border != null)
	f.getBorder().setDirty(ri.border.isDirty());
    f.setDirty(ri.fieldDirty);

    copyFormatAndBorder(ri.format, ri.border);
    fillFormatTab();
    fillBorderTab();
}

}
