package jimm.datavision.gui;
import jimm.datavision.*;
import jimm.datavision.sql.Table;
import jimm.datavision.sql.Column;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;

/**
 * This dialog is used for editing report groups.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class GroupWin extends EditWin
    implements ActionListener, ListSelectionListener
{

/* ================================================================ */
class ListItem {

Column col;
Group group;
int sortOrder;

ListItem(Column col, Group group) {
    this.col = col;
    this.group = group;		// May be null

    if (group == null)
	sortOrder = Group.SORT_ASCENDING;
    else
	sortOrder = group.getSortOrder();
}

public String toString() { return col.fullName(); }
}

/* ================================================================ */

protected static final int NO_CHANGE = 0;
protected static final int ONLY_SORTING_CHANGE = 1;
protected static final int DRASTIC_CHANGE = 2;

protected static final Dimension MINIMUM_LIST_SIZE =
    new Dimension(150, 200);

protected Report report;
protected DesignWin designWin;
protected DefaultListModel ungroupedModel;
protected DefaultListModel groupedModel;
protected JList ungroupedList;
protected JList groupedList;
protected JButton addOne, addAll, removeOne, removeAll;
protected JButton moveUp, moveDown;
protected JRadioButton ascendingRButton, descendingRButton;

/**
 * Constructor.
 *
 * @param designWin the window to which this dialog belongs
 * @param report the...um...I forgot
 */
public GroupWin(DesignWin designWin, Report report) {
    super(designWin, "Groups");
    this.designWin = designWin;
    this.report = report;
    buildWindow();
    pack();
    setVisible(true);
}

/**
 * Builds the window contents.
 *
 * @param whichTab the index of the tab to display when opened
 */
protected void buildWindow() {
    // Left hand columns list and right hand grouped columns list
    ungroupedModel = new DefaultListModel();
    groupedModel = new DefaultListModel();
    fillListModels();

    // Movement buttons
    Box moveButtonPanel = Box.createVerticalBox();
    addOne = new JButton(">");
    addOne.addActionListener(this);
    moveButtonPanel.add(addOne);

    addAll = new JButton(">>>");
    addAll.addActionListener(this);
    moveButtonPanel.add(addAll);

    removeOne = new JButton("<");
    removeOne.addActionListener(this);
    moveButtonPanel.add(removeOne);

    removeAll = new JButton("<<<");
    removeAll.addActionListener(this);
    moveButtonPanel.add(removeAll);

    // "Move Up", "Move Down", "Ascending" and "Descending" radio buttons
    Box orderButtonPanel = Box.createVerticalBox();

    moveUp = new JButton("Move Up");
    moveUp.addActionListener(this);
    orderButtonPanel.add(moveUp);

    moveDown = new JButton("Move Down");
    moveDown.addActionListener(this);
    orderButtonPanel.add(moveDown);

    ButtonGroup bg = new ButtonGroup();

    ascendingRButton = new JRadioButton("Ascending");
    ascendingRButton.addActionListener(this);
    orderButtonPanel.add(ascendingRButton);
    bg.add(ascendingRButton);

    descendingRButton = new JRadioButton("Descending");
    descendingRButton.addActionListener(this);
    orderButtonPanel.add(descendingRButton);
    bg.add(descendingRButton);

    // Ungrouped list and label in panel
    ungroupedList = new JList(ungroupedModel);
    ungroupedList.addListSelectionListener(this);
    ungroupedList.setMinimumSize(MINIMUM_LIST_SIZE);
    ungroupedList.addMouseListener(new MouseAdapter() {
	public void mouseClicked(MouseEvent e) {
	    if (e.getClickCount() == 2)
		moveToGrouped(ungroupedList.locationToIndex(e.getPoint()));
	}
	});
    Box ungroupedBox = Box.createVerticalBox();
    ungroupedBox.add(new JLabel("Columns"));
    ungroupedBox.add(new JScrollPane(ungroupedList));

    // Grouped list and label in panel
    groupedList = new JList(groupedModel);
    groupedList.addListSelectionListener(this);
    groupedList.setMinimumSize(MINIMUM_LIST_SIZE);
    groupedList.addMouseListener(new MouseAdapter() {
	public void mouseClicked(MouseEvent e) {
	    if (e.getClickCount() == 2)
		moveToUngrouped(groupedList.locationToIndex(e.getPoint()));
	}
	});
    Box groupedBox = Box.createVerticalBox();
    groupedBox.add(new JLabel("Group By"));
    groupedBox.add(new JScrollPane(groupedList));

    // Panel containing ungrouped, arrows, grouped, and radio buttons
    JPanel centerPanel = new JPanel();
    centerPanel.add(ungroupedBox);
    centerPanel.add(moveButtonPanel);
    centerPanel.add(groupedBox);
    centerPanel.add(orderButtonPanel);

    // Ok, Apply, Revert, and Cancel Buttons
    JPanel buttonPanel = closeButtonPanel();

    // Add tables and buttons to window
    getContentPane().add(centerPanel, BorderLayout.CENTER);
    getContentPane().add(buttonPanel, BorderLayout.SOUTH);

    // Update button states
    adjustButtons();
}

protected void fillListModels() {
    // First find and add groups in order.
    for (Iterator iter = report.groups(); iter.hasNext(); ) {
	Group group = (Group)iter.next();
	groupedModel.addElement(new ListItem(group.getColumn(), group));
    }

    // Now iterate through all available columns, adding those not already
    // grouped to the ungrouped list.
    Query query = report.getQuery();
    Collection tables = query.getTablesUsed();
    for (Iterator iter = query.getTablesUsed().iterator(); iter.hasNext(); ) {
	Table t = (Table)iter.next();
	for (Iterator cols = t.columns(); cols.hasNext(); ) {
	    Column col = (Column)cols.next();
	    Group group = report.findGroup(col);
	    if (group == null) {
		ListItem item = new ListItem(col, group);
		ungroupedModel.addElement(item);
	    }
	}
    }
}

/**
 * Moves the specified item number from the ungrouped to the grouped list.
 *
 * @param index the item number to move
 */
protected void moveToGrouped(int index) {
    ListItem item = (ListItem)ungroupedModel.elementAt(index);
    ungroupedModel.removeElementAt(index);
    groupedModel.addElement(item);
    adjustButtons();
}

/**
 * Moves the specified item number from the grouped to the ungrouped list.
 *
 * @param index the item number to move
 */
protected void moveToUngrouped(int index) {
    ListItem item = (ListItem)groupedModel.elementAt(index);
    groupedModel.removeElementAt(index);
    ungroupedModel.addElement(item);
    adjustButtons();
}

/**
 * Handles add and delete buttons.
 */
public void actionPerformed(ActionEvent e) {
    DefaultListModel ungroupedModel = (DefaultListModel)ungroupedList.getModel();
    DefaultListModel groupedModel = (DefaultListModel)groupedList.getModel();

    String cmd = e.getActionCommand();
    if (cmd.equals(">")) {
	Object[] items = ungroupedList.getSelectedValues();
	for (int i = 0; i < items.length; ++i) {
	    ungroupedModel.removeElement(items[i]);
	    groupedModel.addElement(items[i]);
	}
	adjustButtons();
    }
    else if (cmd.equals(">>>")) {
	Object[] items = ungroupedModel.toArray();
	for (int i = 0; i < items.length; ++i)
	    groupedModel.addElement(items[i]);
	ungroupedModel.removeAllElements();
	adjustButtons();
    }
    else if (cmd.equals("<")) {
	Object[] items = groupedList.getSelectedValues();
	for (int i = 0; i < items.length; ++i) {
	    ungroupedModel.addElement(items[i]);
	    groupedModel.removeElement(items[i]);
	}
	adjustButtons();
    }
    else if (cmd.equals("<<<")) {
	Object[] items = groupedModel.toArray();
	for (int i = 0; i < items.length; ++i)
	    ungroupedModel.addElement(items[i]);
	groupedModel.removeAllElements();
	adjustButtons();
    }
    else if (cmd.equals("Move Up")) {
	int index = groupedList.getSelectedIndex();
	Object item = groupedModel.remove(index);
	groupedModel.add(index - 1, item);
	groupedList.setSelectedIndex(index - 1);
    }
    else if (cmd.equals("Move Down")) {
	int index = groupedList.getSelectedIndex();
	Object item = groupedModel.remove(index);
	groupedModel.add(index + 1, item);
	groupedList.setSelectedIndex(index + 1);
    }
    else if (cmd.equals("Ascending")) {
	((ListItem)groupedList.getSelectedValue()).sortOrder =
	    Group.SORT_ASCENDING;
    }
    else if (cmd.equals("Descending")) {
	((ListItem)groupedList.getSelectedValue()).sortOrder =
	    Group.SORT_DESCENDING;
    }
    else
	super.actionPerformed(e);
}

public void valueChanged(ListSelectionEvent e) {
    adjustButtons();
}

protected void adjustButtons() {
    boolean selEmpty;

    addOne.setEnabled(!ungroupedList.isSelectionEmpty());
    addAll.setEnabled(ungroupedList.getModel().getSize() > 0);

    selEmpty = groupedList.isSelectionEmpty();
    if (selEmpty) {
	removeOne.setEnabled(false);
	moveUp.setEnabled(false);
	moveDown.setEnabled(false);
	ascendingRButton.setEnabled(false);
	descendingRButton.setEnabled(false);
	ascendingRButton.setSelected(false);
	descendingRButton.setSelected(false);
    }
    else {
	removeOne.setEnabled(true);

	int[] selectionIndices = groupedList.getSelectedIndices();
	if (selectionIndices.length == 1) {
	    int numGroups = groupedModel.size();
	    if (numGroups > 1) {
		int index = selectionIndices[0];
		moveUp.setEnabled(index != 0);
		moveDown.setEnabled(index != numGroups - 1);
	    }
	    else {
		moveUp.setEnabled(false);
		moveDown.setEnabled(false);
	    }

	    ascendingRButton.setEnabled(true);
	    descendingRButton.setEnabled(true);

	    ListItem item = (ListItem)groupedList.getSelectedValue();
	    if (item.sortOrder == Group.SORT_DESCENDING)
		descendingRButton.setSelected(true);
	    else
		ascendingRButton.setSelected(true);
	}
	else {
	    moveUp.setEnabled(false);
	    moveDown.setEnabled(false);
	    ascendingRButton.setEnabled(false);
	    descendingRButton.setEnabled(false);
	    ascendingRButton.setSelected(false);
	    descendingRButton.setSelected(false);
	}
    }
    removeAll.setEnabled(groupedList.getModel().getSize() > 0);
}

protected void doSave(boolean saveRevertInfo) {
    if (saveRevertInfo && !revertInfoList.containsKey(this))
	revertInfoList.put(this, null);
    else if (!saveRevertInfo) {
	// Only modify the report when we are really exiting. This is the
	// "real" save event.
	switch (whatChanged()) {
	case ONLY_SORTING_CHANGE:
	    for (Enumeration e = groupedModel.elements();
		 e.hasMoreElements(); )
	    {
		ListItem item = (ListItem)e.nextElement();
		item.group.setSortOrder(item.sortOrder);
	    }
	    break;
	case DRASTIC_CHANGE:
	    report.removeAllGroups();
	    for (Enumeration e = groupedModel.elements();
		 e.hasMoreElements(); )
	    {
		ListItem item = (ListItem)e.nextElement();
		Group g = item.group;
		if (g == null)
		    g = Group.create(report, item.col);
		g.setSortOrder(item.sortOrder);
		report.addGroup(g);
	    }
	    designWin.rebuildGroups();
	    break;
	case NO_CHANGE:
	    break;
	}
    }
}

/**
 * Determines the severity of the difference between the report and what we
 * have now: <code>NO_CHANGE</code>, <code>ONLY_SORTING_CHANGE</code>, or
 * <code>DRASTIC_CHANGE</code>.
 *
 * @return one of <code>NO_CHANGE</code>, <code>ONLY_SORTING_CHANGE</code>,
 * or <code>DRASTIC_CHANGE</code>
 */
protected int whatChanged() {
    int change = NO_CHANGE;
    Enumeration e = groupedModel.elements();
    Iterator iter = report.groups();
    while (e.hasMoreElements() && iter.hasNext()) {
	ListItem item = (ListItem)e.nextElement();
	Group origGroup = (Group)iter.next();
	if (item.group != origGroup)
	    return DRASTIC_CHANGE;
	if (item.sortOrder != origGroup.getSortOrder())
	    change = ONLY_SORTING_CHANGE;
    }
    if (e.hasMoreElements() || iter.hasNext())
	return DRASTIC_CHANGE;
    return change;		// Either NO_CHANGE or ONLY_SORTING_CHANGE
}

protected void doRevert(Object revertInfo) {
    // Rebuild list models
    ungroupedModel.removeAllElements();
    groupedModel.removeAllElements();
    fillListModels();
    adjustButtons();
}

}
