package jimm.datavision.gui;
import java.awt.Font;
import java.awt.event.*;
import javax.swing.*;

/**
 * Menu creation utilities.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class MenuUtils {

/**
 * Adds a single item to a popup menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item title
 * @param font font; may be <code>null</code>
 */
public static JMenuItem addToMenu(ActionListener listener, JPopupMenu menu,
				  String title, Font font)
{
    JMenuItem item = new JMenuItem(title);
    if (listener != null)
	item.addActionListener(listener);
    if (font != null)
	item.setFont(font);
    menu.add(item);
    return item;
}

/**
 * Adds a single item to a menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item title
 */
public static JMenuItem addToMenu(ActionListener listener, JMenu menu,
				  String title)
{
    return addToMenu(listener, menu, title, title, 0, 0, 0);
}

/**
 * Adds a single item to a menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item title
 * @param action the action string
 */
public static JMenuItem addToMenu(ActionListener listener, JMenu menu,
				  String title, String action)
{
    return addToMenu(listener, menu, title, action, 0, 0, 0);
}

/**
 * Adds a single item to a menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item title
 * @param action the action string
 * @param accelKey the accelerator key
 */
public static JMenuItem addToMenu(ActionListener listener, JMenu menu,
				  String title, String action, int accelKey)
{
    return addToMenu(listener, menu, title, action, accelKey, accelKey,
		     ActionEvent.CTRL_MASK);
}

/**
 * Adds a single item to a menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item title
 * @param action the action string
 * @param accelKey the accelerator key
 * @param shortcutKey the shortcut key
 */
public static JMenuItem addToMenu(ActionListener listener, JMenu menu,
				  String title, String action, int accelKey,
				  int shortcutKey)
{
    return addToMenu(listener, menu, title, action, accelKey, shortcutKey,
		     ActionEvent.CTRL_MASK);
}

/**
 * Adds a single item to a menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item title
 * @param action the action string
 * @param accelKey the accelerator key
 * @param shortcutKey the shortcut key
 * @param mask the modifier mask (for example, ActionEvent.CTRL_MASK)
 */
public static JMenuItem addToMenu(ActionListener listener, JMenu menu,
				  String title, String action, int accelKey,
				  int shortcutKey, int mask)
{
    JMenuItem item = new JMenuItem(title);
    setKeys(item, accelKey, shortcutKey, mask);
    if (action != null)
	item.setActionCommand(action);
    item.addActionListener(listener);
    menu.add(item);
    return item;
}

/**
 * Sets accelerator and shortcut keys. The shortcut key will be the same
 * as the accelerator key, and the shortcut key will use the default
 * modifier mask of <code>ActionEvent.CTRL_MASK</code>.
 *
 * @param accelKey the accelerator key
 */
public static void setKeys(JMenuItem item, int accelKey)
{
    setKeys(item, accelKey, accelKey, ActionEvent.CTRL_MASK);
}

/**
 * Sets accelerator and shortcut keys. The shortcut key will use the default
 * modifier mask of <code>ActionEvent.CTRL_MASK</code>.
 *
 * @param accelKey the accelerator key
 * @param shortcutKey the shortcut key
 */
public static void setKeys(JMenuItem item, int accelKey, int shortcutKey)
{
    setKeys(item, accelKey, shortcutKey, ActionEvent.CTRL_MASK);
}

/**
 * Sets accelerator and shortcut keys.
 *
 * @param accelKey the accelerator key
 * @param shortcutKey the shortcut key
 * @param mask the modifier mask (for example, ActionEvent.CTRL_MASK)
 */
public static void setKeys(JMenuItem item, int accelKey, int shortcutKey,
			   int mask)
{
    if (accelKey != 0)
	item.setMnemonic(accelKey);
    if (shortcutKey != 0)
	item.setAccelerator(KeyStroke.getKeyStroke(shortcutKey, mask));
}

/**
 * Returns a new align menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param useMnemonicKeys if <code>true</code>, mnemonic keys will be
 * assigned to the menu and its items
 * @param font font; may be <code>null</code>
 * @return a new menu
 */
public static JMenu buildAlignMenu(ActionListener listener,
				   boolean useMnemonicKeys, Font font)
{
    JMenu menu = new JMenu("Align");
    menu.setMnemonic(useMnemonicKeys ? KeyEvent.VK_A : 0);

    JMenuItem item = null;
    item = MenuUtils.addToMenu(listener, menu, "Tops", "align_top",
			       useMnemonicKeys ? KeyEvent.VK_T : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, "Middles", "align_middle",
			       useMnemonicKeys ? KeyEvent.VK_M : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, "Bottoms", "align_bottom",
			       useMnemonicKeys ? KeyEvent.VK_B : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, "Lefts", "align_left",
			       useMnemonicKeys ? KeyEvent.VK_L : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener,menu, "Centers", "align_center",
			       useMnemonicKeys ? KeyEvent.VK_C : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, "Rights", "align_right",
			       useMnemonicKeys ? KeyEvent.VK_R : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, "Snap to Grid", "snap_to_grid",
			       useMnemonicKeys ? KeyEvent.VK_S : 0, 0);
    if (font != null) item.setFont(font);

    return menu;
}

/**
 * Returns a new size menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param useMnemonicKeys if <code>true</code>, mnemonic keys will be
 * assigned to the menu and its items
 * @param font font; may be <code>null</code>
 * @return a new menu
 */
public static JMenu buildSizeMenu(ActionListener listener,
				  boolean useMnemonicKeys, Font font)
{
    JMenu menu = new JMenu("Size");
    menu.setMnemonic(useMnemonicKeys ? KeyEvent.VK_A : 0);

    JMenuItem item = null;
    item = MenuUtils.addToMenu(listener, menu, "Same Width", "size_width",
			       useMnemonicKeys ? KeyEvent.VK_W : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, "Same Height", "size_height",
			       useMnemonicKeys ? KeyEvent.VK_H : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, "Same Size", "size_size",
			       useMnemonicKeys ? KeyEvent.VK_S : 0, 0);
    if (font != null) item.setFont(font);

    return menu;
}
}
