package jimm.datavision.gui;
import jimm.datavision.Section;
import jimm.datavision.Field;
import jimm.datavision.TextField;
import jimm.datavision.FormulaField;
import jimm.datavision.ParameterField;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * A section widget is the visual representation of a report section.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
class SectionWidget extends JPanel implements ActionListener {

public static final int LHS_WIDTH = 125;
public static final int PAGE_WIDTH = (int)(72 * 8);
public static final int TOTAL_WIDTH = LHS_WIDTH + PAGE_WIDTH;

protected static Font DEFAULT_POPUP_FONT = new Font("Serif", Font.PLAIN, 10);

static final Color NORMAL_COLOR = Color.white;
static final Color SUPPRESSED_COLOR = Color.lightGray;

protected String name;
protected String popupName;
protected DesignWin designWin;
protected Section section;
protected SectionNameLabel label;
protected SectionFieldPanel fieldPanel;
protected JPopupMenu popup;
protected JMenuItem nameItem, toggleSuppress, shrink;
protected JMenuItem deleteGroup, addGroup, deleteSection, insertSection;

/**
 * An inner class that handles display of the popup menu.
 */
class PopupListener extends MouseAdapter {
public void mousePressed(MouseEvent e) { maybeShowPopup(e); }
public void mouseReleased(MouseEvent e) { maybeShowPopup(e); }
private void maybeShowPopup(MouseEvent e) {
    if (e.isPopupTrigger())
	showPopup(e);
}
}

/**
 * Constructor.
 *
 * @param win parent design window
 * @param sect the report section
 * @param name a name such as "Report Header (a)"
 * @param popupName the string to display as the first, disabled popup item
 */
SectionWidget(DesignWin win, Section sect, String name, String popupName)
{
    super();

    designWin = win;
    section = sect;
    this.name = name;
    this.popupName = popupName;

    setLayout(new SectionLayout());
    setPreferredSize(new Dimension(getWidth(), getHeight()));

    buildPopupMenu();
    addMouseListener(new PopupListener());

    // LHS name of section
    buildDisplayName();

    // Fields
    fieldPanel = new SectionFieldPanel(this);
    fieldPanel.setLayout(null);
    fieldPanel.setBackground(section.isSuppressed()
			     ? SUPPRESSED_COLOR : NORMAL_COLOR);
    add(fieldPanel);
    for (Iterator iter = section.fields(); iter.hasNext(); ) {
	Field f = (Field)iter.next();
	FieldWidget fw = makeWidgetFor(f, this);
	fieldPanel.add(fw, 0);	// Add to top of visual stack.
    }

    // If design window has initiated the drop of a new text field, create
    // one. Else deselect all fields.
    fieldPanel.addMouseListener(new PopupListener() {
	public void mouseClicked(MouseEvent e) {
	    deselectAll();
	    if (designWin.isPlacingNewTextField())
		createNewTextField(e); // Calls acceptNewTextField()
	}
	});

    // Resizer bar
    add(new SectionResizer(this, designWin.sectionContainer));
}

/**
 * Given a field, this factory method creates and returns the appropriate
 * widget.
 *
 * @param f a report field
 * @param sw a section widget; probably either <code>this</code> or
 * <code>null</code>
 */
FieldWidget makeWidgetFor(Field f, SectionWidget sw) {
    FieldWidget fw = null;
    if (f instanceof jimm.datavision.TextField)
	fw = new TextFieldWidget(this, f);
    else if (f instanceof FormulaField)
	fw = new FormulaWidget(this, f);
    else if (f instanceof ParameterField)
	fw = new ParameterWidget(this, f);
    else
	fw = new FieldWidget(this, f);
    return fw;
}

/**
 * Builds the popup menu.
 */
protected void buildPopupMenu() {
    popup = new JPopupMenu(popupName);

    nameItem = addPopupItem(popupName);
    nameItem.setEnabled(false);
    popup.addSeparator();
    toggleSuppress = addPopupItem("Suppress");
    popup.addSeparator();
    shrink = addPopupItem("Shrink to Fit");
    popup.addSeparator();
    deleteGroup = addPopupItem("Delete Group");
    addGroup = addPopupItem("Add Group");
    popup.addSeparator();
    deleteSection = addPopupItem("Delete Section");
    insertSection = addPopupItem("Insert Section Below");
}

/**
 * Adds a single menu item to the popup menu.
 *
 * @return the created item
 */
protected JMenuItem addPopupItem(String text) {
    JMenuItem item = new JMenuItem(text);
    item.setFont(DEFAULT_POPUP_FONT);
    item.addActionListener(this);
    popup.add(item);
    return item;
}

/**
 * Performs some action based on the action command string (the menu
 * item text).
 */
public void actionPerformed(ActionEvent e) {
    String command = e.getActionCommand();
    if (command == null) return;
    if (command.endsWith("Suppress"))
	toggleSuppression();
    else if (command.equals("Delete Group"))
	designWin.deleteGroupContaining(section);
    else if (command.equals("Add Group"))
	designWin.openNewGroupWin();
    else if (command.equals("Delete Section"))
	designWin.deleteSection(section);
    else if (command.equals("Insert Section Below"))
	designWin.insertSectionBelow(section);
    else if (command.equals("Shrink to Fit"))
	shrinkToFit();
}

/**
 * Modifies menu items based on the state of the section.
 */
protected void enablePopupMenuItems() {
    nameItem.setText(popupName);
    toggleSuppress.setText(section.isSuppressed()
			   ? "Don't Suppress" : "Suppress");
    deleteSection.setEnabled(!section.getReport().isOneOfAKind(section));
    deleteGroup.setEnabled(section.getReport().isInsideGroup(section));
}

/**
 * Constructs the section name widget that is displayed to the left of
 * the section.
 */
protected void buildDisplayName() {
    label = new SectionNameLabel(name, this);
    add(label);
}

/**
 * Set the section display name.
 *
 * @param name the new name
 */
public void setDisplayName(String name) {
    this.name = name;
    label.setText(name);
}

/**
 * Set the popup menu name, also displayed as first, disabled menu item.
 *
 * @param popuNpame the new name
 */
public void setPopupName(String popupName) {
    this.popupName = popupName;
    nameItem.setText(popupName);
}

/**
 * Returns the section we are representing.
 *
 * @return a report section
 */
public Section getSection() { return section; }

/**
 * Returns the width of the section, including the left-hand side name.
 *
 * @return the total width
 */
public int getWidth() { return TOTAL_WIDTH; }

/**
 * Returns the height of the section, including the resizer bar.
 *
 * @return the total height
 */
public int getHeight() {
    return (int)section.getHeight() + SectionResizer.HEIGHT;
}

/**
 * Returns the height of the report section.
 *
 * @return report section height
 */
public int getSectionHeight() { return (int)section.getHeight(); }

/**
 * Returns the minimum height the section needs to fit all of its fields.
 *
 * @return minimum height
 */
public int getMinSectionHeight() {
    int minY = 0;
    Component[] fieldWidgets = fieldPanel.getComponents();
    for (int i = 0; i < fieldWidgets.length; ++i) {
	int y = fieldWidgets[i].getBounds().y
	    + fieldWidgets[i].getBounds().height;
	if (y > minY)
	    minY = y;
    }
    return minY;
}

/**
 * Toggles the suppressed flag of the section.
 */
void toggleSuppression() {
    boolean newState = !section.isSuppressed();
    section.setSuppressed(newState);
    fieldPanel.setSuppressed(newState);	// Background color change
}

/**
 * Shrinks this section widget to the minimum height required.
 */
void shrinkToFit() {
    growBy(getMinSectionHeight() - getSectionHeight());
}

/**
 * Grows this section widget to the minimum height required.
 */
void growToFit() {
    int dy = getMinSectionHeight() - getSectionHeight();
    if (dy > 0)
	growBy(dy);
}

/**
 * Resizes the section. Called by resizer bar.
 *
 * @param dy delta y
 */
void growBy(int dy) {
    if (dy == 0) return;

    // Make sure section fits all fields.
    int origHeight = (int)section.getHeight();
    int newHeight = origHeight + dy;
    int minHeight = getMinSectionHeight();
    if (newHeight < minHeight) {
	newHeight = minHeight;
	if (newHeight == origHeight)
	    return;
    }

    section.setHeight(newHeight); // Hight not including resizer
    setPreferredSize(new Dimension(getWidth(), getHeight())); // Incl. resizer
    revalidate();
}

/**
 * Passes on to the design window the request to pick up all selected
 * fields for dragging (not just the specified field). Called from
 * <code>FieldWidget.mousePressed</code>.
 *
 * @param f the field in which the mouse has been clicked
 * @see FieldWidget#mousePressed
 * @see DesignWin#pickUp
 */
void pickUp(FieldWidget f) { designWin.pickUp(f); }

/**
 * Passes on to the design window the request to put down all fields
 * being dragged (not just the specified field). Called from
 * <code>FieldWidget.mouseReleased</code>.
 *
 * @param f the field in which the mouse has been clicked
 * @param origScreenPos the original location of the field in screen
 * coordinates
 * @param mouseScreenPos the location of the mouse in screen coordinates
 * @see FieldWidget#mouseReleased
 * @see DesignWin#putDown
 */
void putDown(FieldWidget f, Point origScreenPos, Point mouseScreenPos) {
    designWin.putDown(f, origScreenPos, mouseScreenPos);
}

/**
 * Tells the window to drag (move, resize) all selected fields. Called
 * from one field widget when it's being manipulated with the mouse.
 *
 * @param action a <code>FieldWidget.ACTION_*</code> constant
 * @param dx delta x
 * @param dy delta y
 * @see FieldWidget#mouseDragged
 * @see DesignWin#dragSelectedWidgets
 */
void dragSelectedWidgets(int action, int dx, int dy) {
    designWin.dragSelectedWidgets(action, dx, dy);
}

/**
 * Selects or deselcts a field widget, possibly deselecting all others
 * everywhere. Called from field widget itself, this passes the request
 * on to the design window.
 *
 * @param FieldWidget a field widget
 * @param makeSelected if <code>true</code>, select this field; else
 * deselect it
 * @param deselectAll if <code>true</code>, all other fields in all sections
 * are deselected first
 * @see DesignWin#select
 */
void select(FieldWidget fieldWidget, boolean makeSelected,
	    boolean deselectOthers)
{
    designWin.select(fieldWidget, makeSelected, deselectOthers);
}

/**
 * Deselects all fields in all sections. Tells the design window to do
 * so.
 *
 * @see DesignWin#deselectAll
 */
void deselectAll() {
    designWin.deselectAll();
}

/**
 * Adds field widget to panel. Does not affect models. Field retains its
 * selection state.
 *
 * @param fw field widget to add
 */
void addField(FieldWidget fw) {
    fieldPanel.add(fw, 0);	// Add to top of visual stack.
    fw.sectionWidget = this;
    fw.setBackground(section.isSuppressed() ? SUPPRESSED_COLOR : NORMAL_COLOR);
    growToFit();
}

/**
 * Removes field widget from panel, but do not change field model's relation
 * with section model. Field retains its selection state.
 * <p>
 * To delete a field widget completely, see
 * <code>DesignWin.deleteSelectedFields</code>.
 *
 * @param fw field widget to remove
 * @see #addField
 * @see DesignWin#deleteSelectedFields
 */
void removeField(FieldWidget fw) {
    fieldPanel.remove(fw);
}

/**
 * Asks the design window to snap the rectangle to it's grid.
 *
 * @param r a rectangle
 */
void snapToGrid(jimm.datavision.Rectangle r) {
    designWin.snapToGrid(r);
}

/**
 * Sets the visibility of all selected fields. Passes the buck to
 * the design window.
 *
 * @see FieldWidget#toggleVisibility
 */
void setFieldVisibility(boolean newVisiblity) {
    designWin.setFieldVisibility(newVisiblity);
}

/**
 * Creates and accepts a new text field.
 */
void createNewTextField(MouseEvent e) {
    Component c = e.getComponent();
    while (!(c instanceof SectionFieldPanel)) {
	java.awt.Rectangle bounds = c.getBounds();
	e.translatePoint(bounds.x, bounds.y);
	c = c.getParent();
	if (c == null)		// Should never happen
	    return;
    }

    // Create text field. Specify null section so the call to
    // FieldWidget.moveToSection() will do the correct thing.
    Field textField = Field.create(null, null, "text", "");
    textField.getBounds()
	.setBounds((double)e.getX(),
		   (double)e.getY() - (int)(Field.DEFAULT_HEIGHT / 2),
		   (double)Field.DEFAULT_WIDTH, (float)Field.DEFAULT_HEIGHT);

    // Create text widget, place it in this section, and select it.
    TextFieldWidget fw = new TextFieldWidget(this, textField);
    fw.moveToSection(this);

    // Accept the drop and start editing.
    designWin.acceptNewTextField();
    fw.startEditing();		// Selects the widget, too
}

/**
 * Displays popup menu, after enabling and disabling menu items.
 *
 * @param e mouse event that caused popup to do its thing
 */
void showPopup(MouseEvent e) {
    enablePopupMenuItems();
    popup.show(e.getComponent(), e.getX(), e.getY());
}

}
