package jimm.datavision.gui;
import jimm.datavision.Report;
import jimm.datavision.Query;
import jimm.datavision.sql.Table;
import jimm.datavision.sql.Column;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;

/**
 * This dialog is used for editing the report query's list of sort
 * orders.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class SortWin extends EditWin
    implements ActionListener, ListSelectionListener
{

/* ================================================================ */
class RevertInfo {
boolean queryDirty;
HashMap sorts;
RevertInfo(Query q) {
    queryDirty = q.isDirty();
    sorts = new HashMap();
    for (Iterator iter = q.sortedColumns(); iter.hasNext(); ) {
	Column col = (Column)iter.next();
	String order = q.sortOrderOf(col);
	sorts.put(col, order);
    }
}
}

/* ================================================================ */
class ListItem {

Column col;
boolean sortedDescending;

ListItem(Column col, String sortOrder) {
    this.col = col;
    if (sortOrder == null)
	sortedDescending = false;
    else
	sortedDescending = (sortOrder.toLowerCase().charAt(0) == 'd');
}

public String toString() { return col.fullName(); }
}

/* ================================================================ */

protected static final Dimension PREFERRED_LIST_SIZE =
    new Dimension(150, 200);

protected Report report;
protected Query query;
protected DefaultListModel unsortedModel;
protected DefaultListModel sortedModel;
protected JList unsortedList;
protected JList sortedList;
protected JButton addOne;
protected JButton addAll;
protected JButton removeOne;
protected JButton removeAll;
protected JRadioButton ascendingRButton;
protected JRadioButton descendingRButton;

/**
 * Constructor.
 *
 * @param designWin the window to which this dialog belongs
 * @param report the...um...I forgot
 */
public SortWin(DesignWin designWin, Report report) {
    super(designWin, "Record Sorting");
    this.report = report;
    query = report.getQuery();

    buildWindow();
    pack();
    setVisible(true);
}

/**
 * Builds the window contents.
 *
 * @param whichTab the index of the tab to display when opened
 */
protected void buildWindow() {
    // Left hand columns list and right hand sorted columns list
    unsortedModel = new DefaultListModel();
    sortedModel = new DefaultListModel();
    fillListModels();

    // Movement buttons
    Box moveButtonPanel = Box.createVerticalBox();

    addOne = new JButton(">");
    addOne.addActionListener(this);
    moveButtonPanel.add(addOne);

    addAll = new JButton(">>>");
    addAll.addActionListener(this);
    moveButtonPanel.add(addAll);

    removeOne = new JButton("<");
    removeOne.addActionListener(this);
    moveButtonPanel.add(removeOne);

    removeAll = new JButton("<<<");
    removeAll.addActionListener(this);
    moveButtonPanel.add(removeAll);

    // "Ascending" and "Descending" radio buttons
    Box orderButtonPanel = Box.createVerticalBox();
    ButtonGroup bg = new ButtonGroup();

    ascendingRButton = new JRadioButton("Ascending");
    ascendingRButton.addActionListener(this);
    orderButtonPanel.add(ascendingRButton);
    bg.add(ascendingRButton);

    descendingRButton = new JRadioButton("Descending");
    descendingRButton.addActionListener(this);
    orderButtonPanel.add(descendingRButton);
    bg.add(descendingRButton);

    // Unsorted list and label in panel
    unsortedList = new JList(unsortedModel);
    unsortedList.addListSelectionListener(this);
    unsortedList.setPreferredSize(PREFERRED_LIST_SIZE);
    unsortedList.addMouseListener(new MouseAdapter() {
	public void mouseClicked(MouseEvent e) {
	    if (e.getClickCount() == 2)
		moveToSorted(unsortedList.locationToIndex(e.getPoint()));
	}
	});
    Box unsortedBox = Box.createVerticalBox();
    unsortedBox.add(new JLabel("Columns"));
    unsortedBox.add(new JScrollPane(unsortedList));

    // Sorted and label in panel
    sortedList = new JList(sortedModel);
    sortedList.addListSelectionListener(this);
    sortedList.setPreferredSize(PREFERRED_LIST_SIZE);
    sortedList.addMouseListener(new MouseAdapter() {
	public void mouseClicked(MouseEvent e) {
	    if (e.getClickCount() == 2)
		moveToUnsorted(sortedList.locationToIndex(e.getPoint()));
	}
	});
    Box sortedBox = Box.createVerticalBox();
    sortedBox.add(new JLabel("Sort By"));
    sortedBox.add(new JScrollPane(sortedList));

    // Panel containing unsorted, arrows, sorted, and radio buttons
    JPanel centerPanel = new JPanel();
    centerPanel.add(unsortedBox);
    centerPanel.add(moveButtonPanel);
    centerPanel.add(sortedBox);
    centerPanel.add(orderButtonPanel);

    // Ok, Apply, Revert, and Cancel Buttons
    JPanel buttonPanel = closeButtonPanel();

    // Add tables and buttons to window
    getContentPane().add(centerPanel, BorderLayout.CENTER);
    getContentPane().add(buttonPanel, BorderLayout.SOUTH);

    // Update button states
    adjustButtons();
}

/**
 * Fills the unsorted and sorted columns lists. Excludes colums that are
 * used in report groups.
 */
protected void fillListModels() {
    Collection tables = query.getTablesUsed();
    for (Iterator iter = query.getTablesUsed().iterator(); iter.hasNext(); ) {
	Table t = (Table)iter.next();
	for (Iterator cols = t.columns(); cols.hasNext(); ) {
	    Column col = (Column)cols.next();
	    if (report.isGroupColumn(col)) // Ignore group columns
		continue;
	    String order = query.sortOrderOf(col);
	    ListItem item = new ListItem(col, order);
	    if (order == null)
		unsortedModel.addElement(item);
	    else
		sortedModel.addElement(item);
	}
    }
}

/**
 * Moves the specified item number from the unsorted to the sorted list.
 *
 * @param index the item number to move
 */
protected void moveToSorted(int index) {
    ListItem item = (ListItem)unsortedModel.elementAt(index);
    unsortedModel.removeElementAt(index);
    sortedModel.addElement(item);
    adjustButtons();
}

/**
 * Moves the specified item number from the sorted to the unsorted list.
 *
 * @param index the item number to move
 */
protected void moveToUnsorted(int index) {
    ListItem item = (ListItem)sortedModel.elementAt(index);
    sortedModel.removeElementAt(index);
    unsortedModel.addElement(item);
    adjustButtons();
}

/**
 * Handles add and delete buttons.
 */
public void actionPerformed(ActionEvent e) {
    DefaultListModel unsortedModel = (DefaultListModel)unsortedList.getModel();
    DefaultListModel sortedModel = (DefaultListModel)sortedList.getModel();

    String cmd = e.getActionCommand();
    if (cmd.equals(">")) {
	Object[] items = unsortedList.getSelectedValues();
	for (int i = 0; i < items.length; ++i) {
	    unsortedModel.removeElement(items[i]);
	    sortedModel.addElement(items[i]);
	}
	adjustButtons();
    }
    else if (cmd.equals(">>>")) {
	Object[] items = unsortedModel.toArray();
	for (int i = 0; i < items.length; ++i)
	    sortedModel.addElement(items[i]);
	unsortedModel.removeAllElements();
	adjustButtons();
    }
    else if (cmd.equals("<")) {
	Object[] items = sortedList.getSelectedValues();
	for (int i = 0; i < items.length; ++i) {
	    unsortedModel.addElement(items[i]);
	    sortedModel.removeElement(items[i]);
	}
	adjustButtons();
    }
    else if (cmd.equals("<<<")) {
	Object[] items = sortedModel.toArray();
	for (int i = 0; i < items.length; ++i)
	    unsortedModel.addElement(items[i]);
	sortedModel.removeAllElements();
	adjustButtons();
    }
    else if (cmd.equals("Ascending")) {
	((ListItem)sortedList.getSelectedValue()).sortedDescending = false;
    }
    else if (cmd.equals("Descending")) {
	((ListItem)sortedList.getSelectedValue()).sortedDescending = true;
    }
    else
	super.actionPerformed(e);
}

public void valueChanged(ListSelectionEvent e) {
    adjustButtons();
}

protected void adjustButtons() {
    boolean selEmpty;

    addOne.setEnabled(!unsortedList.isSelectionEmpty());
    addAll.setEnabled(unsortedList.getModel().getSize() > 0);

    selEmpty = sortedList.isSelectionEmpty();
    if (selEmpty) {
	removeOne.setEnabled(false);
	ascendingRButton.setEnabled(false);
	descendingRButton.setEnabled(false);
	ascendingRButton.setSelected(false);
	descendingRButton.setSelected(false);
    }
    else {
	removeOne.setEnabled(true);

	int[] selectionIndices = sortedList.getSelectedIndices();
	if (selectionIndices.length == 1) {
	    ascendingRButton.setEnabled(true);
	    descendingRButton.setEnabled(true);

	    ListItem item = (ListItem)sortedList.getSelectedValue();
	    if (item.sortedDescending)
		descendingRButton.setSelected(true);
	    else
		ascendingRButton.setSelected(true);
	}
	else {
	    ascendingRButton.setEnabled(false);
	    descendingRButton.setEnabled(false);
	    ascendingRButton.setSelected(false);
	    descendingRButton.setSelected(false);
	}
    }
    removeAll.setEnabled(sortedList.getModel().getSize() > 0);
}

protected void doSave(boolean saveRevertInfo) {
    if (saveRevertInfo && !revertInfoList.containsKey(this))
	revertInfoList.put(this, new RevertInfo(query));

    query.clearSorts();
    for (Enumeration e = sortedModel.elements(); e.hasMoreElements(); ) {
	ListItem item = (ListItem)e.nextElement();
	query.addSort(item.col, item.sortedDescending ? "desc" : "asc");
    }
}

protected void doRevert(Object revertInfo) {
    RevertInfo ri = (RevertInfo)revertInfo;

    query.clearSorts();
    for (Iterator iter = ri.sorts.keySet().iterator(); iter.hasNext(); ) {
	Column col = (Column)iter.next();
	query.addSort(col, (String)ri.sorts.get(col));
    }
    query.setDirty(ri.queryDirty);

    // Rebuild widgets
    unsortedModel.removeAllElements();
    sortedModel.removeAllElements();
    fillListModels();
    adjustButtons();
}

}
