package jimm.datavision.gui;
import jimm.datavision.Field;
import java.awt.Cursor;
import java.awt.event.MouseEvent;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ActionEvent;
import javax.swing.text.Highlighter;
import javax.swing.text.BadLocationException;

/**
 *
 * A text field widget is a field widget that is editable.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
class TextFieldWidget extends FieldWidget implements KeyListener {

protected int origHeight;
protected int lineHeight;
protected boolean changingEditState;

/**
 * Constructor.
 *
 * @param sw section widget in which the field's new widget will reside
 * @param field a report field
 */
TextFieldWidget(SectionWidget sw, Field field) {
    super(sw, field);

    // A reasonable (but possibly incorrect) guess of a text line height.
    jimm.datavision.Format fmt = field.getFormat();
    if (fmt == null) fmt = jimm.datavision.Format.DEFAULT;
    lineHeight = (int)fmt.getSize();

    addKeyListener(this);
}

protected String getPopupNameText() {
    return "Text Field";
}

protected void addCustomPopupItems() {
    MenuUtils.addToMenu(this, popup, "Edit...", POPUP_FONT);
    popup.addSeparator();
}

/**
 * Performs some action based on the action command string (the menu
 * item text).
 */
public void actionPerformed(ActionEvent e) {
    String command = e.getActionCommand();
    if (command == null) return;

    if (command.equals("Edit..."))
	startEditing();
    else
	super.actionPerformed(e);
}

/**
 * if this is a double-click, start editing; else handle the mouse event
 * like a normal field widget.
 *
 * @param e mouse event
 */
public void mouseClicked(MouseEvent e) {
    if (e.getClickCount() == 2)
	startEditing();
    else
	super.mouseClicked(e);
}

/**
 * Makes our text editable and starts editing.
 */
protected void startEditing() {
    // Bug fix: endEditing is called while startEditing is running
    synchronized(this) {
	if (changingEditState)
	    return;
	changingEditState = true;
    }

    if (!selected)
	select(true);
    setEditable(true);
    getCaret().setVisible(true);
    requestFocus();
    origHeight = getBounds().height;

    sectionWidget.designWin.setIgnoreKeys(true);

    removeMouseListener(this);
    removeMouseMotionListener(this);

    changingEditState = false;
}

/**
 * Stores new value and bounds from self into field and makes text
 * non-editable.
 */
protected void endEditing() {
    // Bug fix: endEditing is called while startEditing is running
    synchronized(this) {
	if (changingEditState)
	    return;
	changingEditState = true;
    }

    if (isEditable()) {
	// If we wait until later to grab the text, then we loose text with
	// newlines in it. I don't know why. Go figure. It may be the
	// updates caused by the bounds change that cause it; I don't know.
	String newText = getText();

	setEditable(false);
	getCaret().setVisible(false);

	java.awt.Rectangle bounds = getBounds();
	if (bounds.height != origHeight)
	    field.getBounds().setBounds(bounds);

	// Set text after settings bounds because setting the text triggers
	// an update which in turn re-sets the field's bounds, erasing any
	// changes we made in keyTyped().
	//
	// See also the note above: we must retrieve the text from the edit
	// widget before getting here.
	field.setValue(newText);

	sectionWidget.designWin.setIgnoreKeys(false);

	addMouseListener(this);
	addMouseMotionListener(this);
    }

    changingEditState = false;
}

/**
 * Perform selection; if becoming deselected, ends editing.
 *
 * @param makeSelected new selection state
 */
protected void doSelect(boolean makeSelected) {
    if (makeSelected == false)
	endEditing();
    super.doSelect(makeSelected);
}

/**
 * If this field is being edited, show text cursor. Else call superclass
 * method.
 *
 * @param e a mouse event
 */
protected void cursorForPosition(MouseEvent e) {
    if (isEditable())
	setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
    else
	super.cursorForPosition(e);
}

/**
 * Handles return key by expanding height of editor and backspace and
 * delete by shrinking height if newline is deleted.
 *
 * @param e key event
 */
public void keyTyped(KeyEvent e) {
    char c = e.getKeyChar();
    int pos;
    java.awt.Rectangle bounds;

    switch (c) {
    case KeyEvent.VK_ENTER:
	// We've just grown. Make sure we still fit.
	bounds = getBounds();
	bounds.height += lineHeight;
	setBounds(bounds);

	// Make sure we still fit in section
	sectionWidget.growToFit();
	break;
    case KeyEvent.VK_BACK_SPACE:
	try {
	    pos = getCaretPosition();
	    if (pos > 0 && getText(pos - 1, 1).equals("\n")) {
		bounds = getBounds();
		bounds.height -= lineHeight;
		setBounds(bounds);
	    }
	}
	catch (BadLocationException ex) {}
	break;
    case KeyEvent.VK_DELETE:
	try {
	    pos = getCaretPosition();
	    if (pos < getText().length()
		&& getText(pos, 1).equals("\n"))
	    {
		bounds = getBounds();
		bounds.height -= lineHeight;
		setBounds(bounds);
	    }
	}
	catch (BadLocationException ex2) {}
	break;
    }
}

public void keyPressed(KeyEvent e) {}
public void keyReleased(KeyEvent e) {}

}
