package jimm.datavision.layout;
import jimm.datavision.*;
import jimm.datavision.gui.ExportWin;
import jimm.datavision.gui.DesignWin;
import jimm.datavision.gui.MenuUtils;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.print.*;
import javax.swing.*;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * <code>SwingLE</code> is a layout engine that creates a Swing window.
 * The window can be printed by selecting the appropriate menu item.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class SwingLE extends LayoutEngine {

protected static final Dimension WINDOW_START_SIZE = new Dimension(600, 400);
protected static final String PRINT_ICON =
    "toolbarButtonGraphics/general/Print16.gif";
protected static final String EXPORT_ICON =
    "toolbarButtonGraphics/general/Export16.gif";
protected static final String FIRST_ICON =
    "toolbarButtonGraphics/navigation/Home16.gif";
protected static final String PREV_ICON =
    "toolbarButtonGraphics/navigation/Back16.gif";
protected static final String NEXT_ICON =
    "toolbarButtonGraphics/navigation/Forward16.gif";
protected static final String LAST_ICON =
    "toolbarButtonGraphics/navigation/Down16.gif";

protected JFrame frame;
protected Dimension pageDim;
protected JPanel cardPanel;
protected ArrayList pages;
protected SwingPage pageBeingBuilt;
protected int displayPageNum;	// Starts at 1
protected JLabel pageCountLabel;
protected Action printAction;
protected Action closeAction;
protected Action exportAction;
protected Action goFirstAction;
protected Action goPrevAction;
protected Action goNextAction;
protected Action goLastAction;

/**
 * Constructor.
 *
 * @param report the report to display.
 */
public SwingLE() {
    super();
    pages = new ArrayList();
    pageDim = new Dimension((int)pageWidth(), (int)pageHeight());
}

/**
 * Creates window and displays a blank page.
 */
protected void doStart() {
    frame = new JFrame(report.getTitle());
    makeActions();
    makeMenu(frame);
    frame.getContentPane().add(makeToolbar(), BorderLayout.NORTH);

    // Card panel for displaying pages
    cardPanel = new JPanel();
    cardPanel.setPreferredSize(WINDOW_START_SIZE);
    CardLayout cardLayout = new CardLayout(0, 0);
    cardPanel.setLayout(cardLayout);
    frame.getContentPane().add(cardPanel, BorderLayout.CENTER);

    // New, blank, dummy card we can display until the first page is
    // generated.
    JPanel blankPage = new JPanel();
    blankPage.setBackground(Color.white);
    blankPage.setPreferredSize(pageDim);
    cardPanel.add(blankPage, "blank page");
    cardLayout.show(cardPanel, "blank page");

    displayPageNum = 0;

    frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    frame.addWindowListener(new WindowAdapter() {
	public void windowClosing(WindowEvent e) {
	    close();
	}
	});

    frame.pack();
    frame.setVisible(true);
}

/**
 * Done loading report.
 */
protected void doEnd() {
    printAction.setEnabled(true);
}

/**
 * Creates the actions used by menu items and toolbar widgets.
 */
protected void makeActions() {
    URL url = ClassLoader.getSystemResource(PRINT_ICON);
    printAction = new AbstractAction("Print", new ImageIcon(url, "Print")) {
	public void actionPerformed(ActionEvent e) { printReport(); }
	};
    printAction.putValue(Action.SHORT_DESCRIPTION, "Print");
    printAction.setEnabled(false);

    closeAction = new AbstractAction("Close") {
	public void actionPerformed(ActionEvent e) { close(); }
	};

    url = ClassLoader.getSystemResource(EXPORT_ICON);
    exportAction = new AbstractAction("Export", new ImageIcon(url, "Export")) {
	public void actionPerformed(ActionEvent e) { export(); }
	};
    exportAction.putValue(Action.SHORT_DESCRIPTION, "Export");

    url = ClassLoader.getSystemResource(FIRST_ICON);
    goFirstAction = new AbstractAction("First Page",
				       new ImageIcon(url, "First Page")) {
	public void actionPerformed(ActionEvent e) { displayFirstPage(); }
	};
    goFirstAction.putValue(Action.SHORT_DESCRIPTION, "First Page");
    goFirstAction.setEnabled(false);

    url = ClassLoader.getSystemResource(PREV_ICON);
    goPrevAction = new AbstractAction("Previous Page",
				      new ImageIcon(url, "Previous Page")) {
	public void actionPerformed(ActionEvent e) { displayPrevPage(); }
	};
    goPrevAction.putValue(Action.SHORT_DESCRIPTION, "Previous Page");
    goPrevAction.setEnabled(false);

    url = ClassLoader.getSystemResource(NEXT_ICON);
    goNextAction = new AbstractAction("Next Page",
				      new ImageIcon(url, "Next Page")) {
	public void actionPerformed(ActionEvent e) { displayNextPage(); }
	};
    goNextAction.putValue(Action.SHORT_DESCRIPTION, "Next Page");
    goNextAction.setEnabled(false);

    url = ClassLoader.getSystemResource(LAST_ICON);
    goLastAction = new AbstractAction("Last Page",
				      new ImageIcon(url, "Last Page")) {
	public void actionPerformed(ActionEvent e) { displayLastPage(); }
	};
    goLastAction.putValue(Action.SHORT_DESCRIPTION, "Last Page");
    goLastAction.setEnabled(false);
}

/**
 * Creates the window menu.
 *
 * @param frame the window that will contain the menu
 */
protected void makeMenu(JFrame frame) {
    JMenuBar menuBar = new JMenuBar();
    frame.setJMenuBar(menuBar);

    // File menu
    JMenu menu = new JMenu("File");
    menu.setMnemonic(KeyEvent.VK_F);
    menuBar.add(menu);

    MenuUtils.setKeys(menu.add(printAction), KeyEvent.VK_P);
    menu.addSeparator();
    MenuUtils.setKeys(menu.add(closeAction), KeyEvent.VK_C, KeyEvent.VK_W);

    // View menu
    menu = new JMenu("View");
    menu.setMnemonic(KeyEvent.VK_V);
    menuBar.add(menu);

    MenuUtils.setKeys(menu.add(goFirstAction), KeyEvent.VK_F, KeyEvent.VK_1);
    MenuUtils.setKeys(menu.add(goPrevAction), KeyEvent.VK_P, KeyEvent.VK_2);
    MenuUtils.setKeys(menu.add(goNextAction), KeyEvent.VK_N, KeyEvent.VK_3);
    MenuUtils.setKeys(menu.add(goLastAction), KeyEvent.VK_L, KeyEvent.VK_4);

    // Report menu
    menu = new JMenu("Report");
    menu.setMnemonic(KeyEvent.VK_R);
    menuBar.add(menu);

    MenuUtils.setKeys(menu.add(exportAction), KeyEvent.VK_E);
}

/**
 * Creates and returns a new tool bar.
 */
protected JToolBar makeToolbar() {
    JToolBar bar = new JToolBar(javax.swing.SwingConstants.HORIZONTAL);
    bar.add(printAction);
    bar.addSeparator();
    bar.add(exportAction);
    bar.addSeparator();
    bar.add(goFirstAction);
    bar.add(goPrevAction);
    bar.add(goNextAction);
    bar.add(goLastAction);
    pageCountLabel = new JLabel("Loading first page...");
    bar.add(pageCountLabel);
    return bar;
}

/**
 * Creates a new page.
 */
protected void doStartPage() {
    // Create page and add it to our list of pages
    pageBeingBuilt = new SwingPage();
    pageBeingBuilt.setPreferredSize(pageDim);
    pages.add(pageBeingBuilt);

    // Scrollable pane for page
    JScrollPane scroller = new JScrollPane(pageBeingBuilt);

    // Add new card containing scrollable page
    cardPanel.add(scroller, "" + pages.size());
}

/**
 * Displays the page, if it is the first one.
 */
protected void doEndPage() {
    // If this is the first page, display it.
    if (pages.size() == 1) {
	CardLayout cardLayout = (CardLayout)cardPanel.getLayout();
	cardLayout.show(cardPanel, "" + pages.size());
	displayPageNum = 1;
    }
    updateGoActions();
}

/**
 * Closes this window. Does not call <code>System.exit()</code>.
 */
protected void close() {
    frame.setVisible(false);
    frame.dispose();
}

/**
 * Opens export dialog.
 */
protected void export() {
    new ExportWin(DesignWin.findWindowFor(report), report);
}

/**
 * Updates navigation buttons based on number of pages and current
 * display page.
 */
protected void updateGoActions() {
    int numPages = pages.size();

    boolean canGoBack = numPages > 0 && displayPageNum > 1;
    goFirstAction.setEnabled(canGoBack);
    goPrevAction.setEnabled(canGoBack);

    boolean canGoForward = numPages > 0 && displayPageNum < numPages;
    goNextAction.setEnabled(canGoForward);
    goLastAction.setEnabled(canGoForward);

    pageCountLabel.setText("Page " + displayPageNum + " of " + numPages);
}

/** Performs the "First Page" command. */
protected void displayFirstPage() {
    if (pages.size() > 0 && displayPageNum != 1) {
	displayPageNum = 1;
	CardLayout cardLayout = (CardLayout)cardPanel.getLayout();
	cardLayout.show(cardPanel, "" + displayPageNum);
	updateGoActions();
    }
}

/** Performs the "Next Page" command. */
protected void displayNextPage() {
    if (pages.size() > displayPageNum) {
	++displayPageNum;
	CardLayout cardLayout = (CardLayout)cardPanel.getLayout();
	cardLayout.show(cardPanel, "" + displayPageNum);
	updateGoActions();
    }
}

/** Performs the "Previous Page" command. */
protected void displayPrevPage() {
    if (displayPageNum > 1) {
	--displayPageNum;
	CardLayout cardLayout = (CardLayout)cardPanel.getLayout();
	cardLayout.show(cardPanel, "" + displayPageNum);
	updateGoActions();
    }
}

/** Performs the "Last Page" command. */
protected void displayLastPage() {
    int numPages = pages.size();
    if (numPages > 0 && numPages != displayPageNum) {
	displayPageNum = numPages;
	CardLayout cardLayout = (CardLayout)cardPanel.getLayout();
	cardLayout.show(cardPanel, "" + displayPageNum);
	updateGoActions();
    }
}

/**
 * Creates a new swing field and adds it to the current page.
 *
 * @param field the report field
 * @see SwingField
 */
protected void doOutputField(Field field) {
    String fieldAsString = field.toString();
    if (fieldAsString == null || fieldAsString.length() == 0)
	return;

    // Create label and add to page.
    SwingField label = new SwingField(field, fieldAsString);
    pageBeingBuilt.add(label);

    // Set bounds.
    jimm.datavision.Rectangle bounds = field.getBounds();
    // Page footers are anchored to page bottom.
    int y = (currentSectionType == SECT_PAGE_FOOTER)
	? (int)(pageHeight() - currentSection.getHeight() + bounds.y)
	: (int)(pageHeightUsed + bounds.y);
    label.setBounds((int)bounds.x, y, (int)bounds.width, (int)bounds.height);
}

/**
 * Creates a new line. <em>Unimplemented</em>.
 *
 * @param line a line
 */
protected void doOutputLine(Line line) {}

/**
 * Prints the report.
 */
protected void printReport() {
    PrinterJob printJob = PrinterJob.getPrinterJob();
    Book book = new Book();
    for (Iterator iter = pages.iterator(); iter.hasNext(); )
	book.append((Printable)iter.next(), printJob.defaultPage());
    printJob.setPageable(book);

    if (printJob.printDialog()) {
	try {
	    printJob.print();
	} catch (Exception ex) {
	    ErrorHandler.error(ex);
	}
    }
}

}
