package jimm.datavision.sql;
import jimm.datavision.Draggable;
import java.sql.*;
import java.util.HashMap;

/**
 * A database column. It knows the table to which it belongs, its name,
 * and other metadata. The id of a column is a string of the form
 * "table_name.column_name".
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Column implements Draggable {

protected static HashMap columns = new HashMap();

protected String id;
protected Table table;
protected String name;
protected int type;
protected String typeString;
protected int size;
protected boolean isNullable;
protected String defaultValue;
protected Object value;

/**
 * Given a unique id, returns the column having that id.
 *
 * @param id the id
 * @return the column having that id
 */
public static Column find(String id) {
    return (Column)columns.get(id);
}

/**
 * Constructor.
 *
 * @param table the table to which this column belongs
 * @param rset return value from <code>DatabaseMetaData.getColumns</code>
 * @see java.sql.DatabaseMetaData#getColumns
 */
public Column(Table table, ResultSet rset) throws SQLException {
    this.table = table;
    name = rset.getString("COLUMN_NAME");
    id = table.name + "." + name;
    type = rset.getInt("DATA_TYPE");
    typeString = rset.getString("TYPE_NAME");
    size = rset.getInt("COLUMN_SIZE");
    isNullable = !(rset.getInt("NULLABLE") == DatabaseMetaData.columnNoNulls);
    defaultValue = rset.getString("COLUMN_DEF");

    columns.put(id, this);
}

/**
 * Returns the column's name. To retrieve the full "table.column" name,
 * use <code>fullName</code>.
 *
 * @return a string describing the column's type.
 * @see #fullName
 */
public String getName() { return name; }

/**
 * Returns the type constant (a <code>java.sql.Types</code> value).
 *
 * @return a <code>java.sql.Types</code> value
 * @see java.sql.Types
 */
public int getType() { return type; }

/**
 * Returns the table id. This is a string of the form
 * "table_name.column_name".
 *
 * @return the id string
 */
public String getId() { return id; }

/**
 * Returns the table to which this column belongs.
 *
 * @return the table
 */
public Table getTable() { return table; }

/**
 * Returns the full named of this column. This is a string of the form
 * "table_name.column_name". To retrieve just the column name,
 * use <code>getName</code>.
 *
 * @return the id string
 * @see #getName
 */
public String fullName() {
    return id;
}

public String dragString() {
    return "column:" + fullName();
}

/**
 * Returns a string representation of this column. Calls
 * <code>fullName</code>.
 *
 * @see #fullName
 */
public String toString() {
    return fullName();
}

}
