package jimm.datavision.sql;
import jimm.datavision.Report;
import jimm.datavision.IndentWriter;
import jimm.datavision.Dirtiable;
import jimm.datavision.Writeable;
import jimm.datavision.ErrorHandler;
import jimm.datavision.UserCancellationException;
import java.sql.*;
import java.util.*;

/**
 * A database knows about the tables in a database.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Database implements Dirtiable, Writeable {

protected Report report;
protected String driverClassName;
protected String connInfo;
protected String name;
protected String username;
protected ArrayList tables;
protected Connection conn;
private String password;
protected boolean hasPassword;
protected boolean dirty;

//  /**
//   * Returns a collection of all database names in this database.
//   * <em>Unimplemented; returns an empty collection</em>.
//   *
//   * @return a collection of all database names in the database.
//   */
//  public static ArrayList allDatabaseNames() {
//  // FIX
//      return new ArrayList();
//  }

/**
 * Constructor.
 *
 * @param driverClassName database driver class name
 * @param connInfo database connection info string
 * @param report the report using this database
 * @param name the database name
 * @param username the user name to use when logging in to the database
 * @param password the database password
 */
public Database(String driverClassName, String connInfo, Report report,
		String name, String user)
    throws SQLException, ClassNotFoundException, InstantiationException,
	   IllegalAccessException, UserCancellationException
{
    this(driverClassName, connInfo, report, name, user, null, false);
}

/**
 * Constructor.
 *
 * @param driverClassName database driver class name
 * @param connInfo database connection info string
 * @param report the report using this database
 * @param name the database name
 * @param username the user name to use when logging in to the database
 * @param password the database password
 */
public Database(String driverClassName, String connInfo, Report report,
		String name, String user, String password)
    throws SQLException, ClassNotFoundException, InstantiationException,
	   IllegalAccessException, UserCancellationException
{
    this(driverClassName, connInfo, report, name, user, password, true);
}


/**
 * Constructor.
 *
 * @param driverClassName database driver class name
 * @param connInfo database connection info string
 * @param report the report using this database
 * @param name the database name
 * @param username the user name to use when logging in to the database
 * @param password the database password
 * @param givenPassword if <code>true</code>, the password was passed in
 * to some other constructor
 */
protected Database(String driverClassName, String connInfo, Report report,
		   String name, String user, String password,
		   boolean givenPassword)
    throws SQLException, ClassNotFoundException, InstantiationException,
	   IllegalAccessException, UserCancellationException
{
    this.driverClassName = driverClassName;
    this.connInfo = connInfo;
    this.report = report;
    this.name = name;
    this.username = (user == null) ? "" : user;
    this.password = password;
    tables = null;
    hasPassword = givenPassword;
    initializeConnection();
    loadAllTables();
}

/**
 * Initializes the connection to the database.
 *
 * @return a connection to the database
 */
public void initializeConnection()
    throws ClassNotFoundException, InstantiationException,
	   IllegalAccessException, UserCancellationException
{
    // Keep trying until we succeed. We will only exit if we succeed,
    // if the user cancels, or if something bad happens while obtaining
    // the database connection.
    boolean ok = false;
    while (!ok) {
	if (!hasPassword) {
	    // Calling askForPassword sets our user name and password
	    report.askForPassword(this);
	    hasPassword = true;
	}
	if (password == null)
	    throw new UserCancellationException("Password entry cancelled by user");

	try {
	    if (connInfo == null || connInfo.length() == 0)
		throw new IllegalArgumentException("connection info string is missing");
	    if (username == null || username.length() == 0)
		throw new IllegalArgumentException("user name is missing");
	    if (password == null)
		throw new IllegalArgumentException("password is null");

	    // Load the database JDBC driver
	    Driver d = (Driver)Class.forName(driverClassName).newInstance();
	    DriverManager.registerDriver(d);

	    // Connect to the database
	    conn = DriverManager.getConnection(connInfo, username, password);

	    ok = true;
	}
	catch (SQLException sqle) {	// Force another password request
	    System.err.println("" + sqle.getErrorCode() + ": "
			       + sqle.toString()); // DEBUG
	    hasPassword = false;
	    report.setDatabasePassword(null);
	}
    }
}


/**
 * Returns a connection to the database.
 *
 * @return a connection to the database
 */
public Connection getConnection() { return conn; }

/**
 * Reset key instance variables, closes current connection, and "reloads"
 * all table information (compares new info with existing info and complains
 * if any existing info is not in the new info).
 *
 * @param driverClassName database driver class name
 * @param connInfo database connection info string
 * @param dbName database name
 * @param username the user name to use when logging in to the database
 */
public void reset(String driverClassName, String connInfo, String dbName,
		  String username, String password)
    throws SQLException, ClassNotFoundException, InstantiationException,
	   IllegalAccessException, UserCancellationException
{
    setDriverClassName(driverClassName);
    setConnectionInfo(connInfo);
    setName(dbName);
    setUserName(username);
    this.password = password;
    hasPassword = true;

    if (conn != null) {
	conn.close();
	conn = null;
    }
    loadAllTables();
}

/**
 * Loads information about all tables in the database. If this is not the
 * first time we are loading tables (we were called from <code>reset</code>),
 * then perform a comparison between the metadata obtained from the new
 * and old connections.
 */
protected void loadAllTables() throws SQLException {
    ArrayList newTableList = new ArrayList();
    boolean missingColumn = false;

    DatabaseMetaData dbmd = getConnection().getMetaData();
    ResultSet rset = dbmd.getTables(null, null, "%", null);
    while (rset.next()) {
	String name = rset.getString("TABLE_NAME");
	Table t = Table.find(name);
	if (t == null)
	    t = new Table(this, dbmd, name);
	else {
	    // Make table compare old and new columns, but only complain
	    // the first time we see a missing column.
	    if (!missingColumn && !t.loadColumns(dbmd))
		missingColumn = true;
	}
	newTableList.add(t);
    }
    rset.close();

    // If we are reloading information, make sure all tables in the old list
    // are also in the new list.
    if (tables != null) {
	for (Iterator iter = tables.iterator(); iter.hasNext(); ) {
	    Table t = (Table)iter.next();
	    if (!newTableList.contains(t)) {
		ErrorHandler.error("Warning: table " + t.getName()
				   + " not found in database");
		break;
	    }
	}
    }

    tables = newTableList;
}

/**
 * Returns <code>true</code> if the database parameters have been modified.
 *
 * @return <code>true</code> if the database parameters have been modified
 */
public boolean isDirty() { return dirty; }

/**
 * Sets or clears the dirty bit.
 *
 * @param dirty the new value
 */
public void setDirty(boolean dirty) { this.dirty = dirty; }

/**
 * Returns the driver class name.
 *
 * @return the driver class name
 */
public String getDriverClassName() { return driverClassName; }

/**
 * Sets the driver class name.
 *
 * @param newDriverClassName the driver class name
 */
protected void setDriverClassName(String newDriverClassName) {
    if (driverClassName != newDriverClassName
	&& (driverClassName == null
	    || !driverClassName.equals(newDriverClassName)))
    {
	driverClassName = newDriverClassName;
	setDirty(true);
    }

}

/**
 * Returns the connection info string.
 *
 * @return the connection info string
 */
public String getConnectionInfo() { return connInfo; }

/**
 * Sets the connection info string.
 *
 * @param newConnectionInfo the connection info string
 */
protected void setConnectionInfo(String newConnectionInfo) {
    if (connInfo != newConnectionInfo
	&& (connInfo == null || !connInfo.equals(newConnectionInfo)))
    {
	connInfo = newConnectionInfo;
	setDirty(true);
    }
}

/**
 * Returns the database name.
 *
 * @return the database name
 */
public String getName() { return name; }

/**
 * Sets the name.
 *
 * @param newName the new name
 */
protected void setName(String newName) {
    if (name != newName && (name == null || !name.equals(newName))) {
	name = newName;
	setDirty(true);
    }
}

/**
 * Returns the user name.
 *
 * @return the user name
 */
public String getUserName() { return username; }

/**
 * Sets the user name.
 *
 * @param newUserName the new user name
 */
public void setUserName(String newUserName) {
    if (newUserName == null)
	newUserName = "";

    if (username != newUserName
	&& (username == null || !username.equals(newUserName)))
    {
	username = newUserName;
	setDirty(true);
    }
}

/**
 * Returns the password.
 *
 * @return the password
 */
public String getPassword() { return hasPassword ? password : null; }

/**
 * Sets the password.
 *
 * @param newPassword the new password
 */
public void setPassword(String newPassword) {
    password = newPassword;
    hasPassword = true;
}

/**
 * Returns an iterator over all the tables in the database.
 *
 * @return an iterator over all the tables in the database
 */
public Iterator tables() { return tables.iterator(); }

/**
 * Writes this database and all its tables as an XML tag.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<database driverClassName=\"" + driverClassName + '"');
    out.indent();
    out.indent();
    out.println("connInfo=\"" + connInfo + '"');
    out.println("name=\"" + (name == null ? "" : name) + "\" username=\""
		+ username + "\" />");
    out.outdent();
    out.outdent();
}

}
