package jimm.datavision.sql;
import jimm.datavision.ErrorHandler;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * Represents a database table.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Table {

protected static HashMap tables = new HashMap();


protected String id;
protected Database database;
protected String name;
protected ArrayList columns;
protected boolean usedInReport;

/**
 * Given an id (a table name), returns the table that has that id. If no
 * table with the specified id exists, returns <code>null</code>.
 *
 * @return a table, or <code>null</code> if no table with the specified
 * id exists
 */
public static Table find(Object id) {
    return (Table)tables.get(id);
}

/**
 * Constructor.
 *
 * @param database the database in which this table resides
 * @param dbmd database meta data (used to read column info)
 * @param name the table's name
 */
public Table(Database database, DatabaseMetaData dbmd, String name)
    throws SQLException
{
    id = name;
    this.database = database;
    this.name = name;
    usedInReport = false;
    columns = null;
    loadColumns(dbmd);
    tables.put(id, this);	// Add to list of all tables
}

/**
 * Returns the table name.
 *
 * @return the table name
 */
public String getName() { return name; }

/**
 * Flags this table as being used in the database or not.
 *
 * @param isUsed a boolean
 */
public void setUsedInReport(boolean isUsed) { usedInReport = isUsed; }

/**
 * Returns an iterator over the columns in this table.
 *
 * @return an iterator over the columns in this table
 */
public Iterator columns() { return columns.iterator(); }

/**
 * Loads all columns in the table. Called from the constructor and also
 * by the database when it is checking to make sure that the columns in
 * this table have not changed, even when the database connection information
 * has changed. Returns <code>true</code> if this is the first load or if,
 * on reload, no old columns are missing from the new metadata.
 *
 * @param dbmd database meta data
 * @return <code>true</code> if this is the first load or if, on reload,
 * no old columns are missing from the new metadata
 */
boolean loadColumns(DatabaseMetaData dbmd) throws SQLException {
    ArrayList newColumnList = new ArrayList();

    ResultSet rset = dbmd.getColumns(null, null, name, "%");
    while (rset.next()) {
	Column c =
	    Column.find(getName() + '.' + rset.getString("COLUMN_NAME"));
	if (c == null)
	    c = new Column(this, rset);
	newColumnList.add(c);
    }
    rset.close();

    boolean missingColumn = false;

    // Compare old and new column lists; make sure all columns in the old
    // list are also in the new list.
    if (columns != null) {
	for (Iterator iter = columns.iterator();
	     !missingColumn && iter.hasNext(); )
	{
	    Column c = (Column)iter.next();
	    if (!newColumnList.contains(c)) {
		ErrorHandler.error("Warning: column " + c.fullName()
				   + " not found in database");
		missingColumn = true;
	    }
	}
    }

    columns = newColumnList;
    return !missingColumn;
}

}
